unit FEditDosLevelPack;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, VirtualTrees, ActiveX,
  UMisc, UMasterForms,
  LemTypes, LemCOre, LemDosCmp, LemDosArc,
  LemApp, LemEdit, LemStyles,
  LemDosStyles,
  FAskStyle, FBasicLevelEditor, ActnList, Menus, ComCtrls;

type
  TAddMode = (
    amReplace,
    amAfterSelected,
    amBeforeSelected,
    amAppend
  );

type
  TFormEditDosLevelPack = class(TBasicLevelEditorForm)
    Tree: TVirtualStringTree;
    MainMenu: TMainMenu;
    ActionList: TActionList;
    StatusBar: TStatusBar;
    PopupMenu: TPopupMenu;
    procedure Form_Create(Sender: TObject);
    procedure Form_Destroy(Sender: TObject);
    procedure TreeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType;
      var CellText: WideString);
    procedure TreeChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure TreeKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure TreeDragDrop(Sender: TBaseVirtualTree; Source: TObject;
      DataObject: IDataObject; Formats: TFormatArray; Shift: TShiftState;
      Pt: TPoint; var Effect: Integer; Mode: TDropMode);
    procedure TreeDragAllowed(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; var Allowed: Boolean);
    procedure TreeDragOver(Sender: TBaseVirtualTree; Source: TObject;
      Shift: TShiftState; State: TDragState; Pt: TPoint; Mode: TDropMode;
      var Effect: Integer; var Accept: Boolean);
    procedure ActionList_Update(Action: TBasicAction; var Handled: Boolean);
  private
    fDosArchive   : TDosDatArchiveEx;
//    fFileName     : string;
    fCurrentStyle : TBaseLemmingStyle;
    fLoadedIndex  : Integer;

  { file ... }
//    ActOpenWithEditor   : TAction;
    ActSave             : TAction;
    ActSaveAs           : TAction;

//    ActSaveEditorLevel      : TAction;

    ActSaveEditorAppend     : TAction;
    ActSaveEditorAfter      : TAction;
    ActSaveEditorBefore     : TAction;
    ActSaveEditorReplace    : TAction;

    ActAddFileAppend    : TAction;
    ActAddFileAfter     : TAction;
    ActAddFileBefore    : TAction;
    ActAddFileReplace   : TAction;
    ActRemoveCurrent    : TAction;

    ActMoveLevelUp      : TAction;
    ActMoveLevelDown    : TAction;
  { options ... }
//    ActToggleAutoPreview: TAction;
    ActChangeCurrentStyle: TAction;

    procedure SetFileName(const Value: string);
    procedure FillTree;
    function AutoDetectStyle(const aFileName: string): TBaseLemmingStyle;
    procedure Form_DropFiles(Sender: TObject; const DroppedFiles: TStrings);
    procedure InternalAddFile(Mode: TAddMode);
    procedure InternalAddEditorLevel(Mode: TAddMode);
    procedure InternalRemoveSection;
  { action event handlers }
    procedure ActAddFileAppend_Execute(Sender: TObject);
    procedure ActAddFileBefore_Execute(Sender: TObject);
    procedure ActAddFileAfter_Execute(Sender: TObject);
    procedure ActAddFileReplace_Execute(Sender: TObject);
//    procedure ActSaveEditorLevel_Execute(Sender: TObject);

    procedure ActSaveEditorAppend_Execute(Sender: TObject);
    procedure ActSaveEditorAfter_Execute(Sender: TObject);
    procedure ActSaveEditorBefore_Execute(Sender: TObject);
    procedure ActSaveEditorReplace_Execute(Sender: TObject);


    procedure ActSave_Execute(Sender: TObject);
    procedure ActSaveAs_Execute(Sender: TObject);
    procedure ActRemoveCurrent_Execute(Sender: TObject);
//    procedure ActOpenWithEditor_Execute(Sender: TObject);
//    procedure ActExport_Execute(Sender: TObject);
    procedure ActMoveLevelDown_Execute(Sender: TObject);
    procedure ActMoveLevelUp_Execute(Sender: TObject);
//    procedure ActToggleAutoPreview_Execute(Sender: TObject);
    procedure ActChangeCurrentStyle_Execute(Sender: TObject);
    procedure ActNewLevelPack_Execute(Sender: TObject);
  { internal }
    procedure CreateActions2;
    procedure SetCurrentStyle(const Value: TBaseLemmingStyle);
    function GetFileName: string;
  public
    ActNewLevelPack: TAction;
    procedure SaveInPack;
    property FileName: string read GetFileName write SetFileName;
    property CurrentStyle: TBaseLemmingStyle read fCurrentStyle write SetCurrentStyle;
  end;

function FormEditDosLevelPack: TFormEditDosLevelPack;
function FormEditDosLevelPackActive: Boolean;

implementation

{$R *.dfm}

uses
  FMapEditor;

var
  _Form: TFormEditDosLevelPack;

function FormEditDosLevelPackActive: Boolean;
begin
  Result := _Form <> nil;
end;

function FormEditDosLevelPack: TFormEditDosLevelPack;
var
  i: Integer;
begin
  with Screen do

  for i := 0 to CustomFormCount - 1 do
    if CustomForms[i] is TFormEditDosLevelPack then
    begin
      Result := TFormEditDosLevelPack(CustomForms[i]);
      //Result.CurrentStyle := aStyle;
      Exit;
    end;


  Result := TFormEditDosLevelPack.Create(Application);
//  Result.CurrentStyle := aStyle;

end;

{ TFormDosLevelPacks }

procedure TFormEditDosLevelPack.FillTree;
begin
  with fDosArchive do
  begin
    Tree.RootNodeCount := 0; //virtualtrees
//    fDosArchive.LoadFromFile;
    Tree.RootNodeCount := SectionList.Count;
//    Tree.Invalidate;
  end;
end;

procedure TFormEditDosLevelPack.SetFileName(const Value: string);
var
  Sty: TBaseLemmingStyle;
begin
  Tree.BeginUpdate;
  try
//    fFileName := Value;
    Caption := 'Edit DOS levelpack ' + '[' + ExtractFileName(Value) + ']';
    fDosArchive.FileName := Value;//LoadFromFile(fFileName);
    fDosArchive.LoadFromFile;
    Sty := AutoDetectStyle(Value);
    if Sty <> nil then
      CurrentStyle := Sty;
//    FillTree;
  finally
    FillTree;
    Tree.EndUpdate;
  end;
end;

procedure TFormEditDosLevelPack.Form_Create(Sender: TObject);
var
  S: string;
  Sty: TBaseLemmingStyle;
begin
  _Form := Self;
  fDosArchive := TDosDatArchiveEx.Create;
//  CurrentStyle := DosOrigStyle;
  fLoadedIndex := -1;
  AcceptDraggedFiles := True;
  OnDropFiles := Form_DropFiles;
  Editor := GlobalEditor;

  S := App.EditorSettings.DosPackPreviewStyleName;
  Sty := StyleMgr.FindStyleByName(S);
  if Sty = nil then
    Sty := DosOrigStyle;//StyleMgr[style_dosorig];
  CurrentStyle := Sty;
  CreateActions2;
end;

procedure TFormEditDosLevelPack.Form_Destroy(Sender: TObject);
begin
  if fCurrentStyle <> nil then
    App.EditorSettings.DosPackPreviewStyleName := fCurrentStyle.StyleName
  else
    App.EditorSettings.DosPackPreviewStyleName := '';

  _Form := nil;
  fDosArchive.Free;
end;

procedure TFormEditDosLevelPack.TreeGetText(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
  var CellText: WideString);
var
  B: PBytes;
  S: string;
begin
  if Integer(Node^.Index) > fdosarchive.sectionlist.count-1 then
  begin
    celltext := '<ERROR>';
    exit;
  end;

  with fdosarchive.sectionlist[node^.index].DecompressedData do
  begin
    B := Memory;
    S := stringofchar(' ', 32);
    move(B^[2048-32], s[1], 32);
  end;
  CellText := Trim(S);
end;

procedure TFormEditDosLevelPack.TreeChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
var
  S: TStream;
begin
  if Editor.IsLocked then Exit;
  if node<>tree.FocusedNode then exit;

  S := fdosarchive.sectionlist[node^.index].DecompressedData;
  S.Seek(0, soFromBeginning);
  Assert(fCurrentStyle <> nil);

  //Editor.LoadFromDosArchive(fDosArchive, Node^.Index, fCurrentStyle);
  Editor.LoadFromStream(fDosArchive.SectionList[Node^.Index].DecompressedData,
    fCurrentStyle, lffLVL);

  fLoadedIndex := node^.index; // lemedit
  Editor.Loader := Self;

end;

procedure TFormEditDosLevelPack.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  if Editor.Loader = Self then
    Editor.Loader := nil;

//  GlobalEditor.CloseLevel;
  //GlobalSaver := nil;
//  if Self = GlobalSaver then
  //  Glova
//  if GlobalSaver = Self then
  //  GlobalSaver := nil;
  Action := caFree;
end;

function TFormEditDosLevelPack.AutoDetectStyle(const aFileName: string): TBaseLemmingStyle;
var
  i: Integer;
  P: string;
begin
  P := ExtractFilePath(aFileName);
  for i := 0 to StyleMgr.StyleList.Count - 1 do
  begin
    Result := StyleMgr[i];
    if CompareText(Result.CommonPath, P) = 0 then
      Exit;
  end;
  Result := nil;
end;


procedure TFormEditDosLevelPack.Form_DropFiles(Sender: TObject;
  const DroppedFiles: TStrings);
var
  F: TFileStream;
  TryLevel: Boolean;
  P: TPoint;
  N: PVirtualNode;
begin

  P := Mouse.CursorPos;
  P := Tree.ScreenToClient(P);
  if (p.x < 0) or (p.y < 0) or (p.x > tree.Width - 1) or (p.y > tree.Height - 1) then
    n := nil
  else begin
    N := Tree.GetNodeAt(P.X, P.Y);
  end;  
    //deb([tree.Text[n,0]]);


  F := TFileStream.Create(DroppedFiles[0], fmOpenRead);
  TryLevel := F.Size = 2048;
  F.Free;
  if TryLevel then
  begin
    Tree.BeginUpdate;
    try
      if N <> nil then
        fDosArchive.InsertFileAt(DroppedFiles[0], N^.Index)
      else
        fDosArchive.InsertFileAt(DroppedFiles[0], fDosArchive.SectionList.Count);
    finally
      FillTree;
      Tree.EndUpdate;
    end;
  end
  else
    Filename := DroppedFiles[0];
end;

procedure TFormEditDosLevelPack.SaveInPack;
//var
  //S: TMemoryStream;
begin
  raise exception.create('not implemented');
(*  Tree.RootNodeCount := 0;

//  fDosArchive.ReadOnly := False;
  S := TMemoryStream.Create;
  Editor.SaveToLVLStream(S);
//  fDosArchive.Replace(fLoadedIndex, S);
  S.Free;

  FillTree; *)

end;

procedure TFormEditDosLevelPack.TreeKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  (*
  case Key of
    VK_DELETE:
      ActRemoveCurrent.Execute;
    VK_INSERT:
      ActAddFileBefore.Execute; }
{    VK_UP:
      if ssCtrl in Shift then
        ActMoveLevelUp.Execute;
    VK_DOWN:
      if ssCtrl in Shift then
        ActMoveLevelDown.Execute; }
  end;
  *)
end;

procedure TFormEditDosLevelPack.CreateActions2;

    function NewHolderItem(aParent: TMenuItem; const aCaption: string): TMenuItem;
    begin
      if aParent = nil then
        aParent := PopupMenu.Items;
      Result := TMenuItem.Create(Self);
      Result.Caption := aCaption;
      aParent.Add(Result)
    end;

    procedure NewSep(aParent: TMenuItem);
    var
      M: TMenuItem;
    begin
      if aParent = nil then
        aParent := PopupMenu.Items;
      M := TMenuItem.Create(Self);
      M.Caption := '-';
      aParent.Add(M)
    end;

    function NewAction(aParent: TMenuItem;
                       const aCaption: string;
                       aExecuteHandler: TNotifyEvent;
                       aShortCut: TShortCut = 0): TAction;

    var
      M: TMenuItem;
    begin
      if aParent = nil then
        aParent := PopupMenu.Items;
      Result := TAction.Create(Self);
      Result.Caption := aCaption;
      Result.OnExecute := aExecuteHandler;
      Result.ShortCut  := aShortCut;
      Result.ActionList := ActionList;
      if aParent <> nil then
      begin
        M := TMenuItem.Create(Self);
        M.Action := Result;
        aParent.Add(M);
      end
    end;

var
  MhAdd: TMenuItem;
  MhEditorAdd: TMenuItem;
begin
  ActNewLevelPack := NewAction(nil, 'New levelpack...', ActNewLevelPack_Execute);
  ActSave := NewAction(nil, 'Save levelpack', ActSave_Execute);
  ActSaveAs := NewAction(nil, 'Save levelpack as...', ActSaveAs_Execute);
  NewSep(nil);

//  ActSaveEditorLevel := NewAction(nil, 'Save currently loaded level in archive',
  //  ActSaveEditorLevel_Execute);
//  NewSep(nil);

  MhAdd := NewHolderItem(nil, 'Add LVL file');
  ActAddFileReplace := NewAction(mhAdd, 'Replace current level...',
    ActAddFileReplace_Execute);
  ActAddFileAppend := NewAction(mhAdd, 'At the end...',
    ActAddFileAppend_Execute, ShortCut(VK_INSERT, []));
  ActAddFileBefore := NewAction(mhAdd, 'Before current level...',
    ActAddFileBefore_Execute, ShortCut(VK_INSERT, [ssCtrl]));
  ActAddFileAfter := NewAction(mhAdd, 'After current level...', ActAddFileAfter_Execute);

//  NewSep(nil);

  MhEditorAdd := NewHolderItem(nil, 'Add Editor level');
  ActSaveEditorReplace := NewAction(mhEditorAdd, 'Replace current level',
    ActSaveEditorReplace_Execute);
  ActSaveEditorAppend := NewAction(mhEditorAdd, 'At the end',
    ActSaveEditorAppend_Execute);
  ActSaveEditorAfter := NewAction(mhEditorAdd, 'After current level',
    ActSaveEditorAfter_Execute);
  ActSaveEditorBefore := NewAction(mhEditorAdd, 'Before current level',
    ActSaveEditorBefore_Execute);

  NewSep(nil);


  ActMoveLevelUp := NewAction(nil, 'Move current level up', ActMoveLevelUp_Execute,
    ShortCut(VK_UP, [ssCtrl]));
  ActMoveLevelDown := NewAction(nil, 'Move current level down', ActMoveLevelDown_Execute,
    ShortCut(VK_DOWN, [ssCtrl]));
  ActRemoveCurrent := NewAction(nil, 'Remove current level',
    ActRemoveCurrent_Execute, ShortCut(VK_DELETE, [ssCtrl]));

  NewSep(nil);
  ActChangeCurrentStyle := NewAction(nil, 'Change viewer style...', ActChangeCurrentStyle_Execute);

end;

procedure TFormEditDosLevelPack.ActAddFileAppend_Execute(Sender: TObject);
begin
  InternalAddFile(amAppend);
end;


procedure TFormEditDosLevelPack.ActAddFileBefore_Execute(Sender: TObject);
begin
  InternalAddFile(amBeforeSelected);
end;

procedure TFormEditDosLevelPack.ActAddFileAfter_Execute(Sender: TObject);
begin
  InternalAddFile(amAfterSelected);
end;

procedure TFormEditDosLevelPack.ActAddFileReplace_Execute(Sender: TObject);
begin
  InternalAddFile(amReplace);
end;

procedure TFormEditDosLevelPack.ActRemoveCurrent_Execute(Sender: TObject);
begin
  InternalRemoveSection;
end;

procedure TFormEditDosLevelPack.ActSaveAs_Execute(Sender: TObject);
var
  Dlg: TSaveDialog;
begin
  Dlg := TSaveDialog.Create(nil);
  try
    Dlg.FileName := fDosArchive.FileName;
    if Dlg.Execute then
    begin
      fDosArchive.SaveToFile(Dlg.FileName);
      Caption := 'Edit DOS levelpack ' + '[' + ExtractFileName(FileName) + ']';
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TFormEditDosLevelPack.ActChangeCurrentStyle_Execute(Sender: TObject);
var
  Sty: TBaseLemmingStyle;
begin
  Sty := fCurrentStyle;
  if ExecFormAskStyle(Sty)= mrOK then
    CurrentStyle := Sty; // use property setter (statusbar)
end;

procedure TFormEditDosLevelPack.ActSave_Execute(Sender: TObject);
begin
  fDosArchive.SaveToFile;
end;

procedure TFormEditDosLevelPack.TreeDragDrop(Sender: TBaseVirtualTree;
  Source: TObject; DataObject: IDataObject; Formats: TFormatArray;
  Shift: TShiftState; Pt: TPoint; var Effect: Integer; Mode: TDropMode);
begin
//  windlg(['dataobject', dataobject <> nil]);
  if dataobject<>nil then
  begin
    windlg([length(formats)])
  end;
// virtualtrees
end;

procedure TFormEditDosLevelPack.SetCurrentStyle(const Value: TBaseLemmingStyle);
begin
  fCurrentStyle := Value;
  if fCurrentStyle <> nil then
    StatusBar.SimpleText := 'Current style: ' + fCurrentStyle.StyleDescription;
end;

procedure TFormEditDosLevelPack.InternalAddFile(Mode: TAddMode);
// load and add file
var
  Dlg: TOpenDialog;
  Node: PVirtualNode;
  Ix: Integer;
begin

  Node := Tree.FocusedNode;
  if Node = nil then
    Ix := 0
  else
    Ix := Node^.Index;

  Dlg := TOpenDialog.Create(nil);
  Dlg.FileName := '*.LVL';
  try
    if Dlg.Execute then
    begin
      case Mode of
        amAfterSelected:
          begin
            Tree.BeginUpdate;
            try
              fDosArchive.InsertFileAt(Dlg.FileName, Ix + 1);
            finally
              FillTree;
              Tree.EndUpdate;
            end;
            ///Tree.Invalidate;
          end;
        amBeforeSelected:
          begin
            Tree.BeginUpdate;
            try
              fDosArchive.InsertFileAt(Dlg.FileName, Ix);
            finally
              FillTree;
              Tree.EndUpdate;
            end;
            //Tree.Invalidate;
          end;
        amReplace:
          begin
            Tree.BeginUpdate;
            try
              fDosArchive.OverwriteFileAt(Dlg.FileName, Ix);
            finally
              FillTree;
              Tree.EndUpdate;
            end;
          end;

        amAppend:
          begin
            Tree.BeginUpdate;
            try
              fDosArchive.InsertFileAt(Dlg.FileName, fDosArchive.SectionList.Count); // at end
            finally
              FillTree;
              Tree.EndUpdate;
            end;

          end;

      end;
      //UpdateMatrix;
    end;
  finally
    Dlg.Free;
  end;


end;

procedure TFormEditDosLevelPack.InternalAddEditorLevel(Mode: TAddMode);
var
  Node: PVirtualNode;
  Ix: Integer;
  Mem: TMemoryStream;
begin

  Node := Tree.FocusedNode;
  if Node = nil then
    Ix := 0
  else
    Ix := Node^.Index;

  Mem := TMemoryStream.Create;
  Tree.BeginUpdate;
  try
    Editor.SaveToStream(Mem, lffLVL);
    case Mode of
      amAfterSelected:
        fDosArchive.InsertStreamAt(Mem, Ix + 1);
      amBeforeSelected:
        fDosArchive.InsertStreamAt(Mem, Ix);
      amReplace:
        fDosArchive.OverwriteStreamAt(Mem, Ix);
      amAppend:
        fDosArchive.InsertStreamAt(Mem, fDosArchive.SectionList.Count);
    end;
  finally
    Mem.Free;
    FillTree;
    Tree.EndUpdate;
  end;

end;


procedure TFormEditDosLevelPack.InternalRemoveSection;
var
  Node: PVirtualNode;
//  Ix: Integer;
begin
 Node := Tree.FocusedNode;
  if Node = nil then
  begin
    MessageDlg('No item selected to delete. Please first select the level you want' +
      'to delete', mtInformation, [mbOK], 0);
    Exit;
  end;
  if MessageDlg('Are you sure you want to delete this level?', mtWarning, [mbyes, mbno], 0) <> mryes then
    Exit;

  Tree.BeginUpdate;
  try
    fDosArchive.SectionList.Delete(Node^.Index);
    Tree.DeleteNode(Node);
  finally
    Tree.EndUpdate;
  end;  
end;

procedure TFormEditDosLevelPack.TreeDragAllowed(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex; var Allowed: Boolean);
begin
  Allowed := True;
end;

procedure TFormEditDosLevelPack.TreeDragOver(Sender: TBaseVirtualTree;
  Source: TObject; Shift: TShiftState; State: TDragState; Pt: TPoint;
  Mode: TDropMode; var Effect: Integer; var Accept: Boolean);
//var
 // N: PVirtualNode;
begin
//  Accept := true;
//  deb([effect]);
  {
  N := Tree.GetNodeAt(Pt.X, Pt.Y);
  if N <> nil then
    deb([n^.index])
  else
    deb(['white'])
  }
end;

procedure TFormEditDosLevelPack.ActionList_Update(Action: TBasicAction;
  var Handled: Boolean);
var
  Foc: Boolean;
  FocNode: PVirtualNode;
  HasEditLevel: Boolean;
begin
  FocNode := Tree.FocusedNode;
  Foc := FocNode <> nil;
  HasEditLevel := Assigned(Editor) and
    Assigned(Editor.Style) and Assigned(Editor.Graph)
    and (Editor.Style is TBaseDosStyle);

  Handled := True;

//  ActOpenWithEditor.Enabled := Foc;
  ActSave.Enabled := fDosArchive.FileName <> '';
  ActSaveAs.Enabled := True;
  ActAddFileAppend.Enabled := True;
  ActAddFileAfter.Enabled := Foc;
  ActAddFileBefore.Enabled := Foc;
  ActAddFileReplace.Enabled := Foc;
  ActRemoveCurrent.Enabled := Foc;



  ActSaveEditorAppend.Enabled := HasEditLevel;
  ActSaveEditorAfter.Enabled := Foc and HasEditLevel;
  ActSaveEditorBefore.Enabled     := Foc and HasEditLevel;
  ActSaveEditorReplace.Enabled    := Foc and HasEditLevel;

  ActMoveLevelUp.Enabled := Foc and (Tree.FocusedNode <> Tree.GetFirst);
  ActMoveLevelDown.Enabled := Foc and (Tree.FocusedNode <> Tree.GetLast);

{ options ... }
//  ActToggleAutoPreview: TAction;
  ActChangeCurrentStyle.Enabled := True;




(*  ActSaveEditorLevel.Enabled := Foc and
    (fDosArchive.SectionList.Count > FocNode^.Index) and
    Assigned(Editor) and
    Assigned(Editor.Style) and Assigned(Editor.Graph)
    and (Editor.Style is TBaseDosStyle);


    asdf
  Allowed := Assigned(N) and
    (fDosArchive.SectionList.Count > N^.Index) and
    Assigned(Editor) and
    Assigned(Editor.Style) and Assigned(Editor.Graph)
    and (Editor.Style is TBaseDosStyle);

  *)
  

end;

procedure TFormEditDosLevelPack.ActMoveLevelDown_Execute(Sender: TObject);
var
  N: PVirtualNode;
begin
  N := Tree.FocusedNode;
  if N = Tree.GetLast then
    Exit;
  Tree.BeginUpdate;
  try
    fDosArchive.SectionList.Move(N^.Index, N^.Index + 1);
    Tree.MoveTo(N, Tree.GetNext(N), amInsertAfter, False); // this movenode is just for focusing
  finally
    Tree.EndUpdate;
  end;
end;

procedure TFormEditDosLevelPack.ActMoveLevelUp_Execute(Sender: TObject);
var
  N: PVirtualNode;
begin
  N := Tree.FocusedNode;
  if N = Tree.GetFirst then
    Exit;
  Tree.BeginUpdate;
  try
    fDosArchive.SectionList.Move(N^.Index, N^.Index - 1);
    Tree.MoveTo(N, Tree.GetPrevious(N), amInsertBefore, False); // this movenode is just for focusing
  finally
    Tree.EndUpdate;
  end;
end;

procedure TFormEditDosLevelPack.ActNewLevelPack_Execute(Sender: TObject);
begin
  Tree.BeginUpdate;
  try
    fDosArchive.Close;
    fDosArchive.FileName := '';
    Caption := 'Edit DOS levelpack ' + '[' + ExtractFileName(FileName) + ']';
  finally
    FillTree;
    Tree.EndUpdate;
  end;
end;

procedure TFormEditDosLevelPack.ActSaveEditorAfter_Execute(Sender: TObject);
// save editorlevel after selected level
begin
  InternalAddEditorLevel(amAfterSelected);
end;

procedure TFormEditDosLevelPack.ActSaveEditorAppend_Execute(Sender: TObject);
// save editorlevel as new level at the end of the pack
begin
  InternalAddEditorLevel(amAppend);
end;

procedure TFormEditDosLevelPack.ActSaveEditorBefore_Execute(Sender: TObject);
// save editorlevel before selected level
begin
  InternalAddEditorLevel(amBeforeSelected);
end;

procedure TFormEditDosLevelPack.ActSaveEditorReplace_Execute(Sender: TObject);
// overwrite selected level with editorlevel
begin
  InternalAddEditorLevel(amReplace);
end;

function TFormEditDosLevelPack.GetFileName: string;
begin
  Result := fDosArchive.FileName;
end;

end.

(* OLD TOP-MENU ACTION METHOD
procedure TFormEditDosLevelPack.CreateActions;
var
  TheMenu:TMenu;

    function NewMenubarItem(const aCaption: string): TMenuItem;
    begin
      Result := TMenuItem.Create(Self);
      Result.Caption := aCaption;
      TheMenu.Items.Add(Result);
    end;

    function NewHolderItem(aParent: TMenuItem; const aCaption: string): TMenuItem;
    begin
      Result := TMenuItem.Create(Self);
      Result.Caption := aCaption;
      aParent.Add(Result);
    end;

    procedure NewSep(aParent: TMenuItem);
    var
      M: TMenuItem;
    begin
      if aParent <> nil then
      begin
        M := TMenuItem.Create(Self);
        M.Caption := '-';
        aParent.Add(M);
      end;
    end;

    function NewAction(aParent: TMenuItem;
                       const aCaption: string;
                       aExecuteHandler: TNotifyEvent;
                       aShortCut: TShortCut = 0): TAction;

    var
      M: TMenuItem;
    begin
      Result := TAction.Create(Self);
      Result.Caption := aCaption;
      Result.OnExecute := aExecuteHandler;
      Result.ShortCut  := aShortCut;
      Result.ActionList := ActionList;
      if aParent <> nil then
      begin
        M := TMenuItem.Create(Self);
        M.Action := Result;
        aParent.Add(M);
      end;
    end;


var
  MhAdd: TMenuItem;
begin
  TheMenu := PopupMenu;

  MhAdd := NewHolderItem(mbFile, 'Add direct saved Level');
  MbOptions := NewMenubarItem('Options');

  MbFile.GroupIndex := 5;
  MbOptions.GroupIndex := 6;

  { file }
//  ActOpenWithEditor := NewAction(mbFile, 'Open selected level with editor', ActOpenWithEditor_Execute);

  ActSave := NewAction(mbFile, 'Save', ActSave_Execute);
  ActSaveAs := NewAction(mbFile, 'Save as...', ActSaveAs_Execute);

  ActAddFileBefore := NewAction(mhAdd, 'Before current', ActAddFileBefore_Execute);
  ActAddFileAfter := NewAction(mhAdd, 'After current', ActAddFileAfter_Execute);
  ActAddFileReplace := NewAction(mhAdd, 'Replace current', ActAddFileReplace_Execute);

  ActRemoveCurrent := NewAction(mbFile, 'Remove current', ActRemoveCurrent_Execute);

  NewAction(mbFile, 'export', ActExport_Execute);


  { options }
//ActToggleAutoPreview := NewAction(mbOptions, 'Preview', ActToggleAutoPreview_Execute);
  ActChangeCurrentStyle := NewAction(mbOptions, 'Change viewer style', ActChangeCurrentStyle_Execute);

end;

(*
procedure TFormEditDosLevelPack.ActSaveEditorLevel_Execute(Sender: TObject);
var
  Allowed: Boolean;
  N: PVirtualNode;
//  Sec: TDosDatSectionEx;
  Mem: TMemoryStream;
begin
  N := Tree.FocusedNode;

  Allowed := Assigned(N) and
    (fDosArchive.SectionList.Count > Integer(N^.Index)) and
    Assigned(Editor) and
    Assigned(Editor.Style) and Assigned(Editor.Graph)
    and (Editor.Style is TBaseDosStyle);

  if Allowed then
  begin
    // editor save to tempstream
    Mem := TMemoryStream.Create;
    Tree.BeginUpdate;
    try
      Editor.SaveToStream(Mem, lffLVL);
      fDosArchive.OverwriteStreamAt(Mem, N^.Index);
    finally
      Mem.Free;
      FillTree;
      Tree.EndUpdate;
    end;


(*    Tree.BeginUpdate;
    try
      Sec := fDosArchive.SectionList[N^.Index];
      Sec.DecompressedData.Clear;
      Editor.SaveToStream(Sec.DecompressedData, lffLVL);
    finally
      FillTree;
      Tree.EndUpdate;
    end;
    *)
  end;
end;


