{$include lem_directives.inc}

unit FEditItem;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  math, umisc,  Dialogs, LemEdit, FBasicLevelEditor, StdCtrls, UMasterControls,
fpickterrain,  LemCore, ComCtrls, ExtCtrls, GR32_Image, Buttons;

type
  TTempData = record
    SelCount: Integer;
    SelTypes: TLevelItemTypes; // welke items zitten er in de selectie
    HasObjects: Boolean;
    HasTerrain: Boolean;
    HasSteel: Boolean;

    FirstBaseItem: TBaseLevelItem; // eerste

    // flags
    FirstTerrainErase: Boolean;
    FirstTerrainInvert: Boolean;
    FirstTerrainNoOverwrite: Boolean;

    FirstObjectOnlyTerrain: Boolean;
    FirstObjectInvert: Boolean;
    FirstObjectNoOverwrite: Boolean;

    SameTerrainErase: Boolean;
    SameTerrainInvert: Boolean;
    SameTerrainNoOverwrite: Boolean;

    SameObjectOnlyTerrain: Boolean;
    SameObjectInvert: Boolean;
    SameObjectNoOverwrite: Boolean;

    // positions
    FirstX: Integer;
    FirstY: Integer;
    FirstW: Integer;
    FirstH: Integer;

    SameX: Boolean;
    SameY: Boolean;
    SameW: Boolean;
    SameH: Boolean;

    // id's
    FirstTerrainID: Integer;
    FirstObjectID: Integer;

    SameTerrainID: Boolean;
    SameObjectID: Boolean;

    // index
    TheIndex: Integer;

    // helpers
    //GrSet: Integer;
    GrSet: TBaseGraphicSet;

  end;

type
  TEditItemForm = class(TBasicLevelEditorForm)
    EditX: TEditEx;
    EditY: TEditEx;
    EditW: TEditEx;
    EditH: TEditEx;
    EditTID: TEditEx;
    EditOID: TEditEx;
    EditIx: TEditEx;

    StatusBar: TStatusBar;
    LabelX: TLabel;
    LabelY: TLabel;
    LabelW: TLabel;
    LabelH: TLabel;
    LabelTerrain: TLabel;
    LabelIndex: TLabel;
    LabelObject: TLabel;
    CbTerrainErase: TCheckBox;
    CbTerrainUpside: TCheckBox;
    CbTerrainNoOverwrite: TCheckBox;
    CBObjectOnlyOnTerrain: TCheckBox;
    CbObjectUpside: TCheckBox;
    CbObjectNoOverwrite: TCheckBox;
    Bevel1: TBevel;
    Bevel2: TBevel;
    Bevel3: TBevel;
    ImgTerrain: TImage32;
    ImgObject: TImage32;
    BtnPickTerrain: TSpeedButton;
    BtnPickObject: TSpeedButton;
    procedure Form_Create(Sender: TObject);
    procedure Form_Destroy(Sender: TObject);
    procedure Form_Deactivate(Sender: TObject);
    procedure Edit_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure CbTerrain_Click(Sender: TObject);
    procedure CbObject_Click(Sender: TObject);
    procedure StatusBar_DrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel; const Rect: TRect);
    procedure Edit_Exit(Sender: TObject);
    procedure BtnPickTerrain_Click(Sender: TObject);
    procedure BtnPickObject_Click(Sender: TObject);
  private
    fLabels: TList;
    fEdits: TList;
    fExiting: Boolean;
//    Sending: Boolean;
    Clicking: Boolean;
    Upd: Boolean;
    CurrentInfo: TTempData;
    procedure LMLevelChanged(var Msg: TLMLevelChanged); message LM_LEVELCHANGED;
    procedure LMObjectListChanged(var Msg: TLMObjectListChanged); message LM_OBJECTLISTCHANGED;
    procedure LMTerrainListChanged(var Msg: TLMTerrainListChanged); message LM_TERRAINLISTCHANGED;
    procedure LMSteelListChanged(var Msg: TLMSteelListChanged); message LM_STEELLISTCHANGED;
    procedure LMSelectionChanged(var Msg: TLMSelectionChanged); message LM_SELECTIONCHANGED;

    procedure UpdateIDS(aType: TLevelItemType; V: Integer);
    procedure UpdateEditVar(Sender: TEditEx);
    procedure UpdateCheckBoxVar(Sender: TCheckBox);

    procedure DoPickList(aType: TLevelItemType);

    procedure UpdateInterface;
    procedure DisableInterface;
    procedure ClearContents;
    function GetInfo(var Inf: TTempData): Boolean;
  protected
    procedure DoAfterSetEditor(Value: TLevelEditor); override;
  public
  end;

implementation

{$R *.dfm}

uses
  LemApp;

const
  tag_x                 = 1;
  tag_y                 = 2;
  tag_w                 = 3;
  tag_h                 = 4;
  tag_tid               = 5;
  tag_oid               = 6;
  tag_ter_erase         = 7;
  tag_ter_invert        = 8;
  tag_ter_nooverwrite   = 9;
  tag_obj_onlyterrain   = 10;
  tag_obj_invert        = 11;
  tag_obj_nooverwrite   = 12;
  tag_ix                = 20;

  panel_colorblocks     = 0;
  panel_seltext         = 1;


{ TEditItemForm }

procedure TEditItemForm.LMLevelChanged(var Msg: TLMLevelChanged);
begin
  UpdateInterface;
end;

procedure TEditItemForm.LMObjectListChanged(var Msg: TLMObjectListChanged);
begin
  UpdateInterface;
end;

procedure TEditItemForm.LMSelectionChanged(var Msg: TLMSelectionChanged);
begin
  UpdateInterface;
end;

procedure TEditItemForm.LMSteelListChanged(var Msg: TLMSteelListChanged);
begin
  UpdateInterface;
end;

procedure TEditItemForm.LMTerrainListChanged(var Msg: TLMTerrainListChanged);
begin
  UpdateInterface;
end;

procedure TEditItemForm.Form_Create(Sender: TObject);
const
  IntChars = ['0'..'9', '-'];
begin

   {
   with Constraints do
   begin
     MinWidth := Width;
     MaxWidth := Width;
     MinHeight := Height;
     MaxHeight := Height;
   end; }

  fEdits := TList.Create;
  fEdits.Add(EditX);
  fEdits.Add(EditY);
  fEdits.Add(EditW);
  fEdits.Add(EditH);

  EditTID.Tag := tag_tid;
  EditOID.Tag := tag_oid;

  CbTerrainErase.Tag := tag_ter_erase;
  CbTerrainUpside.Tag := tag_ter_invert;
  CbTerrainNoOverwrite.Tag := tag_ter_nooverwrite;

  CBObjectOnlyOnTerrain.Tag := tag_obj_onlyterrain;
  CbObjectUpside.Tag := tag_obj_invert;
  CbObjectNoOverwrite.Tag := tag_obj_nooverwrite;


  EditX.Tag := tag_x;
  EditY.Tag := tag_y;
  EditW.Tag := tag_w;
  EditH.Tag := tag_h;

  EditIx.Tag := tag_ix;

  EditX.ValidChars := IntChars;
  EditY.ValidChars := IntChars;
  EditW.ValidChars := IntChars;
  EditH.ValidChars := IntChars;
  EditTID.ValidChars := IntChars;
  EditOID.ValidChars := IntChars;
  EditIx.ValidChars := IntChars;


end;

procedure TEditItemForm.Form_Destroy(Sender: TObject);
begin
  fEdits.Free;
end;

procedure TEditItemForm.ClearContents;
begin
  EditTID.Text := '';
  EditOID.Text := '';
  EditIx.Text := '';
  EditX.Text := '';
  EditY.Text := '';
  EditW.Text := '';
  EditH.Text := '';

{  TxtMinX.Caption := '';
  TxtMinY.Caption := '';
  TxtMinW.Caption := '';
  TxtMinH.Caption := '';

  TxtMaxX.Caption := '';
  TxtMaxY.Caption := '';
  TxtMaxW.Caption := '';
  TxtMaxH.Caption := '';}

end;

//procedure

procedure TEditItemForm.DisableInterface;
begin
  //
end;

procedure TEditItemForm.Edit_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  E: TEditEx absolute Sender;
begin
  Assert(Sender is TEditEx, 'sender error');
  case Key of
    VK_RETURN : UpdateEditVar(E);
    VK_F4:
      case E.Tag of
        tag_tid: DoPickList(litTerrain);
        tag_oid: DoPickList(litObject);
      end;
  end;
end;

{procedure TEditItemForm.WriteContentsToSelection(Tag: Integer; Value: Integer)
begin

end;}

procedure TEditItemForm.UpdateIDS(aType: TLevelItemType; V: Integer);
var
  i: Integer;
  B: TBaseLevelItem;
  T: Integer;
begin
  if Editor = nil then
    Exit;

  if clicking or upd then exit;
  //clicking:=true;
  try

  with Editor.SelectorCollection.HackedList do
  begin
    if Count = 0 then
      Exit;
    Editor.BeginUpdate;
    try

      for i := 0 to Count - 1 do
      begin
        B := TSelector(List^[i]).RefItem;
        case aType of
          litTerrain:
            begin
              if b.itemtype = litterrain then
                if tterrain(b).ValidTerrainID(v) then
                   b.Identifier := v;
            end;
          litObject:
            begin
              if b.itemtype = litObject then
                if TInteractiveObject(b).ValidObjectID(v) then
                  b.Identifier := v;
            end;
        end;
      end;

  finally
    Editor.EndUpdate;
  end;

  end;

  finally
    //clicking := false;
  end;


end;

procedure TEditItemForm.UpdateEditVar(Sender: TEditEx);
var
  i: Integer;
  B: TBaseLevelItem;
  V, T: Integer;
begin
  if Editor = nil then
    Exit;

  if not Sender.Modified then
    Exit;

  if clicking or upd then exit;
  clicking:=true;
  try

  with Editor.SelectorCollection.HackedList do
  begin
    if Count = 0 then
      Exit;
    T := Sender.Tag;
    V := StrToIntDef(Sender.Text, MinInt);
    if V = MinInt then raise exception.create('invalid input');


//    Log(['updateeditvarr']);
    Editor.BeginUpdate;
    try

{      if T = tag_x then
      begin

        V := V - CurrentInfo.FirstBaseItem.XPosition;
        Editor.MoveSelection(V, 0);
        Sender.Modified := False;
        Exit;
      end;}

      for i := 0 to Count - 1 do
      begin
        B := TSelector(List^[i]).RefItem;
        case T of
          tag_tid:
            if b.itemtype = litterrain then
              if tterrain(b).ValidTerrainID(v) then
                b.Identifier := v;
          tag_oid:
            if b.itemtype = litObject then
              if TInteractiveObject(b).ValidObjectID(v) then
                b.Identifier := v;
          tag_x:
            begin

              b.XPosition := V;
            end;
          tag_y: b.YPosition := V;
          tag_w:
            if b.itemtype=litSteel then
              b.ItemWidth := V;
          tag_h:
            if b.itemtype=litSteel then
              b.ItemHeight := V;
          tag_ix:
            //if
            if count=1 then
            b.index := V;
        end;
      end;

    Sender.Modified := False;
  finally
    Editor.EndUpdate;
  end;

  end;

  finally
    clicking := false;
  end;


end;

procedure TEditItemForm.UpdateCheckBoxVar(Sender: TCheckBox);
{-------------------------------------------------------------------------------
  This is the main "checkbox change" method, called by the OnClick eventhandler.
  It's rude and stupid but it works.
-------------------------------------------------------------------------------}
var
  i: Integer;
  B: TBaseLevelItem;
  C: Boolean;
  T: Integer;
  Ter: TTerrain;
  Obj: TInteractiveObject;
begin
  if Editor = nil then
    Exit;

  with Editor.SelectorCollection.HackedList do
  begin
    if Count = 0 then
      Exit;
    T := Sender.Tag;
    C := Sender.Checked;

    Editor.BeginUpdate;
    try

      for i := 0 to Count - 1 do
      begin
        B := TSelector(List^[i]).RefItem;

        case T of
          // terrain
          tag_ter_erase:
            if b.itemtype = litterrain then
            begin
              Ter := TTerrain(B);
              if C
              then Ter.TerrainDrawingFlags := Ter.TerrainDrawingFlags + [tdfErase]
              else Ter.TerrainDrawingFlags := Ter.TerrainDrawingFlags - [tdfErase]
            end;

          tag_ter_invert:
            if b.itemtype = litterrain then
            begin
              Ter := TTerrain(B);
              if C
              then Ter.TerrainDrawingFlags := Ter.TerrainDrawingFlags + [tdfInvert]
              else Ter.TerrainDrawingFlags := Ter.TerrainDrawingFlags - [tdfInvert]
            end;

          tag_ter_nooverwrite:
            if b.itemtype = litterrain then
            begin
              Ter := TTerrain(B);
              if C
              then Ter.TerrainDrawingFlags := Ter.TerrainDrawingFlags + [tdfNoOverwrite]
              else Ter.TerrainDrawingFlags := Ter.TerrainDrawingFlags - [tdfNoOverwrite]
            end;

          // object
          tag_obj_onlyterrain:
            if b.itemtype = litObject then
            begin
              Obj := TInteractiveObject(B);
              if C
              then Obj.ObjectDrawingFlags := Obj.ObjectDrawingFlags + [odfOnlyShowOnTerrain]
              else Obj.ObjectDrawingFlags := Obj.ObjectDrawingFlags - [odfOnlyShowOnTerrain]
            end;

          tag_obj_invert:
            if b.itemtype = litObject then
            begin
              Obj := TInteractiveObject(B);
              if C
              then Obj.ObjectDrawingFlags := Obj.ObjectDrawingFlags + [odfInvert]
              else Obj.ObjectDrawingFlags := Obj.ObjectDrawingFlags - [odfInvert]
            end;

          tag_obj_nooverwrite:
            if b.itemtype = litObject then
            begin
              Obj := TInteractiveObject(B);
              if C
              then Obj.ObjectDrawingFlags := Obj.ObjectDrawingFlags + [odfNoOverwrite]
              else Obj.ObjectDrawingFlags := Obj.ObjectDrawingFlags - [odfNoOverwrite]
            end;

        end;
      end;

    finally
      Editor.EndUpdate;
    end;

  end;


end;


procedure TEditItemForm.CbTerrain_Click(Sender: TObject);
var
  C: TCheckBox absolute Sender;
begin
  if Clicking or Upd then Exit; // kut!
  Assert(Sender is TCheckBox, 'sender error');
  Clicking := True;
  try
    UpdateCheckBoxVar(C);
  finally
    Clicking := False;
  end;
end;

procedure TEditItemForm.CbObject_Click(Sender: TObject);
var
  C: TCheckBox absolute Sender;
begin
  if Clicking or Upd then Exit; // kut!
  Assert(Sender is TCheckBox, 'sender error');
  Clicking := True;
  try
    UpdateCheckBoxVar(C);
  finally
    Clicking := False;
  end;
end;

procedure TEditItemForm.DoAfterSetEditor(Value: TLevelEditor);
begin
  UpdateInterface;
end;

procedure TEditItemForm.StatusBar_DrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel; const Rect: TRect);
var
  D: TRect;
  H: INteger;
begin
  if Panel.Index = panel_colorblocks then
  begin
    with StatusBar.Canvas do
    begin
      Pen.Color := clGray;
      D := Rect;
      InflateRect(D, -2, -2);
      D.Right := D.Left + (D.Bottom - D.Top);
      H := D.Bottom - D.Top;

      if CurrentInfo.HasObjects then
        Brush.Color := ObjColorH
      else
        Brush.Color := clBtnFace;
      Rectangle(D);

      if CurrentInfo.HasTerrain then
        Brush.Color := TerColorH
      else
        Brush.Color := clBtnFace;
      RectMove(D, H + 2, 0);
      Rectangle(D);

      if CurrentInfo.HasSteel then
        Brush.Color := StlColorH
      else
        Brush.Color := clBtnFace;
      RectMove(D, H + 2, 0);
      Rectangle(D);
    end;
  end;
end;

function TEditItemForm.GetInfo(var Inf: TTempData): Boolean;
var
  i: Integer;
  B: TBaseLevelItem;
  Flag: Boolean;
  IsFirstTerrain, IsFirstObject: Boolean;
begin

  Result := Editor.SelectorCollection.Count > 0;
  FillChar(Inf, SizeOf(Inf), 0);
  Inf.FirstTerrainID := -1;
  Inf.FirstObjectID := -1;
  with Editor, SelectorCollection.HackedList do
  begin
    Inf.GrSet := Graph; //Statics.GraphicSet;
    if Count = 0 then
      Exit;

    for i := 0 to Count - 1 do
    begin
      B := TSelector(List^[i]).RefItem;
      if i = 0 then
        Inf.FirstBaseItem := B;
      IsFirstTerrain := (B.ItemType = litTerrain) and not (litTerrain in Inf.SelTypes);
      IsFirstObject := (B.ItemType = litObject) and not (litObject in Inf.SelTypes);
      Include(Inf.SelTypes, B.ItemType);
      // very first item
      if i = 0 then
      begin
        Inf.TheIndex := B.Index;
        Inf.SelCount := Count;
        Inf.FirstX := B.XPosition;
        Inf.FirstY := B.YPosition;
        Inf.FirstW := B.ItemWidth;
        Inf.FirstH := B.ItemHeight;
        Inf.SameX := True;
        Inf.SameY := True;
        Inf.SameW := True;
        Inf.SameH := True;
      end
      // following items
      else begin
        Inf.SameX := Inf.SameX and (B.Xposition = Inf.FirstX);
        Inf.SameY := Inf.SameY and (B.Yposition = Inf.FirstY);
        Inf.SameW := Inf.SameW and (B.ItemWidth = Inf.FirstW);
        Inf.SameH := Inf.SameH and (B.ItemHeight = Inf.FirstH);
      end;

      case B.ItemType of
        litTerrain:
          with TTerrain(B) do
          begin
            // first terrain
            if IsFirstTerrain then
            begin
              Inf.FirstTerrainErase := tdfErase in TerrainDrawingFlags;
              Inf.FirstTerrainInvert := tdfInvert in TerrainDrawingFlags;
              Inf.FirstTerrainNoOverwrite := tdfNoOverwrite in TerrainDrawingFlags;
              Inf.SameTerrainErase := True;
              Inf.SameTerrainInvert := True;
              Inf.SameTerrainNoOverwrite := True;
              Inf.FirstTerrainID := TerrainID;
              Inf.SameTerrainID := True;
            end
            // following terrain
            else begin
              Inf.SameTerrainErase := Inf.SameTerrainErase and (Inf.FirstTerrainErase = (tdfErase in TerrainDrawingFlags) );
              Inf.SameTerrainInvert := Inf.SameTerrainInvert and ( Inf.FirstTerrainInvert = (tdfInvert in TerrainDrawingFlags));
              Inf.SameTerrainNoOverwrite := Inf.SameTerrainNoOverwrite and (Inf.FirstTerrainNoOverwrite = (tdfNoOverwrite in TerrainDrawingFlags));
              Inf.SameTerrainID := Inf.SameTerrainID and (Inf.FirstTerrainID = TerrainID);
            end;
          end;
        litObject:
          with TInteractiveObject(B) do
          begin
            // first object
            if IsFirstObject then
            begin
              Inf.FirstObjectOnlyTerrain := odfOnlyShowOnTerrain in ObjectDrawingFlags;
              Inf.FirstObjectInvert := odfInvert in ObjectDrawingFlags;
              Inf.FirstObjectNoOverwrite := odfNoOverwrite in ObjectDrawingFlags;
              Inf.SameObjectOnlyTerrain := True;
              Inf.SameObjectInvert := True;
              Inf.SameObjectNoOverwrite := True;
              Inf.FirstObjectID := ObjectID;
              Inf.SameObjectID := True;
            end
            // following object
            else begin
              Inf.SameObjectOnlyTerrain := Inf.SameObjectOnlyTerrain and (Inf.FirstObjectOnlyTerrain = (odfOnlyShowOnTerrain in ObjectDrawingFlags));
              Inf.SameObjectInvert := Inf.SameObjectInvert and (Inf.FirstObjectInvert = (odfInvert in ObjectDrawingFlags));
              Inf.SameObjectNoOverwrite := Inf.SameObjectNoOverwrite and (Inf.FirstObjectNoOverwrite = (odfNoOverwrite in ObjectDrawingFlags));
              Inf.SameObjectID := Inf.SameObjectID and (Inf.FirstObjectID = ObjectID);
            end;
          end;
        litSteel:
          begin
          end;

      end; // case

    end; // for i

    Inf.HasObjects := litObject in Inf.SelTypes;
    Inf.HasTerrain := litTerrain in Inf.SelTypes;
    Inf.HasSteel := litSteel in Inf.SelTypes;


  end; // with
end;

procedure TEditItemForm.UpdateInterface;
var
  Inf: TTempData;
  HasTerrain, HasObjects: Boolean;

    procedure SetCheckBox(aCheckBox: TCheckBox; First, All: Boolean);
    begin
      if not All
      then aCheckBox.State := cbGrayed
      else aCheckBox.Checked := First;
    end;

begin
  if Clicking or Upd then
    Exit;

  if Editor = nil then
  begin
    DisableInterface;
    Exit;
  end;

  Upd := True;
  try
    GetInfo(Inf);
    CurrentInfo := Inf;


    if Inf.SelCount > 1 then
    begin
      StatusBar.Panels[1].Text := i2s(Inf.SelCount) + ' items selected';
    end
    else if Inf.SelCount = 1 then begin
      StatusBar.Panels[1].Text := '1 item selected';
    end
    else if Inf.SelCount = 0 then
      StatusBar.Panels[1].Text := 'No items selected';

    statusbar.invalidate;
    //

    //general
    EditX.Enabled := Inf.SelCount > 0;
    EditY.Enabled := Inf.SelCount > 0;
    EditW.Enabled := Inf.SelCount > 0;
    EditH.Enabled := Inf.SelCount > 0;
    EditIx.Enabled := Inf.SelCount = 1;
    LabelX.Enabled := Inf.SelCount > 0;
    LabelY.Enabled := Inf.SelCount > 0;
    LabelW.Enabled := Inf.SelCount > 0;
    LabelH.Enabled := Inf.SelCount > 0; //controls
    if Inf.SelCount = 0 then begin
      EditX.Text := '';
      EditY.Text := '';
      EditW.Text := '';
      EditH.Text := '';
    end
    else begin
      EditX.Text := i2s(Inf.FirstX);
      EditY.Text := i2s(Inf.FirstY);
      EditW.Text := i2s(Inf.FirstW);
      EditH.Text := i2s(Inf.FirstH);
    end;

    if (Inf.SelCount = 1)
    then EditIx.Text := i2s(Inf.TheIndex)
    else EditIx.Text := '';
    LabelIndex.Enabled := Inf.SelCount = 1;

    HasTerrain := litTerrain in Inf.SelTypes;
    HasObjects := litObject in Inf.SelTypes;
    with Inf do
    begin
      // terrain
       EditTID.Enabled := HasTerrain;
       LabelTerrain.Enabled := HasTerrain;
       CbTerrainErase.Enabled := HasTerrain;
       CbTerrainUpside.Enabled := HasTerrain;
       CbTerrainNoOverwrite.Enabled := HasTerrain;
       BtnPickTerrain.Enabled := HasTerrain;
      if HasTerrain then
      begin
        EditTID.Text := i2s(FirstTerrainID);
        if Inf.FirstTerrainID >= 0 then
          //ImgTerrain.Bitmap.Assign(PictureMgr.TerrainBitmaps[Inf.GrSet, Inf.FirstTerrainID]);
          ImgTerrain.Bitmap.Assign(GrSet.TerrainBitmaps[Inf.FirstTerrainID]);
        SetCheckBox(CbTerrainErase, Inf.FirstTerrainErase, Inf.SameTerrainErase);
        SetCheckBox(CbTerrainUpside, Inf.FirstTerrainInvert, Inf.SameTerrainInvert);
        SetCheckBox(CbTerrainNoOverwrite, Inf.FirstTerrainNoOverwrite, Inf.SameTerrainNoOverwrite);
      end
      else begin
        EditTID.Text := '';
        ImgTerrain.Bitmap := nil;
        CbTerrainErase.Checked := False;
        CbTerrainUpside.Checked := False;
        CbTerrainNoOverwrite.Checked := False;
      end;
      // objects
       EditOID.Enabled := HasObjects;
       LabelObject.Enabled := HasObjects;
       CBObjectOnlyOnTerrain.Enabled := HasObjects;
       CbObjectUpside.Enabled := HasObjects;
       CbObjectNoOverwrite.Enabled := HasObjects;
       BtnPickObject.Enabled := HasObjects;
      if HasObjects then
      begin
        EditOID.Text := i2s(FirstObjectID);
        if Inf.FirstObjectID >= 0 then
          //ImgObject.Bitmap.Assign(PictureMgr.ObjectBitmaps[Inf.GrSet, Inf.FirstObjectID]);
          //ImgObject.Bitmap.Assign(GrSet.ObjectBitmaps[Inf.FirstObjectID]);
//          ImgObject.Bitmap.Assign(GrSet.ObjectBitmaps[Inf.FirstObjectID]);
          GrSet.GetObjectFrame(Inf.FirstObjectID, 0, ImgObject.Bitmap);
        SetCheckBox(CBObjectOnlyOnTerrain, Inf.FirstObjectOnlyTerrain, Inf.SameObjectOnlyTerrain);
        SetCheckBox(CbObjectUpside, Inf.FirstObjectInvert, Inf.SameObjectInvert);
        SetCheckBox(CbObjectNoOverwrite, Inf.FirstObjectNoOverwrite, Inf.SameObjectNoOverwrite);
      end
      else begin
        EditOID.Text := '';
        ImgObject.Bitmap := nil;
        CBObjectOnlyOnTerrain.Checked := False;
        CbObjectUpside.Checked := False;
        CbObjectNoOverwrite.Checked := False;
      end;

    end;
  finally
    Upd := False;
  end;
end;

procedure TEditItemForm.Form_Deactivate(Sender: TObject);
var
  E: TEditEx;
begin
  if ActiveControl is TEditEx then
  begin
    E := TEditEx(ActiveControl);
    if Assigned(E.OnExit) then
      E.OnExit(E);
  end;
//
end;

procedure TEditItemForm.Edit_Exit(Sender: TObject);
var
  E: TEditEx absolute Sender;
begin
  Assert(Sender is TEditEx, 'sender error (no TEditEx)');
  if E.Modified then
    try
      UpdateEditVar(E);
    except
      E.SetFocus;
      raise;
    end;
end;

procedure TEditItemForm.DoPickList(aType: TLevelItemType);
var
  i: Integer;
begin
  if Editor = nil then
    Exit;
  if not (aType in CurrentInfo.SelTypes) then
    Exit;

  case aType of
    litTerrain:
      begin
        i := CurrentInfo.FirstTerrainID;
        if PickListTerrain(CurrentInfo.GrSet, i) then
          UpdateIDS(aType, i);
      end;

    litObject:
      begin
        i := CurrentInfo.FirstObjectID;
        if PicklistObject(CurrentInfo.GrSet, i) then
          UpdateIDS(aType, i);
      end;
  end;
//    fTerrain.TerrainID := i;
end;

procedure TEditItemForm.BtnPickTerrain_Click(Sender: TObject);
begin
  DoPickList(litTerrain);
end;

procedure TEditItemForm.BtnPickObject_Click(Sender: TObject);
begin
  DoPicklist(litObject);
end;

end.

(*    // eerste item
    B := TSelector(List^[0]).RefItem;
    Inf.TheIndex := B.Index;
    Inf.SelCount := Count;
    Inf.FirstX := B.XPosition;
    Inf.FirstY := B.YPosition;
    Inf.FirstW := B.ItemWidth;
    Inf.FirstH := B.ItemHeight;
    Inf.SameX := True;
    Inf.SameY := True;
    Inf.SameW := True;
    Inf.SameH := True;

    Include(Inf.SelTypes, B.ItemType);
    case B.ItemType of
      litTerrain:
        with TTerrain(B) do
        begin
          Inf.FirstTerrainErase := tdfErase in TerrainDrawingFlags;
          Inf.FirstTerrainInvert := tdfInvert in TerrainDrawingFlags;
          Inf.FirstTerrainNoOverwrite := tdfNoOverwrite in TerrainDrawingFlags;
          Inf.SameTerrainErase := True;
          Inf.SameTerrainInvert := True;
          Inf.SameTerrainNoOverwrite := True;
          Inf.FirstTerrainID := TerrainID;
          Inf.SameTerrainID := True;
        end;
      litObject:
        with TInteractiveObject(B) do
        begin
          Inf.FirstObjectOnlyTerrain := odfOnlyShowOnTerrain in ObjectDrawingFlags;
          Inf.FirstObjectInvert := odfInvert in ObjectDrawingFlags;
          Inf.FirstObjectNoOverwrite := odfNoOverwrite in ObjectDrawingFlags;
          Inf.SameObjectOnlyTerrain := True;
          Inf.SameObjectInvert := True;
          Inf.SameObjectNoOverwrite := True;
          Inf.FirstObjectID := ObjectID;
          Inf.SameObjectID := True;
        end;
    end; // case *)


    1) object dat interface read en write ondersteunt.
    2) object dat interface reader en writer ondersteunt.



    TTerrain = class(IRead, IWrite)
    end;



procedure TEditItemForm.UpdateInterface001;
{-------------------------------------------------------------------------------
  Main routine (big and stupid) to set all contents.
  We really need some good classes to do this all. Always the same shit!
-------------------------------------------------------------------------------}


var
  i: Integer;
  B: TBaseLevelItem;
  FirstX, FirstY, FirstW, FirstH, FirstTID, FirstOID: Integer;
  FirstTerrainFlags: TTerrainDrawingFlags;
  FirstObjFlags: TObjectDrawingFlags;
  SameX, SameY, SameW, SameH, SameTID, SameTerrainFlags, SameObjFlags: Boolean;

  X, Y, W, H: Integer;
  MinX, MinY, MinW, MinH: Integer;
  MaxX, MaxY, MaxW, MaxH: Integer;
  FirstType, T: TLevelItemType;
  SelTypes: TLevelItemTypes;
  TDF: TTerrainDrawingFlags;
  OK: Boolean;

begin
  if clicking then exit;
  upd := true;
  try

  if Editor = nil then
  begin
    DisableInterface;
    Exit;
  end;

  ImgTerrain.Bitmap := nil;
  ImgObject.Bitmap := nil;
  EditTID.Text := '';
  EditOID.Text := '';
  EditIx.Text := '';
  EditX.Text := '';
  EditY.Text := '';
  EditW.Text := '';
  EditH.Text := '';

{  TxtMinX.Caption := '';
  TxtMinY.Caption := '';
  TxtMinW.Caption := '';
  TxtMinH.Caption := '';

  TxtMaxX.Caption := '';
  TxtMaxY.Caption := '';
  TxtMaxW.Caption := '';
  TxtMaxH.Caption := '';}

  CbTerrainErase.State := cbUnchecked;

//  TDF := [tdfErase, tdfUpsideDown, tdfNoOverwrite];

  with Editor.SelectorCollection.HackedList do
  begin

    if Count > 1 then
    begin
      StatusBar.Panels[1].Text := i2s(Count) + ' items selected';
    end
    else if Count = 1 then begin
      StatusBar.Panels[1].Text := '1 item selected';
    end
    else if Count = 0 then
      StatusBar.Panels[1].Text := 'No items selected';

//    if Count = 0 then Exit;

    SameX := False;  SameY := False;  SameW := False;  SameH := False; SameTerrainFlags := False;
    SameObjFlags := False;
    SelTypes := [];
    FirstType := litNone;
    MinX := MaxInt;  MinY := MaxInt;  MinW := MaxInt;  MinH := MaxInt;
    MaxX := MinInt;  MaxY := MinInt;  MaxW := MinInt;  MaxH := MinInt;
    FirstTID := -1;  FirstOID := -1;
    FirstTerrainFlags := [];

    for i := 0 to Count - 1 do
    begin
      B := TSelector(List^[i]).RefItem;
      if i = 0 then
      begin
        FirstX := B.XPosition;
        FirstY := B.YPosition;
        FirstW := B.ItemWidth;
        FirstH := B.ItemHeight;
        SameX := True;
        SameY := True;
        SameW := True;
        SameH := True;

        EditIx.Text := i2s(B.Index);
        if Count <> 1 then
        begin
          EditIx.Text := 'unavailable';
        end;
        EditIx.Enabled := Count = 1;

        EditX.Text := i2s(FirstX);
        EditY.Text := i2s(FirstY);
        EditW.Text := i2s(FirstW);
        EditH.Text := i2s(FirstH);

        if B.ItemType = litTerrain then
        begin
          with TTerrain(B) do
          begin
            CbTerrainErase.Checked := tdfErase in TerrainDrawingFlags;
            CbTerrainUpside.Checked := tdfInvert in TerrainDrawingFlags;
            CbTerrainNoOverwrite.Checked := tdfNoOverwrite in TerrainDrawingFlags;
            FirstTerrainFlags := TerrainDrawingFlags;
            SameTerrainFlags := True;
          end;
        end;

        if B.ItemType = litObject then
        begin
          with TInteractiveObject(B) do
          begin
            CBObjectOnlyOnTerrain.Checked := odfOnlyShowOnTerrain in ObjectDrawingFlags;
            CbObjectUpside.Checked := odfInvert in ObjectDrawingFlags;
            CbObjectNoOverwrite.Checked := odfNoOverwrite in ObjectDrawingFlags;
            FirstObjFlags := ObjectDrawingFlags;
            SameObjFlags := True;
          end;
        end;

        //Continue;
      end; // first item


      X := B.XPosition;
      Y := B.YPosition;
      W := B.ItemWidth;
      H := B.ItemHeight;
      T := B.ItemType;
      Include(SelTypes, T);
//      SameTypes := T = FirstType;

      if T = litTerrain then
        if FirstTID = -1 then
        begin
          EditTID.Text := i2s(B.Identifier);
          FirstTID := B.Identifier;
          try
            ImgTerrain.Bitmap := PictureMgr.TerrainBitmaps[Editor.Statics.GraphicSet, TTerrain(B).TerrainID]
          except
          end;
        end;

      if T = litTerrain then
      begin
        SameTerrainFlags := SameTerrainFlags and (TTerrain(B).TerrainDrawingFlags = FirstTerrainFlags);
        if not SameTerrainFlags then
        begin
          CbTerrainErase.State := cbGrayed;
          CbTerrainUpside.State := cbGrayed;
          CbTerrainNoOverwrite.State := cbGrayed;
        end;
      end;

      if T = litObject then
        if FirstOID = -1 then
        begin
          EditOID.Text := i2s(B.Identifier);
          FirstOID := B.Identifier;
          try
            ImgObject.Bitmap := PictureMgr.ObjectBitmaps[Editor.Statics.GraphicSet, TInteractiveObject(B).ObjectID]
          except
          end;
        end;

      if T = litTerrain then
      begin
        SameObjFlags := SameObjFlags and (TInteractiveObject(B).ObjectDrawingFlags = FirstObjFlags);
        if not SameObjFlags then
        begin
          CBObjectOnlyOnTerrain.State := cbGrayed;
          CbObjectUpside.State := cbGrayed;
          CbObjectNoOverwrite.State := cbGrayed;
        end;
      end;

      MinX := Min(MinX, X);
      MinY := Min(MinY, Y);
      MinW := Min(MinW, W);
      MinH := Min(MinH, H);

      MaxX := Max(MaxX, X);
      MaxY := Max(MaxY, Y);
      MaxW := Max(MaxW, W);
      MaxH := Max(MaxH, H);


      // check x
      if SameX then
      begin
        SameX := SameX and (X = FirstX);
        if not SameX then
//          EditX.Text := '';
      end;
      // check y
      if SameY then
      begin
        SameY := SameY and (Y = FirstY);
        if not SameY then
  //        EditY.Text := '';
      end;
      // check w
      if SameW then
      begin
        SameW := SameW and (W = FirstW);
        if not SameW then
    //      EditW.Text := '';
      end;
      // check h
      if SameH then
      begin
        SameH := SameH and (H = FirstH);
        if not SameH then
      //    EditH.Text := '';
      end;

    end; // for i


//    if SameX then
  //    EditX.Font.Color := clWindowText else EditX.Font.Color := clBlue;

    OK := FirstTID <> -1;
    EditTID.Enabled := OK;
    EditTID.Color := IIF(OK, clWindow, clBtnFace);
{    if OK then
      EditTID.BorderStyle := bsSingle
    else
      EditTID.BorderStyle := bsNone; }
    LabelTerrain.Enabled := OK;
    CbTerrainErase.Enabled := OK;
    CbTerrainUpside.Enabled := OK;
    CbTerrainNoOverwrite.Enabled := OK;

    if OK then
    begin
//      if not SameTerrainFlags then CbTerrainErase.Font.Color := clBlue else
  //      CbTerrainErase.Font.Color := clWindowText;
    end;



    OK := FirstOID <> -1;
    EditOID.Enabled := OK;
    EditOID.Color := IIF(OK, clWindow, clBtnFace);
    {if OK then
      EditOID.BorderStyle := bsSingle
    else
      EditOID.BorderStyle := bsNone; }
    LabelObject.Enabled := OK;
    CBObjectOnlyOnTerrain.Enabled := OK;
    CbObjectUpside.Enabled := OK;
    CbObjectNoOverwrite.Enabled := OK;


    if Count > 0 then
    begin

{      TxtMinX.Caption := i2s(MinX);
      TxtMinY.Caption := i2s(MinY);
      TxtMinW.Caption := i2s(MinW);
      TxtMinH.Caption := i2s(MinH);

      TxtMaxX.Caption := i2s(MaxX);
      TxtMaxY.Caption := i2s(MaxY);
      TxtMaxW.Caption := i2s(MaxW);
      TxtMaxH.Caption := i2s(MaxH);}
    end;

  end;

  finally
    upd:=false;
  end;

end;


