{$include lem_directives.inc}
{$resource lemmixresources.res}

unit FGame;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, Math,
  Dialogs, MMSystem, StdCtrls, TypInfo,
  UMisc,
  GR32, GR32_Image, GR32_Layers,
  LemMisc, LemControls, LemGame, LemTypes, Lemstyles, LemCore,
  UBasePlayingForm;

type
  TGameScroll = (gsNone, gsRight, gsLeft );

  TFormGame = class(TBasePlayingForm)
  { designtime generated }
    Img        : TImage32;
    Label1     : TLabel;
    Label2: TLabel;
    procedure Form_Create(Sender: TObject);
    procedure Form_Destroy(Sender: TObject);
    procedure Form_Activate(Sender: TObject);
    procedure Form_Deactivate(Sender: TObject);
    procedure Form_Close(Sender: TObject; var Action: TCloseAction);
    procedure Form_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Form_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure Form_KeyPress(Sender: TObject; var Key: Char);
    procedure Form_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Form_KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Img_MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure Img_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure Img_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure Img_MouseLeave(Sender: TObject);
  private
  { runtime }
    fGame                 : TLemmingGame;
    SkillToolBar          : TSkillPanelToolbar;
    MinScroll             : Integer;
    MaxScroll             : Integer;
    GameScroll            : TGameScroll;
    ForceUpdateOneFrame   : Boolean;
    MouseClipRect         : TRect;
    PrevCallTime          : Cardinal;
    PrevScrollTime        : Cardinal;
    IdealFrameTimeMS      : Cardinal;
    GameStartTime         : Cardinal;
//    FramesPerSecond       : Double;
    FastForward           : Boolean;
    CanPlay               : Boolean; // true when the game is initialized
    HCursor1, HCursor2    : HCURSOR;
    UseClicksWhenPaused   : Boolean; // true debug mode assignjobs with mouse when paused
    SaveStates            : array[1..12] of Integer;
  { init }
    procedure InitializeCursor;
  { game methods }
    procedure Application_Idle(Sender: TObject; var Done: Boolean);
    procedure Application_Exception(Sender: TObject; E: Exception);
    procedure CheckAdjustReleaseRate;
    procedure CheckScroll;
    procedure CheckResetCursor;
  { game eventhandlers }
    procedure Game_MinimapClick(Sender: TObject; const P: TPoint);
    procedure Game_DebugLemming(L: TLemming);
//    procedure Game_Finished(Sender: TObject);
  { debug or test methods }
    procedure UpdateFrameInfo;
    procedure TestClip;
    procedure ToggleDebugLemming;
    procedure ToggleFrameInfo;
    procedure StartReplay;
    procedure StartReplay2(const aFileName: string);
    procedure GotoSaveState(aTargetIteration: Integer);
    procedure LoadReplay;
    procedure SaveReplay;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    Fakebmp: tbitmap32;
  { method for init and show }
    procedure LoadAndPlay(aLevel: TLevel);
    property Game: TLemmingGame read fGame;
  end;

implementation

{$R *.dfm}

uses
  LemDosStyles;

const
  PLAYCURSOR_DEFAULT = 1;
  PLAYCURSOR_LEMMING = 2;

var
  LemCursorIconInfo : TIconInfo;
  LemSelCursorIconInfo: TIconInfo;

{ TFormGame }

procedure TFormGame.Form_Create(Sender: TObject);
begin
  Color := clBlack;
  fGame := TLemmingGame.Create(nil);
  Game.OnMinimapClick := Game_MinimapClick;
//  Game.OnPutMouse := Game_PutMouse;
//  Game.OnFinish := Game_Finished;
  SkillToolBar := TSkillPanelToolbar.Create(nil);
  SkillToolBar.Parent := Self;
  SkillToolbar.Game := Game;
  IdealFrameTimeMS := 60; // default.
  Application.OnIdle := Application_Idle;
  Application.OnException := Application_Exception;
end;

procedure TFormGame.Form_Destroy(Sender: TObject);
begin
  // interfaces... you cannot trust them. well we're going to change this
  // anyway in the near future
  fGame.InfoPainter := nil;
  SkillToolbar.Game := nil;
  fGame.Free;
  SkillToolbar.Free;

  Application.OnIdle := nil;
  if HCursor1 <> 0 then
    DestroyIcon(HCursor1);
  if HCursor2 <> 0 then
    DestroyIcon(HCursor2);
end;

procedure TFormGame.Application_Idle(Sender: TObject; var Done: Boolean);
var
  CurrTime: Cardinal;
  ForceOne: Boolean;
  TimeForFrame, TimeForScroll, Hyper: Boolean;
begin
  if not CanPlay or not Game.Playing then
    Exit;

  if not Game.HyperSpeed then
    Sleep(1); // relax CPU
  Done := False;
  ForceOne := ForceUpdateOneFrame;
  ForceUpdateOneFrame := False;
  CurrTime := TimeGetTime;
  if Game.CurrentIteration = 0 then
    GameStartTime := CurrTime;

  TimeForFrame := CurrTime - PrevCallTime > IdealFrameTimeMS;
  TimeForScroll := CurrTime - PrevScrollTime > IdealFrameTimeMS;
  Hyper := Game.HyperSpeed;


  if ForceOne or Hyper or FastForward or TimeForFrame or TimeForScroll then
  begin
    PrevCallTime := CurrTime;
    { TODO : This cannot be 100% ok }
    CheckAdjustReleaseRate;

//    if Game.CurrentIteration >= 494 then
  //    Game.Virtuoso;

    if TimeForScroll then
    begin
      PrevScrollTime := CurrTime;
      CheckScroll; // do this before UpdateLemmings, because of screenposition in minimap
    end;
    UpdateFrameInfo; // only in debug mode please

    if ForceOne or not Game.Paused then
      Game.UpdateLemmings;

    if not Hyper then
    begin
      SkillToolBar.RedrawAndUpdateInfo;
      CheckResetCursor;
    end;

    if Hyper then
    begin
      if Game.CurrentIteration >= Game.TargetIteration then
      begin
        //windows.beep(448, 3);
        Game.HyperSpeedEnd;
        SkillToolBar.RedrawAndUpdateInfo;
        CheckResetCursor;
//        showmessage(i2s(game.currenti))
      end;
    end;

  end;

end;

procedure TFormGame.CheckAdjustReleaseRate;
begin
  with Game do
    if SpeedingUpReleaseRate then
    begin
      if not (Replaying and Paused) then
        AdjustReleaseRate(1)
    end
    else if SlowingDownReleaseRate then
    begin
      if not (Replaying and Paused) then
        AdjustReleaseRate(-1)
    end;
end;

procedure TFormGame.CheckScroll;
begin
  case GameScroll of
    gsRight:
      if Img.OffSetHorz > MinScroll * DisplayScale then
      begin
        Img.OffSetHorz := Img.OffSetHorz - DisplayScale * 8;
      end;
    gsLeft:
      if Img.OffSetHorz < MaxScroll * DisplayScale then
        Img.OffSetHorz := Img.OffSetHorz + DisplayScale * 8;
  end;


  //Game.CurrentScreenPosition := Point(Trunc(Img.OffsetHorz), 0);
end;

procedure TFormGame.CheckResetCursor;
begin
  if Screen.Cursor <> Game.CurrentCursor then
  begin
    Img.Cursor := Game.CurrentCursor;
    Screen.Cursor := Game.CurrentCursor;
  end;
end;

procedure TFormGame.Form_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  GameScroll := gsNone;
end;

procedure TFormGame.InitializeCursor;
var
  bmpMask : TBitmap;
  bmpColor : TBitmap;

    procedure scalebmp(bmp:tbitmap; ascale:integer);
    var                         //bad code but it works for now
      b: tbitmap32;
      src,dst:trect;

    begin
      if ascale=1 then exit;
      b:=tbitmap32.create;
      src:=rect(0,0,bmp.width,bmp.height);
      dst:=rect(0,0,bmp.width * ascale, bmp.height*ascale);
      b.setsize(bmp.width*ascale, bmp.height*ascale);
      b.Draw(dst,src, bmp.canvas.handle);
      bmp.Width := b.width;
      bmp.height:=b.height;
      b.drawto(bmp.canvas.handle, 0, 0);// gr32
      b.free;
    end;


begin
  bmpMask := TBitmap.Create;
  bmpColor := TBitmap.Create;

  //bmpMask.LoadFromFile(apppath + 'dosgamecursormask1.bmp');
  //bmpColor.LoadFromFile(apppath+'dosgamecursor1.bmp');
  bmpMask.LoadFromResourceName(HINSTANCE, 'GAMECURSOR_DEFAULT_MASK');
  bmpColor.LoadFromResourceName(HINSTANCE, 'GAMECURSOR_DEFAULT');

  ScaleBmp(bmpMask, DisplayScale);
  ScaleBmp(bmpColor, DisplayScale);

  with LemCursorIconInfo do
  begin
    fIcon := false;
    xHotspot := bmpmask.width div 2;
    yHotspot := bmpmask.Height div 2;
    hbmMask := bmpMask.Handle;
    hbmColor := bmpColor.Handle;
  end;

  HCursor1 := CreateIconIndirect(LemCursorIconInfo);
  Screen.Cursors[PLAYCURSOR_DEFAULT] := HCursor1;//CreateIconIndirect(LemCursorIconInfo);


  img.Cursor := PLAYCURSOR_DEFAULT;
  SkillToolBar.img.cursor := PLAYCURSOR_DEFAULT;
//  Screen.Cursor := MyCursor;

  bmpMask.Free;
  bmpColor.Free;

  //////////

  bmpMask := TBitmap.Create;
  bmpColor := TBitmap.Create;

//  bmpMask.LoadFromFile(apppath + 'dosgamecursormask2.bmp');
//  bmpColor.LoadFromFile(apppath+'dosgamecursor2.bmp');
  bmpMask.LoadFromResourceName(HINSTANCE, 'GAMECURSOR_HIGHLIGHT_MASK');
  bmpColor.LoadFromResourceName(HINSTANCE, 'GAMECURSOR_HIGHLIGHT');

  scalebmp(bmpmask, DisplayScale);
  scalebmp(bmpcolor, DisplayScale);


  with LemSelCursorIconInfo do
  begin
    fIcon := false;
    xHotspot := bmpmask.width div 2;
    yHotspot := bmpmask.Height div 2;
//    xHotspot := 7 * 3;////5;
//    yHotspot := 7 * 3;//15;
    hbmMask := bmpMask.Handle;
    hbmColor := bmpColor.Handle;
  end;

  HCursor2 := CreateIconIndirect(LemSelCursorIconInfo);
  Screen.Cursors[PLAYCURSOR_LEMMING] := HCursor2;

//  Screen.Cursor := MyCursor;

  bmpMask.Free;
  bmpColor.Free;
end;

procedure TFormGame.Form_KeyPress(Sender: TObject; var Key: Char);
// some keypress debugging features
//var
  //OldCanPlay: Boolean;
begin
  case Key of

    // --------- LOWERCASE ------------

    // go 10 game seconds further
    ' ':
            if Game.Playing then
            begin
              if not Game.HyperSpeed then
              begin
                //windows.beep(448, 3);
                Game.HyperSpeedBegin;
                Game.TargetIteration := Game.CurrentIteration + 17*10;
              end
            end;


    // test method ClipCursor
    'c': TestClip;

    // toggle fastforward
    'f': FastForward := not FastForward;

    // hide/show lemming debug line
    'h': if LemmixDebugMode then ToggleDebugLemming;

    // hide/show frame info debug line
    'i': if LemmixDebugMode then ToggleFrameInfo;

    // hide/show lemming foot-pixel
    'l': if LemmixDebugMode then Game.DrawLemmingPixel := not Game.DrawLemmingPixel;

    // do next frame if paused
    'n': {if LemmixDebugMode then}
      if Game.Paused then
        if Game.Replaying then
          ForceUpdateOneFrame := True;

    // disable/enable skill assignment when paused
    'p': if LemmixDebugMode then UseClicksWhenPaused := not UseClicksWhenPaused;

    // start replay
    'r': StartReplay;

    // enable/disable sounds
    's': Game.DoSounds := not Game.DoSounds;

    // save current game to .\Replay\<leveltitle>.lrb
    // in addition a more or less readable version is saved too with the extension ".txt"
    'u': SaveReplay;

    // --------- UPPERCASE ------------

    // disable/enable changing the selected skillbutton with mouse when paused
    'C': if LemmixDebugMode then
           Game.SkillButtonsDisabledWhenPaused := not Game.SkillButtonsDisabledWhenPaused;


    'F':
      begin
            if Game.Playing then
            begin
              if not Game.HyperSpeed then
              begin
                //windows.beep(448, 3);
                Game.HyperSpeedBegin;
                Game.TargetIteration := Game.CurrentIteration + 17*60;
              end
              else begin
                Game.HyperSpeedEnd;
              end;
            end;

      end;


    // load replay file
    'R': LoadReplay;

  end;
end;

procedure TFormGame.ToggleDebugLemming;
begin
  if Assigned(Game.OnDebugLemming) then
  begin
    Label2.Visible := False;
    Game.OnDebugLemming := nil;
    Label2.Caption := '';
  end
  else begin
    Label2.Visible := True;
    Game.OnDebugLemming := Game_DebugLemming;
  end;
end;

procedure TFormGame.ToggleFrameInfo;
begin
  if Label1.Visible then
  begin
    Label1.Visible := False;
    Label1.Caption := '';
  end
  else begin
    Label1.Visible := True;
  end;
end;

procedure TFormGame.Form_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
//var
  //i: Integer;
begin
  case Key of
    VK_ESCAPE:
      Close;
  end;

  if not Game.Playing then
    Exit;

  with Game do
  begin

    if Shift = [ssCtrl] then
    begin
      case Key of
        VK_F1:
          begin
            SaveStates[1] := Game.CurrentIteration;
          end;
      end;
    end
    else if Shift = [ssShift] then
    begin
      case Key of
        VK_F1:
          begin
            if SaveStates[1] <> 0 then
              GotoSaveState(SaveStates[1]);
          end;
      end;
    end

    
    else if Shift = [] then
    begin

      if (Key >= VK_F1) and (Key <= VK_F12) then
      begin
        if Key <> VK_F11 then
  //        if not Game.Paused then
            Game.RegainControl;
        case Key of
          VK_F1: SetSelectedSkill(bskSlower, True);
          VK_F2: SetSelectedSkill(bskFaster, True);
          VK_F3: SetSelectedSkill(bskClimber, True);
          VK_F4: SetSelectedSkill(bskUmbrella, True);
          VK_F5: SetSelectedSkill(bskExplode, True);
          VK_F6: SetSelectedSkill(bskStopper, True);
          VK_F7: SetSelectedSkill(bskBuilder, True);
          VK_F8: SetSelectedSkill(bskBasher, True);
          VK_F9: SetSelectedSkill(bskMiner, True);
          VK_F10: SetSelectedSkill(bskDigger, True);
          VK_F11: SetSelectedSkill(bskPause);
          VK_F12: SetSelectedSkill(bskNuke);
        end;
      end
      // other keys
      else begin
        case Key of
          VK_LEFT: GameScroll := gsLeft;
          VK_RIGHT: GameScroll := gsRight;
        end;
      end;
    end;

  end;

end;

procedure TFormGame.Form_KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin


  if not game.Playing then
    exit;

  with Game do
  begin

    case Key of
      VK_F1: SetSelectedSkill(bskSlower, False);
      VK_F2: SetSelectedSkill(bskFaster, False);
      VK_LEFT: GameScroll := gsNone;
      VK_RIGHT: GameScroll := gsNone;
    end;
  end;
end;

procedure TFormGame.Form_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  with MouseClipRect do
    if (Y >= Img.Top) and (Y <= Img.Top + Img.Height - 1) then
    begin
      if X <= Img.Left + DisplayScale then
        GameScroll := gsLeft
      else if X >= Img.Left + Img.Width - 1 + DisplayScale then
        GameScroll := gsRight
      else
        GameScroll := gsNone;
    end
    else
      GameScroll := gsNone;
end;

procedure TFormGame.Img_MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
var
  HandleClick: Boolean;
begin
  if Game.Playing then
  begin
    Game.RegainControl;
    Game.CursorPoint := Img.ControlToBitmap(Point(X, Y));

    //umisc
    // debug creation of lemming
    if LemmixDebugMode then
    if ssCtrl in Shift then
    if ssAlt in Shift then
    begin
      Game.CreateLemmingAtCursorPoint;//lemgame
      Exit;
    end;
    

    // normal
    Game.RightMouseButtonHeldDown := ssRight in Shift;
    if Button = mbLeft then
    begin
      HandleClick := not Game.Paused or UseClicksWhenPaused;
      if HandleClick then
        Game.ProcessSkillAssignment;
    end;
  end;
end;

procedure TFormGame.Img_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
begin
  if Game.Playing then
  begin
    Game.RightMouseButtonHeldDown := ssRight in Shift;
    Game.CursorPoint := Img.ControlToBitmap(Point(X, Y));
    if Game.Paused then
      Game.HitTest; // maybe move to idle
    if X >= Img.Width - 1 then
      GameScroll := gsRight
    else if X <= 0 then
      GameScroll := gsLeft
    else
      GameScroll := gsNone;
  end;
end;

procedure TFormGame.Img_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
begin
  Game.RightMouseButtonHeldDown := ssRight in Shift;
end;

procedure TFormGame.Img_MouseLeave(Sender: TObject);
begin
  GameScroll := gsNone;
end;

procedure TFormGame.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params.WindowClass do
    Style := Style or CS_OWNDC; // maybe faster screen output
end;

procedure TFormGame.Form_Close(Sender: TObject; var Action: TCloseAction);
begin
  CanPlay := False;
  Game.SetGameResult;
//  if not Game.Replaying then
//    SaveReplay;
  ClipCursor(nil);
end;

procedure TFormGame.LoadAndPlay(aLevel: TLevel);
{-------------------------------------------------------------------------------
  Initialize form and game, after that showmodal
-------------------------------------------------------------------------------}
var
  Par: TDefaultsRec;
begin
//  fakebmp:=tbitmap32.create;
  Par := aLevel.Style.GetParams;
  InitializeScaleAndRect(Par.ViewWidth, Par.ViewHeight);
  Game.Level := aLevel;

  // check superlemming
  if aLevel.Statics.SuperLemming then
    IdealFrameTimeMS := 20
  else
    IdealFrameTimeMS := 60;

  // reload animations
  with aLevel.Style.AnimationSet do
  begin
    ClearData;
    BrickColor := aLevel.Graph.BrickColor;
    Assert(alevel.Graph is TBaseDosGraphicSet, 'class error graph');
    { TODO : get rid of typecasting }
    TBaseDosAnimationSet(aLevel.Style.AnimationSet).AnimationPalette :=
      TBaseDosGraphicSet(aLevel.Graph).PaletteStandard;
    ReadData;
  end;

  // position playing field and skillpanel
  Img.SetBounds(MainRect.Left,
                MainRect.Top,
                RectWidth(MainRect),
                Par.LevelHeight * DisplayScale);

  Img.Scale := DisplayScale;
  Img.ScaleMode := smScale;
  Img.OffsetHorz := -Game.Level.Statics.ScreenPosition * DisplayScale;

  { TODO : get rid of typecasting }
  SkillToolBar.SetStyleAndGraph(Game.Level.Style, TBaseDosGraphicSet(game.Level.Graph), DisplayScale);
  SkillToolBar.Top := Img.Top + Img.Height;
  SkillToolBar.left := Img.Left;

  Game.Renderer.LevelBitmap := Img.Bitmap;
  Game.Renderer.RenderOptions := [];
  Game.Renderer.RenderTypes := [litTerrain];
  Game.Renderer.RenderWorld(Game.Level);
  MinScroll := -(Img.Bitmap.Width - Par.ViewWidth);
  MaxScroll := 0;
  InitializeCursor;
  Mouse.CursorPos := Point(Screen.Width div 2, Screen.Height div 2);
  MouseClipRect := MainRect;
  Inc(MouseClipRect.Right); // we need to add one, windows rect thingy
  Inc(MouseClipRect.Bottom); // we need to add one, windows rect thingy
//  game.renderer.levelbitmap:=fakebmp;
  Game.Start;
  SkillToolbar.RedrawAndUpdateInfo;
  ShowModal;
end;

procedure TFormGame.StartReplay;
begin
  CanPlay := False;
  Game.SetGameResult;
  //if not Game.Replaying then
    //SaveReplay;
  Game.Renderer.RenderWorld(Game.Level);
  Game.Start(True);
  SkillToolbar.RedrawAndUpdateInfo;
  CanPlay := True;
end;

procedure TFormGame.StartReplay2(const aFileName: string);
begin
  CanPlay := False;
  if CompareText(ExtractFileExt(aFileName), '.TXT') = 0
  then Game.Recorder.LoadFromOldTxt(aFileName)
  else

  Game.Recorder.LoadFromFile(aFileName);
  //SaveReplay;       asdf
  Game.Renderer.RenderWorld(Game.Level);

//  FillChar(SaveStates, SizeOf(SaveStates), 0);
  Game.Start(True);
  SkillToolbar.RedrawAndUpdateInfo;
  CanPlay := True;
end;

procedure TFormGame.GotoSaveState(aTargetIteration: Integer);
begin
  CanPlay := False;
  //if Game.HyperSpeed then
    //Game.HyperSpeedEnd;
  Game.Renderer.RenderWorld(Game.Level);
//  FillChar(SaveStates, SizeOf(SaveStates), 0);
  Game.Start(True{Game.Replaying});
  Game.HyperSpeedBegin;
  Game.TargetIteration := aTargetIteration;
  SkillToolbar.RedrawAndUpdateInfo;
  CanPlay := True;
end;

procedure TFormGame.LoadReplay;
var
  OldCanPlay: Boolean;
  Dlg : TOpenDialog;
  s: string;
begin
  OldCanPlay := CanPlay;
  CanPlay := False;
  s:='';
  dlg:=topendialog.create(nil);
  try
//    dlg.DefaultExt := '*.lrb';
    dlg.filename := '*.lrb';
    if dlg.execute then
    begin
      s:=dlg.filename;
    end;
  finally
    dlg.free;
  end;
  if s <> '' then
  begin
    StartReplay2(s);
    exit;
  end;
  CanPlay := OldCanPlay;
end;

procedure TFormGame.Game_MinimapClick(Sender: TObject; const P: TPoint);
{-------------------------------------------------------------------------------
  This is kind of complicated:
  The Dos SkillPanelToolbar calls Game.MiniMapClick when the mouse is down in
  the minimap.
  This method is an eventhandler: Game.OnMiniMapClick and here we scroll the game-image.
-------------------------------------------------------------------------------}
var
//  N: TPoint;
  O: Single;
begin
//  Game.CurrentScreenPosition := Point(P.X, 0);
  O := -P.X * DisplayScale;
  O :=  O + Img.Width div 2;
  if O < MinScroll * DisplayScale then O := MinScroll * DisplayScale;
  if O > MaxScroll * DisplayScale then O := MaxScroll * DisplayScale;
  Img.OffSetHorz := O;
end;

procedure TFormGame.Form_Activate(Sender: TObject);
begin
  OnActivate := nil;
  CanPlay := True;
  ClipCursor(@MouseClipRect);
end;

procedure TFormGame.Form_Deactivate(Sender: TObject);
begin
  ClipCursor(nil);
end;

procedure TFormGame.TestClip;
var
  MouseClipped: Boolean;
  OK, Same: Boolean; Clip: TRect;
  Success: Boolean;
begin
  MouseClipped := ClipCursor(@MouseClipRect);

  Success := MouseClipped;
  if not MouseClipped then
    ShowMessage('ClipCursor failed');

  Clip := EmptyRect;
  OK := GetClipCursor(Clip);
  Success := Success and OK;
  if not OK then
    ShowMessage('Call to GetClipCursor failed')
  else begin

    Same := CompareMem(@Clip, @MouseClipRect, SizeOf(TRect));
    Success := Success and Same;
    if not Same then
      showmessage('GetClipCursor:'+rectstr(clip) + chr(13) +
      'should be:'+rectstr(mousecliprect));
  end;


  if Success then
    canvas.Pixels[0,0] := clwhite else canvas.pixels[0,0] := clred;
  //  ShowMessage('ClipCursor successful')

end;

procedure TFormGame.Game_DebugLemming(L: TLemming);
begin
  Label2.Caption := L.LemHint;
end;

procedure TFormGame.SaveReplay;
var
  i: Integer;
//  L: TStringList;
  path,n: string;
  fn: string;
begin
//  if game.paused then
  //  game.setselectedskill(bskPaused);
  game.SetGameResult;
//  if not UseSaveReplay then
  //  Exit;
//  if game.ReplayList.count = 0 then
  //  e

  path := apppath + 'Replay\';
  if not forcedirectories(path) then
    raise exception.create('cannot create replay directory');


  N := trim(game.level.LevelName);
  if n = '' then
    N := '_noname'
  else
    for i := 1 to Length(n) do
      if n[i] in ['\','/', ':', '*', '?', '"', '<', '>'] then
        n[i] := '_';

//  if game.GameResultRec.gSuccess then
  //  n := 'ok_' + n;

  fn := path + n + '.lrb';

  (*
  i := 1;
  while fileexists(fn + '_'+leadzerostr(i, 3) + '.bin') do
  begin
    Inc(i);
    if i > 9 then
    begin
      i := 1;
      break;
    end;
  end;
  fn := fn + '_'+leadzerostr(i, 3) + '.bin';
  *)


//  Game.ReplayList.SaveToFile('replay.bin');
//  Game.ReplayList.SaveToTxt('replay.txt');
  Game.Recorder.SaveToFile(fn);
  Game.Recorder.SaveToTxt(replacefileext(fn, '.txt'));

  (*
  L := TStringList.Create;

  with Game.ReplayList do
    for i := 0 to Count - 1 do
      with Items[i] do
      begin
        S := i2s(Iteration) + ', ';

        if RecordedAction <> raNone then
        begin
          S := S + GetEnumName(TypeInfo(TRecordedAction), Integer(RecordedAction)) + ', ';
          case RecordedAction of
            raSkillAssignment:
              S := S + GetEnumName(TypeInfo(TBasicLemmingAction), Integer(Skill)) + ', ' +
                   i2s(LemmingIndex);
          end;
        end
        else if ReleaseRateAction <> rrNone then
        begin
          S := S + GetEnumName(TypeInfo(TReleaseRateAction), Integer(ReleaseRateAction)) + ', ';
          S := S + i2s(ReleaseRate);
        end
        else S := '?????';

        L.Add(S);
      end;

  L.SaveToFile(AppPath + 'replay.txt');
  L.Free;

  *)

end;


procedure TFormGame.UpdateFrameInfo;
begin
  if not Label1.Visible then
    Exit;
  Label1.Caption := 'iteration='+LeadZeroStr(Game.CurrentIteration, 6) + ', ' +
                    'clockframe='+LeadZeroStr(Game.ClockFrame, 3)+ ', ' +
                    'interval='+LeadZeroStr(IdealFrameTimeMS, 4);
end;


procedure TFormGame.Application_Exception(Sender: TObject; E: Exception);
begin
  if Game <> nil then
  begin
    CanPlay := False;
    Application.OnIdle := nil;
    if Game.Playing then
      E.Message := 'Error during play:' + chr(13) + chr(13) + E.Message;
    Application.ShowException(E);
  end;
end;

end.
(*
procedure TFormGame.Game_PutMouse(L: TLemming);
var
  P: TPoint;
begin
  P := Img.BitmapToControl(Point(L.LemX, L.LemY));
  P := Img.ClientToScreen(P);
  Mouse.CursorPos := P;
end;

procedure TFormGame.Game_Finished(Sender: TObject);
begin
  Close;
(*  with Img.Bitmap do
  begin
     DrawMode := dmBlend;
     MasterAlpha := 140;

  end; *)

//  asdf
end;

