{$include lem_directives.inc}

unit FLevelMatrix;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, UMx, UMxStrings, UMisc,
  LemApp,
  FBasicLevelEditor,
  LemMisc, LemCore, LemEdit, GR32, GR32_Image, StdCtrls;

type
  TLevelListEditorForm = class(TBasicLevelEditorForm)
    Matrix: TStringMatrix;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Matrix_GetActiveStyle(Sender: TObject; aCol, aRow: Integer; aMatrixStyle, aColStyle, aRowStyle: TCellStyle; var ActiveStyle: TCellStyle);
    procedure Matrix_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Matrix_KeyPress(Sender: TObject; var Key: Char);
    procedure Matrix_PaintCell(Sender: TObject; Params: TCellParams; Data: TCellData);
  private
//    fLink: TLevelEditorLink;
    procedure LMLevelChanged(var Msg: TLMLevelChanged); message LM_LEVELCHANGED;
    procedure LMSelectionChanged(var Msg: TLMSelectionChanged); message LM_SELECTIONCHANGED;
    procedure LMObjectListChanged(var Msg: TLMObjectListChanged); message LM_OBJECTLISTCHANGED;
    procedure LMTerrainListChanged(var Msg: TLMTerrainListChanged); message LM_TERRAINLISTCHANGED;
    procedure LMSteelListChanged(var Msg: TLMSteelListChanged); message LM_STEELLISTCHANGED;
    procedure UpdateMatrix(aItem: TBaseLevelItem = nil);
    procedure UpdateMatrixSelection;
//    function GetLevelEditor: TLevelEditor;
  protected
    procedure DoAfterSetEditor(Value: TLevelEditor); override;

//    procedure SetLevelEditor(const Value: TLevelEditor);
  public
  //  property LevelEditor: TLevelEditor read GetLevelEditor write SetLevelEditor;
  end;

var
  LevelListEditorForm: TLevelListEditorForm;

implementation

{$R *.dfm}

const
  col_ix = 0;
  col_id = 1;
  col_xpos = 2;
  col_ypos = 3;
  col_width = 4;
  col_height = 5;
  col_name = 6;

  obj_column = 0;

  TAG_SELECTED = Bit0;

{ TLevelListEditorForm }

procedure TLevelListEditorForm.UpdateMatrix(aItem: TBaseLevelItem = nil);
{-------------------------------------------------------------------------------
  Rebuild level matrix. Care must be taken to delete all invalid pointers!
-------------------------------------------------------------------------------}
var
  i: Integer;
  T: TTerrain;
  O: TInteractiveObject;
  S: TSteel;
  Lev: TLevelEditor;
  R: Integer;
  GN, GO: string;
  Gr: Integer;
  NewCount: Integer;
begin
  Lev := {Level}Editor;
  if Lev = nil then
    Exit;
  if aItem <> nil then
  begin
//    Matrix.Invalidate;
  //  Exit;
  end;

  with Matrix do
  begin
    Gr := Lev.Statics.GraphicSet;
    if Lev.Graph <> nil then
      GN := Lev.Graph.GraphicSetName
    else
      GN := '';
//    GN := Lev.Graph
//    Gr.Gr//''GraphicSetToDescription(Gr);
    PaintLock;
    try
    NewCount := Lev.TerrainCollection.Count +
                Lev.InteractiveObjectCollection.Count +
                Lev.SteelCollection.Count;
{    if NewCount = 0 then
    begin
      NewCount := 1;
      Cells[]
    end; }

    if RowCount <> NewCount then
    begin
      RowCount := NewCount;
      if RowCount > 0 then   //umx
        Row := 0;
    end;
    R := 0;

    for i := 0 to RowCount - 1 do
    begin
      Objects[obj_column, i] := nil;
      Rows[i].Tag := 0;
    end;

    for i := 0 to Lev.InteractiveObjectCollection.Count - 1 do
    begin
      O := Lev.InteractiveObjectCollection[i];
      Cells[col_ix, R] := i2s(i);
      Cells[col_id, R] := i2s(O.ObjectID);
      Cells[col_xpos, R] := i2s(O.ObjectX);
      Cells[col_ypos, R] := i2s(O.ObjectY);
//      GO := GraphicObjectToDescription(Gr, O.ObjectID);
//      if GO = '' then
        GO := 'Object';
      Cells[col_name, R] := GO;// + '(' + GN + ')';

      Objects[obj_column, R] := O;

      Rows[R].Style.Colors.BackColor := ObjColor;//RGB(239, 254, 237);
//      Rows[R].Style.Colors.BackColor := RGB(227, 254, 222);
      Inc(R);
    end;
    for i := 0 to Lev.TerrainCollection.Count - 1 do
    begin
      T := Lev.TerrainCollection[i];
      Cells[col_ix, R] := i2s(i);
      Cells[col_id, R] := i2s(T.TerrainId);
      Cells[col_xpos, R] := i2s(T.TerrainX);
      Cells[col_ypos, R] := i2s(T.TerrainY);
      Cells[col_name, R] := 'Terrain';

      Objects[obj_column, R] := T;
//        252, 239, 226
        Rows[R].Style.Colors.BackColor := TerColor;//RGB(252, 239, 226);
//      Rows[R].Style.Colors.BackColor := RGB(247, 217, 187); DIT IS DONKERDER
      Inc(R);
    end;
    for i := 0 to Lev.SteelCollection.Count - 1 do
    begin
      S := Lev.SteelCollection[i];
      Cells[col_ix, R] := i2s(i);
      Cells[col_id, R] := '';
      Cells[col_xpos, R] := i2s(S.SteelX);
      Cells[col_ypos, R] := i2s(S.SteelY);
      Cells[col_width, R] := i2s(S.SteelWidth);
      Cells[col_height, R] := i2s(S.SteelHeight);
      Cells[col_name, R] := 'Steel';
      Objects[obj_column, R] := S;
      Rows[R].Style.Colors.BackColor := StlColor;//RGB(240, 240, 240);
      Inc(R);
    end;
    finally
    PaintUnlock;
    end;
  end;
end;

procedure TLevelListEditorForm.UpdateMatrixSelection;
var
  O: TObject;
  i: Integer;
  R: TMxRow;
  Lev: TLevelEditor;
begin
  Lev := {Level}Editor;
  if Lev = nil then
    Exit;
  with Matrix do
  begin
    for i := 0 to RowCount - 1 do
    begin
      R := Rows[i];
      R.Tag := 0;
      O := Objects[obj_column, i];
      if O is TBaseLevelItem then
        if Lev.SelectorCollection.FindBaseItem(TBaseLevelItem(O)) >= 0 then
          R.Tag := TAG_SELECTED;
    end;
  end;
  Matrix.Invalidate;
end;

procedure TLevelListEditorForm.FormCreate(Sender: TObject);
begin
//  fLink := TLevelEditorLink.Create(nil, Self);
//  fCellData := TTerrainMatrixCellData.Create;
end;

procedure TLevelListEditorForm.FormDestroy(Sender: TObject);
begin
//  fLink.Free;
  //fCellData.Free;
end;

{function TLevelListEditorForm.GetLevelEditor: TLevelEditor;
begin
  Result := fLink.LevelEditor;
end;

procedure TLevelListEditorForm.SetLevelEditor(const Value: TLevelEditor);
begin
  fLink.LevelEditor := Value;
end;}

procedure TLevelListEditorForm.LMLevelChanged(var Msg: TLMLevelChanged);
begin
  UpdateMatrix;
end;

procedure TLevelListEditorForm.LMSelectionChanged(var Msg: TLMSelectionChanged);
begin
  UpdateMatrixSelection;
end;

procedure TLevelListEditorForm.LMObjectListChanged(var Msg: TLMObjectListChanged);
begin
  UpdateMatrix(Msg.Obj);
end;

procedure TLevelListEditorForm.LMSteelListChanged(var Msg: TLMSteelListChanged);
begin
  UpdateMatrix(Msg.Steel);
end;

procedure TLevelListEditorForm.LMTerrainListChanged(var Msg: TLMTerrainListChanged);
begin
  UpdateMatrix(Msg.Terrain);
end;


procedure TLevelListEditorForm.Matrix_GetActiveStyle(Sender: TObject; aCol,
  aRow: Integer; aMatrixStyle, aColStyle, aRowStyle: TCellStyle;
  var ActiveStyle: TCellStyle);
begin
//
  ActiveStyle := aRowStyle;
end;

procedure TLevelListEditorForm.Matrix_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  O: TObject;
  Lev: TLevelEditor;
begin
exit;
  Lev := {Level}Editor;
  if Lev = nil then
    Exit;
  with Matrix do
  if Shift = [] then
    case Key of
      VK_DELETE:
        begin
          {O := Objects[obj_column, row];
          if O is TBaseLevelItem then
            O.Free; }
        end;
    end;


end;

procedure TLevelListEditorForm.Matrix_KeyPress(Sender: TObject; var Key: Char);
var
  Lev: TLevelEditor;
  O: TObject;
begin
  Lev := {Level}Editor;
  if Lev = nil then
    Exit;
  case Key of
    ' ':
      begin
        with Matrix do
        begin
          O := Objects[obj_column, row];
          if not (ssShift in GetKeyBoardShiftState) then
          begin
            if O is TInteractiveObject then
            begin
              Lev.SelectedObject := TInteractiveObject(O);
            end
            else if O is TTerrain then
            begin
              Lev.SelectedTerrain := TTerrain(O);
            end
            else if O is TSteel then
            begin
              Lev.SelectedSteel := TSteel(O);
            end;
          end
          else begin
            Lev.ExtendSelection(TBaseLevelItem(O), True);
          end;
        end;
      end;
    'v':
      begin
        with Matrix do
          O := Objects[obj_column, row];
        if not (O is TBaseLevelItem) then
          Exit;
        Lev.ExtendSelection(TBaseLevelItem(O), True);
      end;
  end; // case
end;

procedure TLevelListEditorForm.Matrix_PaintCell(Sender: TObject; Params: TCellParams; Data: TCellData);
begin

  with Matrix, Params do
    if not (dsFocused in pDrawState) then
      if not (dsHighlightRow in pDrawState) then
    if Rows[pRow].Tag and TAG_SELECTED <> 0 then
    begin
      pBackColor := clRed;
      pTextColor := clWhite;
    end;
end;

procedure TLevelListEditorForm.DoAfterSetEditor(Value: TLevelEditor);
begin
  UpdateMatrix;
end;

end.

