{$include lem_directives.inc}

unit FMapEditor;

{-------------------------------------------------------------------------------
  One of the possible Lemmix mainforms: the level editor. The other mainform is
  planned to be a lemming-playing form.
  o There are several develop forms. they are *NOT* allowed in exeversions!
  o The only designtime menuitem is the "file" item, just to see
    something designtime. The rest is created runtime.
-------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms, TypInfo,
  ClipBrd, FileCtrl, IniFiles, Dialogs, Math,
  Types, Menus, ActnList, ComCtrls, ExtCtrls, StdCtrls,
  // kernel + kernel 3dparty
  UMasterForms, UMisc, UEvents, UFiles, UTools, UWinTools, USplit,
  GR32, GR32_Image, GR32_Layers, GifImage,
  // develop
  {$ifdef develop}
lemreplay, lemdosarc,lemdoscmp,  lemdevelop,
  {LemConvert, }FTest, FShowSPRFiles, FShowDOSGroundFile,
  {$endif}
  // lemmings units
  LemDosFiles, LemStyles, LemStrings, LemTypes, LemFiles, LemMisc, LemCore, LemEdit,
  FEditItem, FEditStatics, FLevelMatrix, FAbout, FBasicLevelEditor,
  FPickTerrain, FAlignSelection, FNewLevel, FPasteSpecial,
  FDosDatDecompression, FDosDecompress, //FViewDosLevelPack,
  FWizLemminiCompiler, FAskStyle, {FOpenDosDat, }FVgaSpecCreator, FValidationReport,
  FVgaSpecViewer, FCompileStyle,
  // lemmings forms
  FGame, FViewlevels, LemDosStyles, LemLemminiStyles, LemGraphics,
  FEditDosLevelPack,
  FOptions,
  ToolWin, ImgList, USplitPanels;
  {fdraw, }

type
  TMapEditorForm = class;

  TClickAction = (
    caNone,
    caSelect
  );

  (*
  {-------------------------------------------------------------------------------
    Class to keep track of different modes of loading.
    The form keeps an instance of this class so that the form is aware of this
    edit-mode.
    So when loading a level in the editor
  -------------------------------------------------------------------------------}
  TLevelEditController = class
  private
  protected
  public
  published
  end;
  *)

  TMapEditorForm = class(TBasicLevelEditorForm)
    Img: TImgView32;
    StatusBar: TStatusBar;
    Splitter1: TSplitter;
    SplitPanel1: TSplitPanel;
    ActionList: TActionList;
    MainMenu: TMainMenu;
    MbFile: TMenuItem;
    MapPopup: TPopupMenu;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ImageList: TImageList;
    procedure Form_Create(Sender: TObject);
    procedure Form_Destroy(Sender: TObject);
    procedure Form_Show(Sender: TObject);
    procedure Img_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Img_KeyPress(Sender: TObject; var Key: Char);
    procedure Img_MouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
    procedure Img_MouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
    procedure Img_MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure Img_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure Img_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure Img_StartDrag(Sender: TObject; var DragObject: TDragObject);
    procedure Img_DragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    procedure Img_DragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure Img_EndDrag(Sender, Target: TObject; X, Y: Integer);
    procedure FormContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
    procedure ActionList_Update(Action: TBasicAction; var Handled: Boolean);
    procedure Button2Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
  private
    fRenderer                  : TLevelRenderer;
    fLocked                    : Boolean;
    fHitList                   : TList;
    fScaleIndex                : Integer;
    fDragger                   : TDragObject;
    fSelectionLayer            : TPositionedLayer; // the gr32_image layer to paint selectionrectangles
    fScreenStartLayer          : TPositionedLayer; // the gr32_image layer to paint the screenstart of the level
    fDrawingSelectionRect      : Boolean;
    fSelectionRect             : TRect;
    fSelectionRectVisible      : Boolean;
//    fCurrentClickTerrain       : Integer;
    DragSelection              : Boolean;
    DragIP                     : TPoint; // TPoint startdrag als Image coordinaat
    DragBP                     : TPoint; // TPoint startdrag als Bitmap coordinaat
//    fDragBitmap                : TBitmap32;
    LastHitItem                : TBaseLevelItem;
    fImageIsDragging           : Boolean;
    fMouseOnSelection          : Boolean;
    fCommandStr                : string;
    fCommandChangeTicks        : Cardinal;
//    fCopyPoint                 : TPoint;
    fInternalGettingDragBitmap : Boolean;
    fSelectorColors            : array[litObject..litSteel] of TColor32;
    fStaticsForm               : TFormEditStatics;
    fListForm                  : TLevelListEditorForm;
    fEditItemForm              : TEditItemForm;
  { menubar }
    MbEdit                     : TMenuItem; // menubar
    MbSelection                : TMenuItem;
    MbView                     : TMenuItem;
//    MbStyles                   : TMenuItem;
    MbTools                    : TMenuItem;
    MbHelp                     : TMenuItem;
    {$ifdef develop}
    MhRecentFiles              : TMenuItem; // menuholder recentfiles
    MbDevelop                  : TMenuItem;
    {$endif}
  { actions }
  { file ... }
    ActNewLevel                : TAction;
    ActNewDosLevelPack         : TAction;
    ActNewDosExtGraphic        : TAction;
    ActOpenFile                : TAction;
    ActReopenFile              : TAction;
    ActCloseLevel              : TAction;
    ActSaveAs                  : TAction;
    ActOptions                 : TAction;
    ActQuit                    : TAction;
  { edit ... }
    ActInsertObject            : TAction;
    ActInsertObjectLocal       : TAction; // local menu
    ActInsertTerrain           : TAction;
    ActInsertTerrainLocal      : TAction; // local menu
    ActInsertSteel             : TAction;
    ActInsertSteelLocal        : TAction; // local menu
    ActCopy                    : TAction;
    ActCut                     : TAction;
    ActPaste                   : TAction;
    ActPasteSpecial            : TAction;
  { selection ... }
    ActAlignSelection          : TAction;
    ActDosAlignSelection       : TAction;
  { view ... }
    ActShowInspector           : TAction;
    ActShowStatics             : TAction;
    ActShowList                : TAction;
    ActShowImage               : TAction;
    ActToggleTerrain           : TAction;
    ActToggleObjects           : TAction;
    ActToggleSteel             : TAction;
    ActToggleSteelBack         : TAction;
    ActToggleScreenStart       : TAction;
    ActToggleTriggerAreas      : TAction;
    ActShowDosLevelPackEditor  : TAction;
    ActZoomIn                  : TAction;
    ActZoomOut                 : TAction;
  { styles ... }
    ActStyleCompiler           : TAction;
    //ActAddStyle                : TAction;
  { tools ... }
    ActValidateLevel           : TAction;
    ActDosDatDecompress        : TAction;
    ActPlay                    : TAction; 
  { help ... }
    ActAbout                   : TAction;
  { level messages }
    procedure LMLevelChanging(var Msg: TLMLevelChanging); message LM_LEVELCHANGING;
    procedure LMLevelChanged(var Msg: TLMLevelChanged); message LM_LEVELCHANGED;
    procedure LMStaticsChanged(var Msg: TLMStaticsChanged); message LM_STATICSCHANGED;
    procedure LMObjectListChanged(var Msg: TLMObjectListChanged); message LM_OBJECTLISTCHANGED;
    procedure LMTerrainListChanged(var Msg: TLMTerrainListChanged); message LM_TERRAINLISTCHANGED;
    procedure LMSteelListChanged(var Msg: TLMSteelListChanged); message LM_STEELLISTCHANGED;
    procedure LMSelectionChanged(var Msg: TLMSelectionChanged); message LM_SELECTIONCHANGED;
  { internal methods image editing }
    procedure SelectAllUnderRect(const aRect: TRect);
    function GetSelectionDragBitmap(Bmp: TBitmap; var HotX, HotY: Integer): Boolean;
    function LevelActive: Boolean;
    function ZoomIn(MouseX: Integer = MinInt; MouseY: Integer = MinInt): Boolean;
    function ZoomOut(MouseX: Integer = MinInt; MouseY: Integer = MinInt): Boolean;
//    procedure UpdateScreenCaption;
    procedure UpdatePanelResolution;
    procedure UpdatePanelScale;
    procedure UpdateSelectionLayer;
    function CalcCenteredInsertionPoint: TPoint;
    procedure CreateActionsAndMenus;
    function MapScaleIndex(const aScale: Single): Integer;
    procedure HandleCommandStr;
    procedure Rerender;
    function CheckCanDrag: Boolean;
  { internal methods }
    procedure InternalPlayLevel(aLevel: TLevel);
    procedure InternalInsertObject(P: TPoint);
    procedure InternalInsertTerrain(P: TPoint);
    procedure InternalInsertSteel(P: TPoint);
    procedure InternalOpenFile(const aFileName: string; aType: Integer);
  { private eventhandlers }
    procedure SelectionLayer_Paint(Sender: TObject; Buffer: TBitmap32);
    procedure ScreenStartLayer_Paint(Sender: TObject; Buffer: TBitmap32);
    procedure Form_DropFiles(Sender: TObject; const Files: TStrings);
  { app events }
  { action eventhandlers }
  { file ... }
    procedure ActNewLevel_Execute(Sender: TObject);
    procedure ActNewDosLevelPack_Execute(Sender: TObject);
    procedure ActNewDosExtGraphic_Execute(Sender: TObject);
    procedure ActOpenFile_Execute(Sender: TObject);
    procedure ActOpenRecentFile_Execute(Sender: TObject);
    procedure ActCloseLevel_Execute(Sender: TObject);
    procedure ActQuit_Execute(Sender: TObject);
    procedure ActSaveAs_Execute(Sender: TObject);
    procedure ActOptions_Execute(Sender: TObject);
  { edit ... }
    procedure ActInsertTerrain_Execute(Sender: TObject);
    procedure ActInsertObject_Execute(Sender: TObject);
    procedure ActInsertSteel_Execute(Sender: TObject);
    procedure ActCopy_Execute(Sender: TObject);
    procedure ActCut_Execute(Sender: TObject);
    procedure ActPaste_Execute(Sender: TObject);
    procedure ActPasteSpecial_Execute(Sender: TObject);
  { selection ... }
    procedure ActAlignSelection_Execute(Sender: TObject);
    procedure ActDosAlignSelection_Execute(Sender: TObject);
  { view ... }
    procedure ActShowInspector_Execute(Sender: TObject);
    procedure ActShowStatics_Execute(Sender: TObject);
    procedure ActShowList_Execute(Sender: TObject);
    procedure ActShowImage_Execute(Sender: TObject);
    procedure ActToggleTerrain_Execute(Sender: TObject);
    procedure ActToggleObjects_Execute(Sender: TObject);
    procedure ActToggleSteel_Execute(Sender: TObject);
    procedure ActToggleSteelBack_Execute(Sender: TObject);
    procedure ActToggleTriggerAreas_Execute(Sender: TObject);
    procedure ActToggleScreenStart_Execute(Sender: TObject);
    procedure ActZoomIn_Execute(Sender: TObject);
    procedure ActZoomOut_Execute(Sender: TObject);
    procedure ActViewLevels_Execute(Sender: TObject);
    procedure ActShowDosLevelPackEditor_Execute(Sender: TObject);
  { styles ... }
    procedure ActStyleCompiler_Execute(Sender: TObject);
  { tools ... }
    procedure ActValidateLevel_Execute(Sender: TObject);
    procedure ActDosDatDecompress_Execute(Sender: TObject);
    procedure ActPlay_Execute(Sender: TObject);
  { help ... }
    procedure ActAbout_Execute(Sender: TObject);

  {$ifdef develop}
    procedure ActTest_Execute(Sender: TObject);
    procedure ActShowSpriteFiles_Execute(Sender: TObject);
    procedure ActShowDOSGroundFile_Execute(Sender: TObject);
    procedure ActPackExtGraphics_Execute(Sender: TObject);
    procedure ActWizardLemmini_Execute(Sender: TObject);
  {$endif}

    procedure SetCommandStr(const Value: string);
    property CommandStr: string read fCommandStr write SetCommandStr;

  protected
    fCC: string;
    procedure DoAfterSetEditor(Value: TLevelEditor); override;

  public


//    procedure SetEditInfo(Who: TObject);
    procedure UpdateScreenCaption;

  end;

var
  GlobalEditorForm: TMapEditorForm;

implementation

{$R *.dfm}

uses
  LemApp, LemDialogs,
  LemWinStyles;

const
  panel_resolution = 0;
  panel_scale = 1;
  panel_imgcoord = 2;
  panel_levelcoord = 3;
  panel_selection = 4;
  panel_commandstr = 5;

//  OFD_ALL = 1;
  OFD_RAWLEVEL = 1;
  OFD_LEMMINI = 2;
  OFD_DOSLEVELPACK = 3;
  OFD_VGASPEC = 4;

  {OFD_DOSLEVELXXX = 6;
  OFD_DOSGROUNDXXDAT = 7;
  OFD_VGAGRXDAT = 8;
  OFD_UNKNOWNDAT = 9; }

  OpenDlgFilters : array[1..4] of string = (
//    'All files (*.*)|*.*|', // ofd_all
    'Raw Level file (*.lvl)|*.lvl|', // ofd_rawlevel
    'Lemmini file (*.ini)|*.ini|', // ofd_lemmini
    'DOS Levelpack file (*.dat)|*.dat|', // ofd_doslevelpack
    'DOS Extended Graphics file (vgaspec?.dat)|vgaspec?.dat|'
  );
//    'DOS Ground info file (ground??.dat)|ground??.dat|',
  //  'DOS VGA info file (vgagr?.dat)|vgagr?.dat|',
    //'DOS Generic DAT file (*.dat)|*.dat|'
//  );

  SaveAsFilters : array[1..2] of string = (
    'Lemmini file (*.ini)|*.ini|',
    'Raw Level file (*.lvl)|*.lvl'
  );


  // dos original levelpack
  // dos original extracted level

  // dos ohno levelpack
  // dos ohno extracted level

  //

  // winlemmings low resolution extracted level
  // winlemmings high resolution extracted level

  // lemmini level

type
  TFileType = (
    ftNone,
    ftLevel,
    ftLevelPack
  );



const
  AVAILABLESCALECOUNT = 11;

  AvailableScales: array[0..AVAILABLESCALECOUNT - 1] of Single = (
    0.125,
    0.25,
    0.5,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    8
  );

const
//  ScaleStrings: array[0..10] of string =

  AvailableScaleNames : array[0..AVAILABLESCALECOUNT - 1] of string = (
    '1:8', '1:4', '1:2', '1:1', '2:1', '3:1', '4:1', '5:1', '6:1', '7:1', '8:1'
  );


type
  TEditorDragObject = class(TDragControlObject)
  private
    fDragImageList: TDragImageList;
    fEditorForm: TMapEditorForm;
//    fDragX, fDragY: Integer;
  protected
    function GetDragImages: TDragImageList; override;
  public
    constructor Create(aControl: TControl); override;
    destructor Destroy; override;
  published
  end;

  TRecentFileAction = class(TAction)
  private
    fFileType: string; // Level, LevelPack
    fFileInfo: string; // style classname if it's a level
    fFileName: string;
    fFileFormat: string;
  protected
  public
    property FileName: string read fFileName;
    property FileType: string read fFileType;
    property FileInfo: string read fFileInfo;
    property FileFormat: string read fFileFormat;
//    property FileInfo: string read fFileInfo write fFileInfo;
  published
  end;

  TFileTypeInfo = class
  private
    fDefaultFileMask: string;
  protected
  public
    property DefaultFileMask: string read fDefaultFileMask;
  published
  end;

  TFileMgr = class
  private
    fFileTypeList: TStringList;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    property FileTypeList: TStringList read fFileTypeList;
//    property utools
  published
  end;


{ TEditorDragObject }

constructor TEditorDragObject.Create(aControl: TControl);
begin
  inherited;
end;

destructor TEditorDragObject.Destroy;
begin
  //windlg('drag.destroy');
  fDragImageList.Free;
  inherited;
end;

function TEditorDragObject.GetDragImages: TDragImageList;
var
  Bmp: TBitmap;
  Ix: Integer;
//  Buf, Scaler: TBitmap32;
  hx,hy:integer;
begin
//  Result := nil; exit;

  if fDragImageList = nil then
    fDragImageList := TDragImageList.Create(nil);

  Result := fDragImageList;
  Result.Clear;
  begin

//    ix:=0;
    with fDragImageList do
    if Count = 0 then
    begin
      Bmp := TBitmap.Create;
      //Scaler := TBitmap32.Create;

      fEditorForm.GetSelectionDragBitmap(bmp, hx, hy);
      {
      bmp.width := 50;
      bmp.height := 50;
      bmp.canvas.Brush.color := clred;
      bmp.canvas.fillrect(rect(0, 0, 50, 50));
      }

      {Buf := PictureMgr.TerrainBitmaps[fGraphicSet, fTerrain.TerrainId];
      Scaler.Width := Buf.Width * 2;
      Scaler.Height := Buf.Height * 2;
      Scaler.Draw(Rect(0, 0, buf.Width * 2, buf.height * 2), rect(0,0,buf.width, buf.Height), buf);
      Bmp.Assign(Scaler);
      Scaler.Free;}



      Width := Bmp.Width;
      Height := Bmp.Height;
//      Deb([width]);
      Ix := Result.AddMasked(Bmp, clBlack);
      //Deb([ix]);
      Bmp.Free;
//Deb(['ok']);
//      Width := TestImg.Picture.Bitmap.Width;
  //    Height := TestImg.Picture.Bitmap.Height;
    //  Add(TestImg.Picture.Bitmap, nil);
      Result.SetDragImage(Ix, hx, hy);//fEditorForm.DragP.X, fEditorForm.DragP.Y);
    end;

//    Deb([ix]);


  end;

end;

{ TFileMgr }

constructor TFileMgr.Create;
//var
  //i: Integer;
begin
  inherited;
  fFileTypeList := TStringList.Create;
  fFileTypeList.Add('level,lvl');
  fFileTypeList.Add('level,lemmini');
  fFileTypeList.Add('levelpack,dos');

end;

destructor TFileMgr.Destroy;
begin
  fFileTypeList.Free;
  inherited;
end;

{ TMapEditorForm }

procedure TMapEditorForm.Form_Create(Sender: TObject);
{-------------------------------------------------------------------------------
  So what do we need:
  o Objects
    o A Link that sends change notifications from the leveleditor.
    o A Bitmap renderer to display a lemming level.
    o A HitList for mouse selection.
    o A Selection layer for displaying selection on top of the image.
    o A ScreenStart layer for displaying the initial screen
    o A Dragobject for dragging selection around the image.
    o Inspector forms
  o Dragging:
    o Let the form and all its subcontrols display dragimages.
    o Accept dragged files for opening a level.
  o Display:
    o Colors for selection
  o Actions, mainmenu, local menus
-------------------------------------------------------------------------------}
begin
  fRenderer := TLevelRenderer.Create(Self);
  fRenderer.RenderOptions := fRenderer.RenderOptions -
                             [roCollisionGridBuilding{, roInitialScreenPosRect}] + [roSteelNoOverwrite];
//                             + [roShowTriggerAreas];
  fRenderer.LevelBitmap := Img.Bitmap;
  //fRenderer.Resolution := GlobalResolution;
  //fRenderer.OnRenderItem := Renderer_RenderItem;
  fHitList := TList.Create;
  fScaleIndex := MapScaleIndex(App.ScreenSettings.EditorScale);
  Img.Scale := AvailableScales[fScaleIndex];
  Img.Height := iif(App.EditorSettings.MapHeight > 0, App.EditorSettings.MapHeight, Img.Height);
  Img.Options := [pboWantArrowKeys, pboAutoFocus];
  fSelectionLayer := TPositionedLayer.Create(Img.Layers);
  fSelectionLayer.MouseEvents := False;
  fSelectionLayer.Scaled := True;
  fSelectionLayer.OnPaint := SelectionLayer_Paint;
  fScreenStartLayer := TPositionedLayer.Create(Img.Layers);
  fScreenStartLayer.MouseEvents := False;
  fScreenStartLayer.Scaled := True;
  fScreenStartLayer.OnPaint := ScreenStartLayer_Paint;
  fScreenStartLayer.Visible := False;
  fSelectorColors[litObject] := clLime32;
  fSelectorColors[litTerrain] := clYellow32;
  fSelectorColors[litSteel] := clRed32;
  fStaticsForm := TFormEditStatics.Create(Self);
  fListForm := TLevelListEditorForm.Create(Self);
  fEditItemForm := TEditItemForm.Create(Self);
  fDragger := TEditorDragObject.Create(Img);
  TEditorDragObject(fDragger).fEditorForm := Self;
  SetDisplayDragImage(Self);
  AcceptDraggedFiles := True;
  OnDropFiles := Form_DropFiles;
  CreateActionsAndMenus;
  fRenderer.RenderLevel(nil);
end;

procedure TMapEditorForm.CreateActionsAndMenus;
{-------------------------------------------------------------------------------
  We create the menu + actions at runtime.
-------------------------------------------------------------------------------}

    function NewMenubarItem(const aCaption: string): TMenuItem;
    // local function to add menubar item to the mainmenu
    begin
      Result := TMenuItem.Create(Self);
      Result.Caption := aCaption;
      MainMenu.Items.Add(Result);
    end;

    function NewHolderItem(aParent: TMenuItem; const aCaption: string): TMenuItem;
    // local function to add a menuitem (for holding submenus) to the mainmenu
    begin
      Result := TMenuItem.Create(Self);
      Result.Caption := aCaption;
      aParent.Add(Result);
    end;

    procedure NewSep(aParent: TMenuItem);
    // local function to add a seperator to the main menu
    var
      M: TMenuItem;
    begin
      if aParent <> nil then
      begin
        M := TMenuItem.Create(Self);
        M.Caption := '-';
        aParent.Add(M);
      end;
    end;

    function NewAction(aParent: TMenuItem;
                       const aCaption: string;
                       aExecuteHandler: TNotifyEvent;
                       aShortCut: TShortCut = 0): TAction;

    var
      M: TMenuItem;
    begin
      Result := TAction.Create(Self);
      Result.Caption := aCaption;
      Result.OnExecute := aExecuteHandler;
      Result.ShortCut  := aShortCut;
      Result.ActionList := ActionList;
      if aParent <> nil then
      begin
        M := TMenuItem.Create(Self);
        M.Action := Result;
        aParent.Add(M);
      end;
    end;


    function NewMapPopupAction(aParent: TMenuItem;
                               const aCaption: string;
                               aExecuteHandler: TNotifyEvent): TAction;

    var
      M: TMenuItem;
    begin
      Result := TAction.Create(Self);
      Result.Caption := aCaption;
      Result.OnExecute := aExecuteHandler;
      Result.ActionList := ActionList;
      M := TMenuItem.Create(Self);
      M.Action := Result;
      if aParent <> nil then
        aParent.Add(M)
      else
        MapPopup.Items.Add(M);
    end;


    (*
    procedure AddRecentFiles;
//    const
  //    PrefixChars: array[0..9 +
    var
      Act: TRecentFileAction;
      M: TMenuItem;
      i: Integer;
      C: Char;
      S: string;
    begin
      Assert(MhRecentFiles <> nil, 'recentfiles menu item error');
      with App.RecentFilesSettings do
      begin
        for i := 0 to FileStrings.Count - 1 do
        begin
//          Deb([FileNames[i]]);
          Act := TRecentFileAction.Create(Self);
          S := FileStrings[i];
          Act.fFileName := SplitString(S, 0, ',');
          Act.fFileType := SplitString(S, 1, ',');
          Act.fFileInfo := SplitString(S, 2, ',');
          Act.fFileFormat := SplitString(S, 3, ',');
          if i < 10 then
            C := i2s(i)[1]
          else if i < 36 then
            C := Chr(Ord('A') + i - 10);
          Act.Caption := C + ' ' + Act.FileName;
//          Act.fFileType := FileTypes[i];
//          Act.fFileInfo := FileInfo[i];
          Act.OnExecute := ActOpenRecentFile_Execute;
          M := TMenuItem.Create(Self);
          M.Action := Act;

          MhRecentFiles.Add(M);
          MhRecentFiles.Enabled := True;
        end;
      end;
    end;
    *)


var
  A: TAction;
  MhNew: TMenuItem;
//  MhZoomHolder: TMenuItem; // holder for all zooms
  MhViewLevel: TMenuItem;
//  M: TMenuItem;
  i: Integer;
begin

  MbEdit      := NewMenuBarItem('&Edit');
  MbSelection := NewMenuBarItem('&Selection');
  MbView      := NewMenuBarItem('&View');
  MbTools     := NewMenuBarItem('&Tools');
  MbHelp      := NewMenuBarItem('&Help');

  {
  MbFile.GroupIndex := 1;
  MbSelection.GroupIndex := 2;
  MbView.GroupIndex := 3;
  MbTools.GroupIndex := 4;
  MbHelp.GroupIndex := 255; }

//  MainMenu.AutoMerge := True;


  {$ifdef develop}
  MbDevelop := NewMenuBarItem('&Develop');
  {$endif}

  // file actions
  MhNew := NewHolderItem(mbFile, '&New');
  ActNewLevel := NewAction(MhNew, '&Level...', ActNewLevel_Execute);
  NewSep(MhNew);
  ActNewDosLevelPack := NewAction(MhNew, '&DOS Level pack...', ActNewDosLevelPack_Execute);
  ActNewDosExtGraphic := NewAction(MhNew, '&DOS Extended Graphic File...', ActNewDosExtGraphic_Execute);

  ActOpenFile := NewAction(MbFile, '&Open...', ActOpenFile_Execute, ShortCut(Ord('O'), [ssCtrl]));


  {
  MhRecentFiles := NewHolderItem(mbFile, '&Reopen');
  MhRecentFiles.Enabled := False;
  AddRecentFiles;
  }
  

  ActCloseLevel := NewAction(MbFile, '&Close level...', ActCloseLevel_Execute, ShortCut(Ord('W'), [ssCtrl]));
  ActSaveAs := NewAction(MbFile, 'Save &As...', ActSaveAs_Execute);
//  ActSaveInDosPack := NewAction(MbFile, 'Save in DOS levelpack', ActSaveInDosPack_Execute);
  ActOptions := NewAction(MbFile, 'Settings...', ActOptions_Execute);
  NewSep(MbFile);
  ActQuit := NewAction(MbFile, 'E&xit', ActQuit_Execute, ShortCut(Ord('X'), [ssAlt]));


  // edit actions
  ActInsertTerrain := NewAction(MbEdit, 'Insert &Terrain', ActInsertTerrain_Execute, ShortCut(Ord('T'), [ssCtrl, ssShift]));
  ActInsertTerrainLocal := NewMapPopupAction(nil, 'Insert &Terrain', ActInsertTerrain_Execute);

  ActInsertObject := NewAction(MbEdit, 'Insert &Object', ActInsertObject_Execute, ShortCut(Ord('O'), [ssCtrl, ssShift]));
  ActInsertObjectLocal := NewMapPopupAction(nil, 'Insert &Object', ActInsertObject_Execute);

  ActInsertSteel := NewAction(MbEdit, 'Insert &Steel', ActInsertSteel_Execute, ShortCut(Ord('S'), [ssCtrl, ssShift]));
  ActInsertSteelLocal := NewMapPopupAction(nil, 'Insert &Steel', ActInsertSteel_Execute);

  ActCopy := NewAction(MbEdit, '&Copy', ActCopy_Execute, ShortCut(Ord('C'), [ssCtrl]));
  ActCut := NewAction(MbEdit, '&Cut', ActCut_Execute, ShortCut(Ord('X'), [ssCtrl]));
  ActPaste := NewAction(MbEdit, '&Paste', ActPaste_Execute, ShortCut(Ord('V'), [ssCtrl]));
  ActPasteSpecial := NewAction(MbEdit, 'Paste spe&cial', ActPasteSpecial_Execute, ShortCut(Ord('V'), [ssCtrl, ssShift]));


  // selection actions
  ActAlignSelection := NewAction(MbSelection, '&Align...', ActAlignSelection_Execute);
  ActDosAlignSelection := NewAction(MbSelection, '&Align to grid used by DOS...', ActDosAlignSelection_Execute);

  // view actions
  ActZoomIn := NewAction(MbView, 'Zoom in', ActZoomIn_Execute);
  ToolButton1.Action := ActZoomIn;
  ActZoomOut := NewAction(MbView, 'Zoom out', ActZoomOut_Execute);
  ToolButton2.Action := ActZoomOut;
//  A :=
  NewSep(MbView);
  {
  mhZoomHolder := NewHolderItem(mbView, 'Zoom');
  for i := 0 to AVAILABLESCALECOUNT - 1 do
  begin
    A := NewAction(mhZoomHolder, Chr(Ord('a') + i) + ' ' + AvailableScaleNames[i], nil)
  end; }
//  asdf



  ActShowImage := NewAction(MbView, 'Map editor', ActShowImage_Execute, ShortCut(VK_F12, []));
  ActShowInspector := NewAction(MbView, 'Inspector window', ActShowInspector_Execute, ShortCut(VK_F11, []));
  ActShowStatics := NewAction(MbView, 'Level properties window', ActShowStatics_Execute, ShortCut(VK_F10, []));
  ActShowList := NewAction(MbView, 'Level List window', ActShowList_Execute, ShortCut(VK_F9, []));
  NewSep(MbView);
  ActToggleTerrain := NewAction(MbView, 'Terrain rendering', ActToggleTerrain_Execute, ShortCut(VK_F5, []));
  ActToggleObjects := NewAction(MbView, 'Object rendering', ActToggleObjects_Execute, ShortCut(VK_F6, []));
  ActToggleSteel := NewAction(MbView, 'Steel rendering', ActToggleSteel_Execute, ShortCut(VK_F7, []));
  ActToggleSteelBack := NewAction(MbView, 'Steel in background', ActToggleSteelBack_Execute, ShortCut(VK_F8, []));
  ActToggleScreenStart := NewAction(MbView, 'Screen start', ActToggleScreenStart_Execute);
  ActToggleTriggerAreas := NewAction(MbView, 'Trigger areas', ActToggleTriggerAreas_Execute);
  NewSep(MbView);
  MhViewLevel := NewHolderItem(MbView, 'Levels viewer');
  for i := 0 to StyleMgr.StyleList.Count - 1 do
  begin
    A := NewAction(mhViewLevel, StyleMgr[i].StyleDescription, ActViewLevels_Execute);
    A.Tag := i + 100;
  end;
  ActShowDosLevelPackEditor := NewAction(MbView, 'Dos Levelpack editor', ActShowDosLevelPackEditor_Execute);


  // tools actions
  ActValidateLevel := NewAction(MbTools, 'Validate Level', ActValidateLevel_Execute);
  NewSep(mbTools);
  ActStyleCompiler := NewAction(MbTools, 'Compiler...', ActStyleCompiler_Execute);
  NewSep(mbTools);
  ActDosDatDecompress := NewAction(MbTools, 'Dos File Decompression...', ActDosDatDecompress_Execute);
  ActPlay := NewAction(MbTools, 'Play this level', ActPlay_Execute, ShortCut(VK_F2, []));
//  {$ifndef develop}
//  ActPlay.Enabled := False;
//  {$endif} lemstrings

  // help actions
  ActAbout := NewAction(MbHelp, 'About...', ActAbout_Execute);





  // test actions
  {$ifdef develop}
  NewAction(MbDevelop, 'Play', ActPlay_Execute, ShortCut(VK_F2, []));
  NewAction(MbDevelop, 'Test...', ActTest_Execute, ShortCut(VK_F3, []));
  NewAction(MbDevelop, 'Show WinLemmings SPR Files', ActShowSpriteFiles_Execute);
  NewAction(MbDevelop, 'Show DOS Ground File Form', ActShowDOSGroundFile_Execute);
  NewSep(MbDevelop);
  NewAction(MbDevelop, 'Pack extended graphics', ActPackExtGraphics_Execute);
  NewSep(MbDevelop);
  NewAction(MbDevelop, 'Lemmini Compiler Wizard', ActWizardLemmini_Execute);
  {$endif}


  // init
  ActToggleTerrain.Checked := True;
  ActToggleObjects.Checked := True;
//  ActToggleSteel.Checked := True;
  ActToggleSteelBack.Checked := True;

end;

procedure TMapEditorForm.Form_Destroy(Sender: TObject);
begin
  if Editor <> nil then
    App.EditorSettings.LastFile := Editor.FileName;

//  fEditorMapLink.Free;
  App.ScreenSettings.EditorScale := Img.Scale;
  App.EditorSettings.MapHeight := Img.Height;
  fHitList.Free;
  fDragger.Free;
end;

procedure TMapEditorForm.Form_Show(Sender: TObject);
var
  Item: TFormSettingItem;
begin
//  ShowMessage('1');
  Item := TFormSettingItem.Create(nil);
  try
    App.ReadForm(Self, Item);
    if Item.Height > 0 then BEGIN
    Self.WindowState := Item.WinState;
    //aForm.Visible := Item.Visible;
    if Self.WindowState = wsNormal then
    begin
    Self.Left := Item.Left;
    Self.Top := Item.Top;
    Self.Width := Item.Width;
    Self.Height := Item.Height;
    end;
//    if Item.StayOnTop then
  //     SaForm.FormStyle := fsStayOnTop;
  END;
  finally
    Item.Free;
  end;
//  ShowMessage('2');

(*

  *)

{  if not fInspectorForm.Visible then
  begin
    fInspectorForm.Left := 300;
    fInspectorForm.Top := 320;
    fInspectorForm.Show;
  end; }
  exit;
  if not fStaticsForm.Visible then
  begin
  //  fStaticsForm.Left := 400;
//    fStaticsForm.Top := 330;
    fStaticsForm.Show;
  end;
  if not fListForm.Visible then
  begin
  //  fListForm.Left := 500;
//    fListForm.Top := 320;
    fListForm.Show;
  end;
  if not fEditItemForm.Visible then
  begin
  //  fEditItemForm.Left := 550;
//    fEditItemForm.Top := 330;
    fEditItemForm.Show;
  end;
end;

procedure TMapEditorForm.UpdateSelectionLayer;
//var
  //Sel: TTerrain;
 // R: TRect;
begin
  (*
  Sel := LevelEditor.SelectedTerrain;
  if Sel <> nil then
  begin
    R.Left := Sel.TerrainX;
    R.Top := Sel.TerrainY;
    R.Right := R.Left + Sel.ItemWidth;
    R.Bottom := R.Top + Sel.ItemHeight;
//    R := MakeRect(fSelectionLayer.GetAdjustedLocation);
//    R := Rect(0, 0, 10, 10);
//    Deb([rectstr(r)]);

    fSelectionLayer.Location := FloatRect(R);
//    Deb([trunc(r.left)]);
  end; *)
//  fSelectionLayer.Visible := {Level}Editor.SelectorCollection.Count > 0; //Sel <> nil;
  fSelectionLayer.Changed;
//  else fSelection
end;

function TMapEditorForm.MapScaleIndex(const aScale: Single): Integer;
{-------------------------------------------------------------------------------
  Internal function returns the index of AvailableScales, closest to aScale.
  0.0 is mapped to a scale of 1.0
-------------------------------------------------------------------------------}
var
  i: Integer;
begin
  Result := 3;
  if aScale = 0.0 then
    Exit;
  for i := Low(AvailableScales) to High(AvailableScales) do
    if aScale = AvailableScales[i] then
    begin
      Result := i;
      Exit;
    end;
end;

procedure TMapEditorForm.Img_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
{-------------------------------------------------------------------------------
  A lot of edit functions are available with the keyboard:
  o Left, Right, Up, Down: Moving
  o Delete: Deleting
  o Insert: Inserting
  o A/O/T, Selecting
-------------------------------------------------------------------------------}
var
//  H, X, Y: Integer;
  P: TPoint;
  //B: TBaseLevelItem;
  //Dist: Integer;
  //Steel: TSteel;

    (*
    function GetDist: Integer;
    const
      Distances: array[1..3] of Integer = (1, 32, 8);
    begin
      Result := 3;

      if Shift = [] then
        Result := 3
      else if Shift = [ssCtrl] then
        Result := 1
      else if Shift = [ssCtrl, ssShift] then
        Result := 2;

      Result := Distances[Result];
    end; *)

    function GetDist: Integer;
    begin
      Result := App.EditorSettings.DefaultArrowKeysDistance;

//      if Shift = [] then
  //      Result := 3
//      else
      if Shift = [ssCtrl] then
        Result := App.EditorSettings.DefaultArrowKeysCtrlDistance;
//      else if Shift = [ssCtrl, ssShift] then
  //      Result := 2;

    //  Result := Distances[Result];
      if (Result < 0) or (Result > 128) then
        Result := 1; 
    end;

begin
  if not LevelActive then Exit;
  case Key of
//    Ord('C')

    VK_TAB:
      begin
        //editor.selectnext
        if ssCtrl in Shift  then
        begin
          editor.SelectNextItem(not (ssShift in Shift));
//          if Editor.SelectorCount = 0 then
{            if Editor.TerrainCollection.Count > 0 then
             begin
               Editor.SelectedTerrain := Editor.TerrainCollection[0];
               Exit;
             end; }


{          if Editor.SelectorCount > 0 then
            if Editor.TerrainCollection.Count > 0 then
             begin
               Editor.SelectedTerrain := Editor.TerrainCollection[0];
               Exit;
             end; }



        end;
          //if Editor.SelectorCount = 1 then
          begin

            //H := Editor.Selectors[0].RefItem.Index;
//            Deb(['next', h]);
          end;
      end;
    VK_RETURN:
      begin
        HandleCommandStr;
      end;
    VK_LEFT:
      begin
       Editor.MoveSelection(-GetDist, 0);
       Img.Update;
      end;
    VK_RIGHT:
      begin
        begin
          (*
          if Shift = [ssShift] then
            if Editor.SelectorCount = 1 then
              if Editor.SelectedSteel <> nil then
              begin
                Steel := Editor.SelectedSteel;
                Steel.SteelWidth := Steel.SteelWidth + 4{GetDist};
                Img.Update;
                Exit;
              end; *)
          Editor.MoveSelection(GetDist, 0);
          Img.Update;
        end;
      end;
    VK_UP:
      begin
        Editor.MoveSelection(0, -GetDist);
        Img.Update;
      end;
    VK_DOWN:
      begin
        Editor.MoveSelection(0, GetDist);
        Img.Update;
      end;
    VK_DELETE:
      begin
        Editor.DeleteSelection;
      end;
    VK_INSERT:
      begin
        ActInsertTerrain.Execute;
      end;

    VK_F10:
      if Shift = [ssShift] then
      begin
        P := Point(Img.Left + Img.Width div 2, Img.Top + Img.Height div 2);
        P := Img.ClientToScreen(P);
        MapPopup.Popup(P.X, P.Y);
      end;

  end;
end;

procedure TMapEditorForm.Img_KeyPress(Sender: TObject; var Key: Char);
//var
  //T: TTerrain;
  //p: tpoint;
//  S: string
begin
(*  if not (Key in ['0'..'9']) then
    if Key <> #13 then
    if Key <> #8 then
      CommandStr := ''; *)

  case Key of
    'i':
      begin
(*        if editor.selectorcount <> 1 then exit;
        p.x := editor.Selectors[0].RefItem.xposition;
        p.y := editor.Selectors[0].RefItem.yposition;
//        p:=img.BitmapToControl(p);
        img.ScrollToCenter(p.x, p.y); *)
      end;

    #8:
      begin
        CommandStr := CutRight(CommandStr, 1);
      end;
    '0'..'9':
      begin
        if GetTickCount - fCommandChangeTicks > 1000 then
          CommandStr := '';
        if Length(fCommandStr) = 1 then
          CommandStr := CommandStr + Key
        else
          CommandStr := Key;

      end;
    '=', '+', '.', '>' : ZoomIn;//(MinInt, MinInt);
    '-', ',', '<'      : ZoomOut;
    'a':
      begin
        Editor.SelectAll;
      end;

    'o':
      begin
        Editor.SelectAll([litObject]);
      end;
    'O':
      begin
        Editor.FilterSelection([litObject])
      end;
    't':
      begin
        Editor.SelectAll([litTerrain]);
      end;
    'T':
      begin
        Editor.FilterSelection([litTerrain])
      end;
    's':
      begin
        Editor.SelectAll([litSteel]);
      end;
    'S':
      begin
        Editor.FilterSelection([litSteel])
      end;

    'C', 'c':
      begin
        Editor.SelectorCollection.Clear;
      end;
    'H', 'h':
      begin
//         Deb(['toggle hide']);
        fSelectionLayer.Visible := not fSelectionLayer.Visible;
      end;
  end;
end;

procedure TMapEditorForm.Img_MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
{-------------------------------------------------------------------------------
  Here is a lot of the mouselogic when editing a level.
  o Left button (plain) selects an object under the mousecursor.
    If the object is not selected it will be now.
    If the object is already selected drag/drop begins.
    If it's a doubleclick and we only hit air, then the selection is cleared.
  o Left button + Ctrl begins drawing a selection rectangle. Further handling in
    MouseMove and MouseUp. When we're in this mode the mouse is clipped.
  o Left button + Shift extends the selection with the hitted object.
    Then we follow the settings:
    if DirectSelect is False: If the hitted object is already selected
    it is removed from the selection.
    if DirectSelect is True: Dragging is controlled in mousemove
-------------------------------------------------------------------------------}
var
  P: TPoint;
  Hit: Integer;
  Lev: TLevelEditor;
  R: TRect;
  Filt: TLevelItemTypes;
begin

   fMouseOnSelection := False;
   DragSelection := False;
//  Img.SetFocus;
  if not LevelActive then Exit;
  if layer<>nil then exit;
  Lev := Editor;

{  if not fDrawingSelectionRect and (Button = mbRight) then
  begin
    P := Img.ClientToScreen(Point(X, Y));
    MapPopup.Popup(P.X, P.Y);
    Exit;
  end; }

  // selection rectangle
  if (Button = mbLeft) and (Shift * [ssCtrl, ssShift, ssAlt] = [ssCtrl]) then
  begin
    fDrawingSelectionRect := True;
    R := Img.BoundsRect;
    R.TopLeft := ClientToScreen(R.TopLeft);
    R.BottomRight := ClientToScreen(R.BottomRight);
    ClipCursor(@R);
    with fSelectionRect do
    begin
      TopLeft := Point(X, Y);
      BottomRight := TopLeft;
    end;
  end

  else if (Button = mbLeft) or (Button = mbRight) then
  begin
    P := Img.ControlToBitmap(Point(X, Y));
    fHitList.Clear; // belangrijk
    Filt := [litObject, litTerrain, litSteel] * fRenderer.RenderTypes; // chique
    if Filt = [] then
      Exit;

    Hit := fRenderer.HitTest(Lev, p.x, p.y, fHitList, Filt, 1);

    if Hit >= 0 then
      LastHitItem := fHitList.List^[0]
    else
      LastHitItem := nil;

    // nothing hitted
    if Hit < 0 then
    begin
      // air-click with right button
      if Button = mbRight then
      begin
        P := Img.ClientToScreen(Point(X, Y));
        MapPopup.Popup(P.X, P.Y);
        Exit;
      end
      // clear selection if left button
      else begin
        if App.EditorSettings.ClearSelectionOnAirClick then
          Lev.SelectorCollection.Clear
        else if ssDouble in Shift then
          Lev.SelectorCollection.Clear;
      end
    end
    // there is a hit
    else if Hit >= 0 then
    begin
      if Button = mbRight then
        Exit;
      // no shift keys
      if Shift * [ssCtrl, ssShift, ssAlt] = [] then
      begin
        // drag already made selection
        if Lev.SelectorCollection.FindBaseItem(TBaseLevelItem(fHitList[0])) >= 0 then
        begin
          if not CheckCanDrag then Exit;
          fMouseOnSelection := True;
          if not App.EditorSettings.DirectDragOnSelect then
          begin
            DragIP := Point(X, Y);
            DragBP := Img.ControlToBitmap(DragIP);
//            Img.Update;
            Img.BeginDrag(False);
            Exit;
          end;
        end
        // make new selection
        else begin
          if TObject(fHitList[0]) is TTerrain then
            Lev.SelectedTerrain := fHitList[0]
          else if TObject(fHitList[0]) is TInteractiveObject then
            Lev.SelectedObject := fHitList[0]
          else if TObject(fHitList[0]) is TSteel then
            Lev.SelectedSteel := fHitList[0];
          fMouseOnSelection := True;
        end;
      end
      // shift key pressed
      else if Shift * [ssCtrl, ssShift, ssAlt] = [ssShift] then
      begin
        Lev.ExtendSelection(fHitList, True);
      end;
    end; // hit >= 0

  end;
end;

procedure TMapEditorForm.Img_MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
type
  TMouseMoveAction = (mmNone, mmDrawRect, mmDrag);
var
//  DoRect, DoDrag: Boolean; UMISC
  MM: TMouseMoveAction;
  P: TPoint;
begin
  StatusBar.Panels[panel_imgcoord].Text := i2s(X) + ':' + i2s(Y);
  P := Img.ControlToBitmap(Point(X, Y));
  StatusBar.Panels[panel_levelcoord].Text := i2s(P.X) + ':' + i2s(P.Y);
  MM := mmNone;
  if fDrawingSelectionRect then
//  and (ssCtrl in Shift) then
    MM := mmDrawRect
  else if not fImageIsDragging
  and fMouseOnSelection
  and (ssLeft in Shift)
  and  App.EditorSettings.DirectDragOnSelect
  and CheckCanDrag then //disable dragging when zoomed in
    MM := mmDrag;

  case MM of
    mmNone:
      begin
        Exit;
      end;
    mmDrawRect:
      with fSelectionRect, Img.Canvas do
      begin
        Brush.Style := bsClear;
        Pen.Color := clWhite;
        Pen.Style := psDot;
        Pen.Mode := pmXOR;
        if fSelectionRectVisible then
          Rectangle(fSelectionRect);
        BottomRight := Point(X, Y);
        Rectangle(fSelectionRect);
        fSelectionRectVisible := True;
      end;
    mmDrag:
      begin
        DragIP := Point(X, Y);
        DragBP := Img.ControlToBitmap(DragIP);
//        with dragbp do
  //        deb(['startdrag', x,y]);
        //DraggedItem := fHitList.List^[0];
        Img.Update;
        Img.BeginDrag(False);
      end;
  end;
end;

procedure TMapEditorForm.Img_MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
var
  DoSelect: Boolean;
  R: TRect;
//  P: TPoint;
begin
//  Deb(['mu']);
  DoSelect := False;
  ClipCursor(nil);
  fMouseOnSelection := False;
  (*

  if DragSelection then
  begin
    DragSelection := False;
//    Deb(['dropselection']);
//    Deb([dragx, x]);
    // delete if not in map
    P := Img.ControlToBitmap(Point(X,Y));
    LevelEditor.MoveSelection(P.X - DragP.X, P.Y - DragP.Y);
    Exit;
  end; *)

  if fDrawingSelectionRect then
  //  if ssCtrl in Shift then
      with fSelectionRect do
      begin
        with Img.Canvas do
        begin
          Brush.Style := bsClear;
          Pen.Color := clWhite;
          Pen.Style := psDot;
          Pen.Mode := pmXOR;
          if fSelectionRectVisible then
            Rectangle(fSelectionRect);
          DoSelect := True;
        end;
      end;
  fDrawingSelectionRect := False;
  fSelectionRectVisible := False;
  if DoSelect then
  begin
//    R := fSelectionRect;
    R.TopLeft := Img.ControlToBitmap(fSelectionRect.TopLeft);
    R.BottomRight := Img.ControlToBitmap(fSelectionRect.BottomRight);
    if R.Bottom < R.Top then SwapInts(r.bottom, r.top);
    if R.Right < R.Left then SwapInts(r.right, r.left);
    SelectAllUnderRect(R);
  end;
end;


procedure TMapEditorForm.Img_MouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
//deb([mousepos.y]);
  ZoomIn(MousePos.X, MousePos.Y);
  Handled := True;
end;

procedure TMapEditorForm.Img_MouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
  ZoomOut(MousePos.X, MousePos.Y);
  Handled := True;
end;

function TMapEditorForm.LevelActive: Boolean;
begin
  Result := (Editor <> nil)
end;


procedure TMapEditorForm.SelectionLayer_Paint(Sender: TObject; Buffer: TBitmap32);
var
  R: TRect;
//  FR: TFloatRect;
//  Sel: TTerrain;
  i: Integer;
  S: TSelector;
begin
  if fLocked then Exit;
  if fInternalGettingDragBitmap then Exit;
  if Editor = nil then
    Exit;
//  if LevelEditor.IsUpdating then
  //  Exit;

  with Editor.SelectorCollection do
    for i := 0 to HackedList.Count - 1 do
    begin
      S := Items[i];
      R := S.RefItem.Bounds;
      R := MakeRect(fSelectionLayer.GetAdjustedRect(FloatRect(R)));
      Buffer.FrameRectS(R.Left, R.Top, R.Right, R.Bottom, fSelectorColors[S.RefItem.ItemType]);
{      if i = 0 then
      begin
        InflateRect(R, -1, -1);
        Buffer.FrameRectS(R.Left, R.Top, R.Right, R.Bottom, fSelectorColors[S.RefItem.ItemType]);
      end; }
    end;

end;

procedure TMapEditorForm.ScreenStartLayer_Paint(Sender: TObject; Buffer: TBitmap32);
var
  R: TRect;
//  FR: TFloatRect;
//  Sel: TTerrain;
//  i: Integer;
//  S: TSelector;
  Def: TDefaultsRec;
begin
  if fLocked then Exit;
  if Editor = nil then
    Exit;
  if Editor.Style = nil then exit;
  if Editor.Graph = nil then exit;
//  if LevelEditor.IsUpdating then
  //  Exit;

  with Editor, Statics do
  begin
//    deb(['reevaluate screenstartlayer_paint']);
//    raise Exception.Create('reevaluate screenstartlayer_paint');

    Style.GetStyleParams(Def);

    R := Rect(0, 0, Def.ViewWidth, Def.LevelHeight);
    RectMove(R, ScreenPosition, 0);
    R := MakeRect(fScreenStartLayer.GetAdjustedRect(FloatRect(R)));
    Buffer.FrameRectS(R.Left, R.Top, R.Right, R.Bottom, clWhite32);

  end;
end;

function TMapEditorForm.ZoomIn(MouseX: Integer = MinInt; MouseY: Integer = MinInt): Boolean;
var
  P, Q: TPoint;
//  Xo, Yo: Single;
//  DistX : Integer;
  //L: TImgView32;
//  r:trect;
begin
  Result := fScaleIndex < High(AvailableScales);
  //L := Img;
  if Result then
  with Img do
  begin

    Inc(fScaleIndex);
    if (MouseX > MinInt) and (MouseY > MinInt) then
      Q := Img.ControlToBitmap(Point(MouseX, MouseY));
//    else
  //    Q := Img.ControlToBitmap(Point(Img.Width div 2, Img.Height div 2));
    Scale := AvailableScales[fScaleIndex];

    if Editor <> nil then
    begin
      if Editor.SelectorCount > 0 then
      begin
        P := Editor.SelectorCollection.Center;
        Img.ScrollToCenter(P.X, P.Y);
      end
      else if (MouseX > MinInt) and (MouseY > MinInt) then
      begin
        Img.ScrollToCenter(Q.X, Q.Y);
      end
      else begin
        //Img.ScrollToCenter(Img.Bitmap.Width div 2, Img.Bitmap.Height div 2);
      end;
    end;

  end;
  UpdatePanelScale;
end;

function TMapEditorForm.ZoomOut(MouseX: Integer = MinInt; MouseY: Integer = MinInt): Boolean;
var
  Q, P: TPoint;
begin
  Result := fScaleIndex > 0;
  if Result then
  begin
    if (MouseX > MinInt) and (MouseY > MinInt) then
      Q := Img.ControlToBitmap(Point(MouseX, MouseY));


    Dec(fScaleIndex);
    Img.Scale := AvailableScales[fScaleIndex];
    if Editor <> nil then
    begin
      if Editor.SelectorCount > 0 then
      begin
        P := Editor.SelectorCollection.Center;
//      else begin
//        P := Point(Img.Width div 2, Img.Height div 2);
  //      P.X := Img.Bitmap.Width div 2;
    //    P.X := Img.Bitmap.Height div 2;
        Img.ScrollToCenter(P.X, P.Y);
      end
      else if (MouseX > MinInt) and (MouseY > MinInt) then
      begin
        //P := Img.ControlToBitmap(Point(MouseX, MouseY));
        Img.ScrollToCenter(Q.X, Q.Y);
      end;
    end;
  end;
  UpdatePanelScale;
end;

procedure TMapEditorForm.SelectAllUnderRect(const aRect: TRect);
var
  i: Integer;
  T: TTerrain;
  O: TInteractiveObject;
  S: TSteel;
  R, Dummy: TRect;
begin
  with Editor do
  begin
    SelectorCollection.BeginUpdate;
    SelectorCollection.Clear;

    if litObject in fRenderer.RenderTypes then
    for i := 0 to InteractiveObjectCollection.Count - 1 do
    begin
      O := InteractiveObjectCollection[i];
      R.Left := O.ObjectX;
      R.Top := O.ObjectY;
      R.Right := R.Left + O.ItemWidth;
      R.Bottom := R.Top + O.ItemHeight;
{      if PtInRect(aRect, R.TopLeft)
      or PtInRect(aRect, R.BottomRight)
      or PtInRect(R, aRect.TopLeft)
      or PtInRect(R, aRect.BottomRight) then }
      if IntersectRect(Dummy, R, aRect) then
      begin
        with SelectorCollection.Add do
          RefItem := O;
//        Deb([t.terrainid, rectstr(r)]);
      end;
    end;

    if litTerrain in fRenderer.RenderTypes then
    for i := 0 to TerrainCollection.Count - 1 do
    begin
      T := TerrainCollection[i];
      R.Left := T.TerrainX;
      R.Top := T.TerrainY;
      R.Right := R.Left + T.ItemWidth;
      R.Bottom := R.Top + T.ItemHeight;
{      if PtInRect(aRect, R.TopLeft)
      or PtInRect(aRect, R.BottomRight)
      or PtInRect(R, aRect.TopLeft)
      or PtInRect(R, aRect.BottomRight) then}
      if IntersectRect(Dummy, R, aRect) then
      begin
        with SelectorCollection.Add do
          RefItem := T;
//        Deb([t.terrainid, rectstr(r)]);
      end;
    end;


    if litSteel in fRenderer.RenderTypes then
    for i := 0 to SteelCollection.Count - 1 do
    begin
      S := SteelCollection[i];
      R.Left := S.SteelX;
      R.Top := S.SteelY;
      R.Right := R.Left + S.ItemWidth;
      R.Bottom := R.Top + S.ItemHeight;
{      if PtInRect(aRect, R.TopLeft)
      or PtInRect(aRect, R.BottomRight)
      or PtInRect(R, aRect.TopLeft)
      or PtInRect(R, aRect.BottomRight) then }
      if IntersectRect(Dummy, R, aRect) then
      begin
        with SelectorCollection.Add do
          RefItem := S;
//        Deb([t.terrainid, rectstr(r)]);
      end;
    end;

    SelectorCollection.EndUpdate;


  end;
end;

function TMapEditorForm.GetSelectionDragBitmap(Bmp: TBitmap; var HotX, HotY: Integer): Boolean;
{-------------------------------------------------------------------------------
  This routine is and may only be called when dragging.
-------------------------------------------------------------------------------}
var
//  i: Integer;
  Lev: TLevelEditor;
  D, R: TRect;
 // First: Boolean;
  Pic: TBitmap32;
//  Sca: Integer;

begin
  Lev := Editor;
  Result := Assigned(Lev) and (Lev.SelectorCollection.Count > 0);
  R := EmptyRect;
//  First := True;

  R := LastHitItem.Bounds;

//  Dec(R.Right);
//  Dec(R.Bottom);//oops

//  Deb([rectwidth(R)]);

//  HotX := DragIP

  HotX := Round((DragBP.X - R.Left) * Img.Scale);
  HotY := Round((DragBP.Y - R.Top) * Img.Scale);

  //inc(hotx);
//  inc(hoty);
//  Deb([hotx,hoty]);

  D := ZeroTopLeftRect(R);
  D.Right := Round(D.Right * Img.Scale);
  D.Bottom := Round(D.Bottom * Img.Scale);

  Bmp.Width := RectWidth(D)-1;
  Bmp.Height := RectHeight(D)-1;


//  Deb([bmp.width]);

  case LastHitItem.ItemType of
    litTerrain:
      begin
        // PictureMgr.TerrainBitmaps[Editor.Statics.GraphicSet, LastHitItem.Identifier];
        Pic := Editor.Graph.TerrainBitmaps[LastHitItem.Identifier];
        Pic.DrawTo(Bmp.Canvas.Handle, ZeroTopLeftRect(D),ZeroTopLeftRect(R));
      end;
    litObject:
      begin
        //Pic := PictureMgr.ObjectBitmaps[Editor.Statics.GraphicSet, LastHitItem.Identifier];
        Pic := Editor.Graph.ObjectBitmaps[LastHitItem.Identifier];
        Pic.DrawTo(Bmp.Canvas.Handle, ZeroTopLeftRect(D),ZeroTopLeftRect(R));
      end;
    else
      Img.Bitmap.DrawTo(Bmp.Canvas.Handle, D, R);
  end;
//    Pic := PictureMgr.
end;

procedure TMapEditorForm.LMSelectionChanged(var Msg: TLMSelectionChanged);
begin
  StatusBar.Panels[panel_selection].Text := i2s(Editor.SelectorCollection.Count) + ' item(s) selected';
  UpdateSelectionLayer;
end;

procedure TMapEditorForm.LMLevelChanging(var Msg: TLMLevelChanging);
begin
  fLocked := True;
end;

procedure TMapEditorForm.LMLevelChanged(var Msg: TLMLevelChanged);
var
  Def: TDefaultsRec;
begin
  fLocked := False;
  ReRender;
  UpdateSelectionLayer;

  UpdateScreenCaption;
  UpdatePanelScale;
  UpdatePanelResolution;

  if (LCF_LOADFROMFILE and Msg.LevelChangeFlags <> 0)
  or (LCF_LOADFROMSTREAM and Msg.LevelChangeFlags <> 0)then
  begin
//    Deb(['aha']);

    if Editor.Style = nil then Exit;
    editor.style.GetStyleParams(def);
    Img.ScrollToCenter(Editor.Statics.ScreenPosition + Def.LevelHeight, 0);
//  Deb(['total update', editor.filename]);

//  deb(['we need to reevaluate scroll'])
  end;
end;

procedure TMapEditorForm.LMStaticsChanged(var Msg: TLMStaticsChanged);
begin
  if Msg.ChangeInfo = 0 then
    Img.Invalidate //???
  else
    ReRender;

end;

procedure TMapEditorForm.LMObjectListChanged(var Msg: TLMObjectListChanged);
begin
  ReRender;
//  fRenderer.RenderLevel(Editor);
end;

procedure TMapEditorForm.LMTerrainListChanged(var Msg: TLMTerrainListChanged);
begin
  ReRender;
//  fRenderer.RenderLevel(Editor);
end;

procedure TMapEditorForm.LMSteelListChanged(var Msg: TLMSteelListChanged);
begin
  ReRender;
//  fRenderer.RenderLevel(Editor);
end;

procedure TMapEditorForm.Img_StartDrag(Sender: TObject; var DragObject: TDragObject);
begin
  fImageIsDragging := True;
  Img.Update;
  DragObject := fDragger;
end;

procedure TMapEditorForm.Img_DragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept := True;
end;

procedure TMapEditorForm.Img_DragDrop(Sender, Source: TObject; X, Y: Integer);
var
  P, S: TPoint;
//  A, B: TPoint;

    function Snap(I: Integer): Integer;
//    var
  //    M: Integer;
    begin
//      M := I mod 8;
  //    Deb([m]);
      Result := I;// and not 3;
    end;

begin
//  if DragSelection then
  begin
    DragSelection := False;


    S := DragBP;
    P := Img.ControlToBitmap(Point(X, Y));

//    with p do
  //    deb(['dragdrop', x,y]);


//    A :=
    //if fDragger <> nil then
      //FreeAndNil(fDragger);

    Editor.MoveSelection(Snap(P.X - S.X), Snap(P.Y - S.Y));

  //  deb(['delta', p.X-s.x, p.Y-s.y])
//    deb

//    if Target is TDragObject then
  //    Target.Free;
  //  Exit;
  end;
end;

procedure TMapEditorForm.Img_EndDrag(Sender, Target: TObject; X, Y: Integer);
begin
  fImageIsDragging := False;
end;

procedure TMapEditorForm.Form_DropFiles(Sender: TObject; const Files: TStrings);
//var
  //Lev: TLevelEditor;
begin
  ShowMessage('Dragging files here not finished yet. Use File|Open to open files');
  exit;
  (*
{  with TAskStyleForm.Create(nil) do
  begin
    SHowModal;
    Free;
  end;     }
  if Files.Count <> 1 then
    Exit;
  Lev := Editor;
  if Lev = nil then
    Exit;
  Lev.LoadFromFile(Files[0], STYLE_DOSORIG{STYLE_WINLORES}, lffLVL{lffAutoDetect});

  ForceForegroundWindow(Handle);
  *)
end;

procedure TMapEditorForm.UpdateScreenCaption;
var
  Lev: TLevelEditor;
//  F: string;
//  T: string;
begin
  Lev := Editor;
  if Lev = nil then
    Caption := 'Lemmix'
  else
    Caption := 'Lemmix [' +
    MinimizeName(Lev.FileName, Self.Canvas, 256) + '] ' + Lev.LevelName;
//  else if Lev.LoaderObject is TFormEditDosLevelPack then
  //  Caption := 'Lemmix [' +
    //MinimizeName(FormEditDosLevelPack.FileName, Self.Canvas, 256) + '] ' +
//    i2s(FormEditDosLevelPack.tree.selectedindex) +
//    Lev.LevelName;
end;

procedure TMapEditorForm.UpdatePanelResolution;
begin
  if Editor = nil then
    Exit;
  StatusBar.Panels[panel_resolution].Text := i2s(Img.Bitmap.Width) + ' * ' + i2s(Img.Bitmap.Height);
end;

procedure TMapEditorForm.UpdatePanelScale;
begin
  StatusBar.Panels[panel_scale].Text := availableScaleNames[fScaleIndex];
end;

procedure TMapEditorForm.ActNewLevel_Execute(Sender: TObject);
var
  OK: Boolean;
//  S: Integer;
//  R: TResolution;
  S: TBaseLemmingStyle;
  G: TBaseGraphicSet;
begin
  if Editor = nil then
    Exit;
  OK := ExecuteNewLevelForm(S, G);
  if not OK then
    Exit;
  Editor.BeginUpdate;
  try
    Editor.CloseLevel;
//    Editor.GraphicSet := G;
//    Editor.
    Editor.Style := S;
    Editor.Graph := G;
    Editor.Statics.GraphicSet := G.GraphicSetId; //lemcore
    Editor.Statics.GraphicSetEx := G.GraphicSetIdExt; //lemcore
//    Editor.Resolution := R;
//    PictureMgr.Resolution := R;
//    fRenderer.Resolution := R;
  finally
    Editor.EndUpdate;
  end;
end;

procedure TMapEditorForm.ActNewDosLevelPack_Execute(Sender: TObject);
begin
  with FormEditDosLevelPack do
  begin
    ActNewLevelPack.Execute;
    Show;
  end;
{  ShowMessage('Not implemented yet. ' +
  'You can open (and modify) levelpacks but not create them yet.' +
  'Creation can be done by creating a new empty file (with notepad for example). After that open it with Lemmix');
  }
end;

procedure TMapEditorForm.ActNewDosExtGraphic_Execute(Sender: TObject);
begin
  ExecuteFormVgaSpecCreator;
end;

procedure TMapEditorForm.ActOpenFile_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  Action Execute eventhandler to open a level-file.
  History is saved.
-------------------------------------------------------------------------------}
var
  Path: string;
  Dlg: TOpenDialog;
  i: Integer;
  F: string;
  Ix: Integer;
  Exec: Boolean;
//  S: string;
begin

  if Editor = nil then
    Exit;
  Path := App.HistorySettings.EditorLastOpenDir;
  if Path = '' then
    Path := ApplicationPath;

  Ix := 0;
  Dlg := TOpenDialog.Create(nil);
  try
    Dlg.InitialDir := Path;

    for i := Low(OpenDlgFilters) to High(OpenDlgFilters) do
      Dlg.Filter := Dlg.Filter + OpenDlgFilters[i];
    Dlg.FilterIndex := App.HistorySettings.EditorLastOpenFilterIndex;
    Exec := Dlg.Execute;
    if Exec then
    begin
      App.HistorySettings.EditorLastOpenDir := ExtractFilePath(Dlg.FileName);
      App.HistorySettings.EditorLastOpenFilterIndex := Dlg.FilterIndex;
      F := Dlg.FileName;
      ix := Dlg.FilterIndex;

//      Editor.LoadFromFile(Dlg.FileName, FileNameToFormat(Dlg.FileName));

//    R := Rect(0, 0, Defaults[Resolution].ViewWidth, Defaults[Resolution].ViewHeight); //lemstartup
//    RectMove(R, ScreenPosition, 0);
//    R := MakeRect(fScreenStartLayer.GetAdjustedRect(FloatRect(R)));
//    Buffer.FrameRectS(R.Left, R.Top, R.Right, R.Bottom, clWhite32);
//
//      Img.ScrollToCenter(Editor.Statics.ScreenPosition + Defaults[Editor.Resolution].ViewWidth div 2, 0);
    end;
  finally
    Dlg.Free;
  end;

  //
  if Exec then
    InternalOpenFile(F, Ix);

end;

procedure TMapEditorForm.ActOpenRecentFile_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  This eventhandler is triggered by a TRecentFileAction. So we know the file
  and what file.
-------------------------------------------------------------------------------}
var
  Act: TRecentFileAction absolute Sender;
  Sty: TBaseLemmingStyle;
  Typ: TLemmingFileFormat;
  { TODO : checks }
begin

  Assert(Sender is TRecentFileAction, 'sender error');
  with Act do
  begin

    if comparetext(act.filetype,'level')=0 then
    begin
      Sty := StyleMgr.StyleByName(act.FileInfo);
      Typ := TLemmingFileFormat(GetEnumValue(typeinfo(TLemmingFileFormat), act.FileFormat));
      Editor.LoadFromFile(FileName, Sty, Typ);
    end;

  //    Editor.LoadFromFile(FileName, STYLE_LEMMINI, lffLemmini);
//    else if comparetext(act.filetype,'dosorig')=0 then
  //    InternalOpenFile(FileName, OFD_DOSLEMMINI)

  end;

end;


procedure TMapEditorForm.ActCloseLevel_Execute(Sender: TObject);
begin
  Assert(Editor <> nil, 'editor = nil');
  Editor.CloseLevel;
end;

procedure TMapEditorForm.ActQuit_Execute(Sender: TObject);
begin
  Close;
end;


procedure TMapEditorForm.ActSaveAs_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  Action Execute eventhandler to save a level-file.
  History is saved.
-------------------------------------------------------------------------------}
var
  Path, F: string;
  Dlg: TSaveDialog;
  i: Integer;
  TheFormat: TLemmingFileFormat;
begin
  if Editor = nil then
    Exit;
  Path := App.HistorySettings.EditorLastSaveDir;
  if Path = '' then
    Path := ApplicationPath;
  Dlg := TSaveDialog.Create(nil);
  try
    //Dlg.OnTypeChange := SaveDlg_TypeChange;
    for i := Low(SaveAsFilters) to High(SaveAsFilters) do
      Dlg.Filter := Dlg.Filter + SaveAsFilters[i];
    Dlg.InitialDir := Path;

    case Editor.CurrentFormat of
      lffLVL:
        Dlg.FilterIndex := 2;
      lffLemmini:
        Dlg.FilterIndex := 1;
    else
      Dlg.FilterIndex := App.HistorySettings.EditorLastSaveFilterIndex;
    end;

    if Editor.FileName <> '' then
      Dlg.FileName := Editor.FileName;
    if Dlg.Execute then
    begin
      F := Dlg.FileName;
      case Dlg.FilterIndex of
        1: TheFormat := lffLemmini;
        2: TheFormat := lffLVL;
      else
        raise Exception.Create('SaveAs: file format error');
      end;
      F := Editor.GetAdjustedFileName(F, TheFormat);
      App.HistorySettings.EditorLastSaveDir := ExtractFilePath(F);
      App.HistorySettings.EditorLastSaveFilterIndex := Dlg.FilterIndex;
      Editor.SaveToFile(F, TheFormat);
    end;
  finally
    Dlg.Free;
  end;

  //TCustomLevelValidator.ValidateLevel(Editor);

end;



procedure TMapEditorForm.ActOptions_Execute(Sender: TObject);
begin
  ExecuteFormOptions;
end;


procedure TMapEditorForm.ActAlignSelection_Execute(Sender: TObject);
var
  Horz: TItemHorzAlign;
  Vert: TItemVertAlign;
begin
  if Editor = nil then
    Exit;
  if ExecuteAlignSelectionForm(Horz, Vert) then
    Editor.AlignSelection(Horz, Vert);
end;

procedure TMapEditorForm.ActDosAlignSelection_Execute(Sender: TObject);
begin
  if Editor = nil then
    Exit;
  Editor.DosAlignSelection;
end;


function TMapEditorForm.CalcCenteredInsertionPoint: TPoint;
{-------------------------------------------------------------------------------
  Calculate the "for the eye" best point to insert a new object into the map.
  It returns a bitmap coordinate (not an image coordinate!)
  If the center of the image is not a bitmap point then the scale is most likely
  very small and we take the center of the bitmap as result.
-------------------------------------------------------------------------------}
begin
  Result := Point(Img.Width div 2, Img.Height div 2);
  Result := Img.ControlToBitmap(Result);
  if not PtInRect(Img.Bitmap.BoundsRect, Result) then
    Result := Point(Img.Bitmap.Width div 2, Img.Bitmap.Height div 2);
end;

procedure TMapEditorForm.ActInsertObject_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  Action event handler.
  There are two actions that are attached to this eventhandler:
  1) ActInsertObject from the main menu
  2) ActInsertObject from the local menu
  In the case of the local menu we take the popup point as location
-------------------------------------------------------------------------------}
var
  P: TPoint;
//  O: TInteractiveObject;
begin
  if Sender = ActInsertObjectLocal then
  begin
    // convert global screen coordinate
    P := MapPopup.PopupPoint; // this is a global screen coordinate
    P := Img.ScreenToClient(P);
    P := Img.ControlToBitmap(P);
  end
  else
    P := CalcCenteredInsertionPoint;

  InternalInsertObject(P);
end;

procedure TMapEditorForm.ActInsertTerrain_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  See also: ActInsertObject_Execute
-------------------------------------------------------------------------------}
var
  P: TPoint;
//  O: TInteractiveObject;
begin
  if Sender = ActInsertTerrainLocal then
  begin
    // convert global screen coordinate
    P := MapPopup.PopupPoint; // this is a global screen coordinate
    P := Img.ScreenToClient(P);
    P := Img.ControlToBitmap(P);
  end
  else
    P := CalcCenteredInsertionPoint;

  InternalInsertTerrain(P);
end;

procedure TMapEditorForm.ActInsertSteel_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  See also: ActInsertObject_Execute
-------------------------------------------------------------------------------}
var
  P: TPoint;
begin
  if Sender = ActInsertSteelLocal then
  begin
    // convert global screen coordinate
    P := MapPopup.PopupPoint; // this is a global screen coordinate
    P := Img.ScreenToClient(P);
    P := Img.ControlToBitmap(P);
  end
  else
    P := CalcCenteredInsertionPoint;

  InternalInsertSteel(P);
end;
(*
var
  BP: TPoint;
  S: TSteel;
begin
  if Editor = nil then
    Exit;
  Editor.BeginUpdate;
  try
    BP := CalcCenteredInsertionPoint;
    S := Editor.SteelCollection.Add;
    S.SteelX := BP.X;
    S.SteelY := BP.Y;
    S.SteelWidth := 40;
    S.SteelHeight := 40;
    Editor.SelectedSteel := S;
  finally
    Editor.EndUpdate;
  end;
end;*)

procedure TMapEditorForm.ActCopy_Execute(Sender: TObject);
begin
  if Editor = nil then
    Exit;
  Editor.CopyToClipBoard;
end;

procedure TMapEditorForm.ActCut_Execute(Sender: TObject);
begin
  if Editor = nil then
    Exit;
  Editor.CopyToClipBoard;
  Editor.DeleteSelection;
end;

procedure TMapEditorForm.ActPaste_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  Action eventhandler. If
-------------------------------------------------------------------------------}
begin
  if Editor = nil then
    Exit;
(*  with Editor do
    if SelectorCount > 0 then
      Editor.PasteFromClipBoard(4, 4)
    else                          *)
      Editor.PasteFromClipBoard;//(fSelectionRect.Left, fSelectionRect.Top);
end;

procedure TMapEditorForm.ActPasteSpecial_Execute(Sender: TObject);
var
  H, V, C: Integer;
begin
  if Editor = nil then
    Exit;
  if ExecutePasteSpecialForm(H, V, C) then
  begin
    Editor.PasteFromClipBoard(H, V, C);
  end;
end;


procedure TMapEditorForm.ActShowInspector_Execute(Sender: TObject);
begin
  fEditItemForm.Show;
end;

procedure TMapEditorForm.ActShowList_Execute(Sender: TObject);
begin
  fListForm.Show;
end;

procedure TMapEditorForm.ActShowStatics_Execute(Sender: TObject);
begin
  fStaticsForm.Show;
end;


procedure TMapEditorForm.ActShowImage_Execute(Sender: TObject);
begin
  if Img.CanFocus then
    Img.SetFocus;
end;

procedure TMapEditorForm.HandleCommandStr;
var
  B: TBaseLevelItem;
  I: Integer;
begin
  if Length(fCommandStr) < 1 then
    Exit;
  if Editor = nil then
    Exit;
  if Editor.SelectorCount <> 1 then
    Exit;

   I := StrToIntDef(fCommandStr, -1);
   if I = -1 then
     Exit;
   B := Editor.SelectorCollection[0].RefItem;
   case B.ItemType of
     litObject:
       begin
         if TInteractiveObject(B).ValidObjectID(s2i(fCommandStr)) then
           TInteractiveObject(B).ObjectID := s2i(fCommandStr);
       end;
     litTerrain:
       begin
         if TTerrain(B).ValidTerrainID(s2i(fCommandStr)) then
           TTerrain(B).TerrainID := s2i(fCommandStr);
       end;
   end;

//   CommandStr := '';

end;

procedure TMapEditorForm.SetCommandStr(const Value: string);
begin
  fCommandStr := Value;
  StatusBar.Panels[panel_commandstr].Text := fCommandStr;
  fCommandChangeTicks := GetTickCount;
end;


procedure TMapEditorForm.DoAfterSetEditor(Value: TLevelEditor);
begin
//  fInspectorForm.Editor := Value;
  fStaticsForm.Editor := Value;
  fListForm.Editor := Value;
  fEditItemForm.Editor := Value;
end;

procedure TMapEditorForm.FormContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
begin
//  inherited;
//  Deb(['context', sender.classname]);
  handled := true;
end;

procedure TMapEditorForm.ActToggleTerrain_Execute(Sender: TObject);
begin
  with fRenderer do
  begin
    if litTerrain in RenderTypes
    then RenderTypes := RenderTypes - [litTerrain]
    else RenderTypes := RenderTypes + [litTerrain];
    ActToggleTerrain.Checked := litTerrain in RenderTypes;
    ReRender;
  //  fRenderer.RenderLevel(Editor);
  end;
end;

procedure TMapEditorForm.ActToggleObjects_Execute(Sender: TObject);
begin
  with fRenderer do
  begin
    if litObject in RenderTypes
    then RenderTypes := RenderTypes - [litObject]
    else RenderTypes := RenderTypes + [litObject];
    ActToggleObjects.Checked := litObject in RenderTypes;
    ReRender;
   //  fRenderer.RenderLevel(Editor);
  end;
end;

procedure TMapEditorForm.ActToggleSteel_Execute(Sender: TObject);
begin
  with fRenderer do
  begin
    if litSteel in RenderTypes
    then RenderTypes := RenderTypes - [litSteel]
    else RenderTypes := RenderTypes + [litSteel];
    ActToggleSteel.Checked := litSteel in RenderTypes;
    ReRender;
  //  fRenderer.RenderLevel(Editor);
  end;
end;

procedure TMapEditorForm.ActToggleSteelBack_Execute(Sender: TObject);
begin
  with fRenderer do
  begin
    if roSteelNoOverwrite in RenderOptions
    then RenderOptions := RenderOptions - [roSteelNoOverwrite]
    else RenderOptions := RenderOptions  + [roSteelNoOverwrite];
    ActToggleSteelBack.Checked := roSteelNoOverwrite in RenderOptions;
    ReRender;
//    fRenderer.RenderLevel(Editor);
  end;
end;

procedure TMapEditorForm.ActToggleTriggerAreas_Execute(Sender: TObject);
begin
  with fRenderer do
  begin
    if roShowTriggerAreas in RenderOptions
    then RenderOptions := RenderOptions - [roShowTriggerAreas]
    else RenderOptions := RenderOptions  + [roShowTriggerAreas];
    ActToggleTriggerAreas.Checked := roShowTriggerAreas in RenderOptions;
    ReRender;
//    fRenderer.RenderLevel(Editor);
  end;
end;

procedure TMapEditorForm.ActToggleScreenStart_Execute(Sender: TObject);
begin
  fScreenStartLayer.Visible := not fScreenStartLayer.Visible;
  ActToggleScreenStart.Checked := fScreenStartLayer.Visible;
end;

procedure TMapEditorForm.ActViewLevels_Execute(Sender: TObject);
{-------------------------------------------------------------------------------
  Only valid if tag is set.
-------------------------------------------------------------------------------}
var
  A: TAction absolute Sender;
begin
  Assert(Sender is TAction, 'Sender error');
  if A.Tag >= 100 then
    if A.Tag <= 300 then
      FormViewLevels(StyleMgr[a.Tag - 100]).Show;
end;

procedure TMapEditorForm.ActShowDosLevelPackEditor_Execute(Sender: TObject);
begin
  FormEditDosLevelPack.Show;
end;


procedure TMapEditorForm.ActZoomIn_Execute(Sender: TObject);
begin
  ZoomIn;
end;

procedure TMapEditorForm.ActZoomOut_Execute(Sender: TObject);
begin
  ZoomOut;
end;


procedure TMapEditorForm.ActStyleCompiler_Execute(Sender: TObject);
begin
  with TFormCompileStyle.create(nil) do
  begin
    ShowModal;
    Free;
  end;
end;


procedure TMapEditorForm.ActValidateLevel_Execute(Sender: TObject);
var
  L: TStringList;
  OK: Boolean;
begin
  if Editor = nil then
    Exit;
  L := TStringList.Create;
  try
    OK := TCustomLevelValidator.ValidateLevel(Editor, L);
    if OK then
      L.Insert(0, 'Level OK')
    else
      L.Insert(0, 'Errors found');
    ExecuteFormValidationReport(L);
  finally
    L.Free;
  end;
end;

procedure TMapEditorForm.ActDosDatDecompress_Execute(Sender: TObject);
begin
  ExecuteFormDosDecompress;
end;

procedure TMapEditorForm.ActPlay_Execute(Sender: TObject);
begin
  InternalPlayLevel(Editor);
end;

procedure TMapEditorForm.ActionList_Update(Action: TBasicAction; var Handled: Boolean);
var
  Edit, {LevelOpen, }Select, MultiSelect, Clip: Boolean;
  n: Integer;
begin
  Edit         := Assigned(Editor) and Assigned(Editor.Style) and Assigned(Editor.Graph);
  Select       := Edit and (Editor.SelectorCollection.HackedList.Count > 0);
  MultiSelect  := Select and (Editor.SelectorCollection.HackedList.Count > 1);
  Clip         := ClipBoard.HasFormat(CF_TEXT);

  ActNewLevel.Enabled := Assigned(Editor);
//  ActOpenFile.Enabled             := Edit;
  ActCloseLevel.Enabled := Edit;
  ActSaveAs.Enabled := Edit;
//  ActSaveInDosPack.Enabled := Edit and (Editor.Loader is TFormEditDosLevelPack) and
  //                           FormEditDosLevelPackActive; // for safety

  ActInsertObject.Enabled := Edit and (Editor.Graph.MetaObjectCollection.HackedList.Count > 0);
  ActInsertObjectLocal.Enabled := ActInsertObject.Enabled;

  ActInsertTerrain.Enabled := Edit and (Editor.Graph.MetaTerrainCollection.HackedList.Count > 0);
  ActInsertTerrainLocal.Enabled := ActInsertTerrain.Enabled;

  ActInsertSteel.Enabled := Edit;
  ActInsertSteelLocal.Enabled := ActInsertSteel.Enabled;

  ActCopy.Enabled := Edit and Select;
  ActCut.Enabled := Edit and Select;
  ActPaste.Enabled := Edit and Clip;
  ActPasteSpecial.Enabled := Edit and Clip;

  ActAlignSelection.Enabled := Edit and MultiSelect;
  ActDosAlignSelection.Enabled := Edit and Select;

  if Action.Tag >= 100 then
    if Action.Tag <= 300 then
    begin
      n := Action.Tag - 100;
      TAction(Action).Enabled := Assigned(StyleMgr[n].LevelLoader);
    end;

  Handled := True;
end;



procedure TMapEditorForm.ActAbout_Execute(Sender: TObject);
begin
  ShowAboutForm;
  //MessageDlg(SAboutStr, mtInformation, [mbOK], 0);
end;

{$ifdef develop}

procedure TMapEditorForm.ActTest_Execute(Sender: TObject);
begin
{  with TTestForm.Create(nil) do
  begin
    ShowModal;
    Free;
  end;
}
end;

procedure TMapEditorForm.ActShowSpriteFiles_Execute(Sender: TObject);
begin
  {
  with TFormShowSPR.Create(nil) do
  begin
    ShowModal;
    Free;
  end;
  }
end;

procedure TMapEditorForm.ActShowDOSGroundFile_Execute(Sender: TObject);
begin

  with TDOSGroundFileForm.Create(nil) do
  begin
    ShowModal;
    Free;
  end;
  
end;

procedure TMapEditorForm.ActPackExtGraphics_Execute(Sender: TObject);
begin
//  PackExtendedGraphics;
end;

procedure TMapEditorForm.ActWizardLemmini_Execute(Sender: TObject);
begin
  with TWizardLemminiCompiler.Create(nil) do
  begin
    ShowModal;
    Free;
  end;
end;

{$endif}

procedure TMapEditorForm.InternalPlayLevel(aLevel: TLevel);
var
  Mem: TMemoryStream;
//  S: string;
begin
  if not LevelActive then
    Exit;

  // try Lemmini
  if aLevel.Style is TLemminiStyle then
  begin
    Mem := TMemoryStream.Create;
    try
      Clipboard.AsText := Editor.SaveToLemminiString;
    finally
      Mem.Free;
    end;
    Exit;
  end;

  // try internal dos player
  if not (aLevel.Style is TBaseDosStyle) then
  begin
    ShowMessage('Currentlevel is not a DOS level');
    Exit;
  end;

  with TFormGame.Create(nil) do
  begin
    aLevel.LockLevel;
    LoadAndPlay(aLevel);
    Free;
    aLevel.UnLockLevel;
  end;
end;

procedure TMapEditorForm.InternalInsertObject(P: TPoint);
var
  O: TInteractiveObject;
begin
  if Editor = nil then
    Exit;
  Editor.BeginUpdate;
  try
    O := Editor.InteractiveObjectCollection.Add;
    O.ObjectX := P.X;
    O.ObjectY := P.Y;
    Editor.SelectedObject := O;
  finally
    Editor.EndUpdate;
  end;
end;

procedure TMapEditorForm.InternalInsertTerrain(P: TPoint);
var
  T: TTerrain;
begin
  if Editor = nil then
    Exit;
  Editor.BeginUpdate;
  try
    T := Editor.TerrainCollection.Add;
    T.TerrainX := P.X;
    T.TerrainY := P.Y;
    Editor.SelectedTerrain := T;
  finally
    Editor.EndUpdate;
  end;
end;

procedure TMapEditorForm.InternalInsertSteel(P: TPoint);
var
  S: TSteel;
begin
  if Editor = nil then
    Exit;
  Editor.BeginUpdate;
  try
    S := Editor.SteelCollection.Add;
    S.SteelX := P.X;
    S.SteelY := P.Y;
    S.SteelWidth := 40;
    S.SteelHeight := 40;
    Editor.SelectedSteel := S;
  finally
    Editor.EndUpdate;
  end;
end;


procedure TMapEditorForm.InternalOpenFile(const aFileName: string; aType: Integer);
{-------------------------------------------------------------------------------
  Key routine to open a file. lemstrings
  Use the OFD_XXX consts as aType parameter
-------------------------------------------------------------------------------}
var
  aStyle: TBaseLemmingStyle;
begin
  aStyle := nil; // important
  case aType of
    OFD_DOSLEVELPACK://OFD_DOSLEVELXXX, OFD_UNKNOWNDAT:
      begin


        FormEditDosLevelPack.FileName := aFileName;
        FormEditDosLevelPack.Show;

        {

        with TViewDosLevelPackForm.Create(nil) do
        begin
          FileName := aFileName;
          ShowModal;
          Free;
        end;

        }

      end;
    OFD_RAWLEVEL:
      begin
        if Dlg_AskStyle(aStyle) = mrOK then
        begin
//          if aStyle is TBaseWinStyle then
  //          ShowMessage('Be patient, this style is slow!');
          Editor.LoadFromFile(aFileName, aStyle, lffLVL);
          Editor.Loader := Self;
//          GlobalSaver := Self;
          //LoaderObject := Self;
        end;
      end;
    OFD_LEMMINI:
      begin
        Editor.LoadFromFile(aFileName, LemminiStyle, lffLemmini);
        Editor.Loader := Self;
//        GlobalSaver := Self;
        //LoaderObject := Self;
      end;
    OFD_VGASPEC:
      begin
        ExecuteFormVgaSpecViewer(aFileName);
      end;
  end;


end;

procedure TMapEditorForm.Rerender;
begin
  fRenderer.RenderLevel(Editor)
end;

function TMapEditorForm.CheckCanDrag: Boolean;
var
  Def: TDefaultsRec;
begin
  Result := True;
  exit;
  Result := False;
  if Editor = nil then exit;
  if editor.graph =nil then exit;
  editor.Style.GetStyleParams(def);
  if def.Resolution = 1.0 then
    Result := fScaleIndex <= 5
  else result := fScaleIndex <= 4;
end;

procedure TMapEditorForm.Button2Click(Sender: TObject);
begin
end;

{ txlist }

procedure TMapEditorForm.Button4Click(Sender: TObject);
begin
//  formviewlevels(StyleMgr[style_dosohno]).show;
  DosOrigStyle.AnimationSet.ReadData;

  ShowImageForm(dosorigstyle.AnimationSet.AnimationBitmaps[0]);
//  DosHoliday93Style.AnimationSet.ReadData;
end;

end.
(*
//actnlst

function TMapEditorForm.GetSelectionDragBitmap(Bmp: TBitmap; var HotX, HotY: Integer): Boolean;
var
  i: Integer;
  Lev: TLevelEditor;
  D, R: TRect;
  First: Boolean;
begin
  Lev := Editor;
  Result := Assigned(Lev) and (Lev.SelectorCollection.Count > 0);
  R := EmptyRect;
  First := True;

//  if LastHitItem <> nil then
  //  Deb([lasthititem.Index]);


  with Lev.SelectorCollection do
    for i := 0 to Count - 1 do
    begin
      with Items[i].RefItem do
      begin
        if First then
        begin
          First := False;
          R := Bounds;
        end
        else begin
          R.Left := Min(R.Left, XPosition);
          R.Top := Min(R.Top, YPosition);
          R.Right := Max(R.Right, XPosition + ItemWidth);
          R.Bottom := Max(R.Bottom, YPosition + ItemHeight);
        end;
      end;
    end;


  HotX := Round((DragP.X - R.Left) * Img.Scale);
  HotY := Round((DragP.Y - R.Top) * Img.Scale);

//  Deb([hotx, hoty]);

  D := ZeroTopLeftRect(R);
  D.Right := Round(D.Right * Img.Scale);
  D.Bottom := Round(D.Bottom * Img.Scale);

  Bmp.Width := RectWidth(D);
  Bmp.Height := RectHeight(D);

  Img.Bitmap.DrawTo(Bmp.Canvas.Handle, D, R);



  {
  if Bmp.Width > 200 then
    Bmp.Width := 200;
  if Bmp.Height > 200 then
    Bmp.Height := 200;
   }
//  Img.Bitmap.Fi
(*  with Bmp.Canvas do
  begin
    brush.Color := clBlue;
    fillrect(d);
  end; *)

//  Bmp.Canvas.Draw(Img.Bitmap, )    lemstrings
end;


procedure ShowLevelEditor;
var
  E: TLevelEditor;
  F: TMapEditorForm;
  O: TLemmingObjectInspector;
  List: TLevelListEditorForm;
  St: TEditStaticsForm;
  I: TEditItemForm;
begin
  E := TLevelEditor.Create(Application);
  //L := TLevelEditorLink.Create(E.Mgr, Self);
  F := TMapEditorForm.Create(Application);
  F.Editor := E;

  F.ShowModal;

  (*
  O := TLemmingObjectInspector.Create(application);
  O.Left := 0;
  O.Show;
  O.Editor := E;

  St := TEditStaticsForm.Create(Application);
  St.Left := 0;
  St.Show;
  St.Editor := E;

  List := TLevelListEditorForm.Create(Application);
  List.Left := 0;
  List.Show;
  List.LevelEditor := E;

  f.fInspectorForm := o;
  f.fStaticsForm := st;
  f.fListForm := list; *)


{  I := TEditItemForm.Create(Application);
  I.Show;
  I.Editor := E;}


(*  E.LoadFromFile('LVL0091.LVL'); *)
//  E.LoadFromFile('d:\lemmings downloads\lemmini\level\orig\LVL0092.LVL', lffLVL);
(*  E.LoadFromFile('d:\lemmings downloads\lemmini\level\orig\LVL0094.LVL'); *)


//  E.LoadFromFile('lvl1016.ini', lffLemmini);

//  f.ShowModal; ustartup

end;
//lemedit sysutils udebug
//umx
//ibevents
//dialogs

// special opendialogs
//open Dos Original compressed levelpack *.dat
//     Dos OhNo compressed levelpack

//lemapp



(*

  style: name



*)

//create vgaspec
{
 select bitmap
 check 960x160
 check count colors
 select filename
 check overwrite
 create file
 check size (maybe too big)
 write it




}

(*

MinimizeName function

hortens a fully qualified path name so that it can be drawn with a specified length limit.

ProcessPath procedure

arses a file name into its constituent parts.

UnixPathToDosPath function

Converts a Unix-compatible path specification to a Dos-compatible path specification.


[20:26:05] gb zegt: file name utilities

*)

                         //lemstartup


//udebug lemstrings



(*


*)


procedure TMapEditorForm.Button2Click(Sender: TObject);
var
  R: TReplayItems;
    procedure add(it: integer; rt: trecordedtype;
      askill:TBasicLemmingAction;lem:integer);
    begin
      with r.add do
      begin
        iteration:=it;
        rectyp:=rt;//assignskill;
        skill:=askill;
        lemmingindex:=lem;
      end;
    end;

    procedure rr(it: integer; rt: trecordedtype;
      rr:integer);
    begin
      with r.add do
      begin
        iteration:=it;
        rectyp:=rt;//assignskill;
        releaserate:=rr;
      end;
    end;
    

begin
  R:=TReplayItems.create;
  try
    add(58, rtAssignSkill, baClimbing, 1);
    add(63, rtAssignSkill, baClimbing, 2);
    add(67, rtAssignSkill, baClimbing, 3);
    add(71, rtAssignSkill, baClimbing, 4);
    add(75, rtAssignSkill, baClimbing, 5);
    add(78, rtAssignSkill, baClimbing, 6);
    add(83, rtAssignSkill, baClimbing, 7);
    add(87, rtAssignSkill, baClimbing, 8);
    add(94, rtAssignSkill, baFloating, 10);
    add(98, rtAssignSkill, baFloating, 11);
    add(103, rtAssignSkill, baClimbing, 10);
    add(214, rtAssignSkill, baBashing, 9);
    add(222, rtAssignSkill, baMining, 0);
    add(239, rtAssignSkill, baFloating, 1);
    add(241, rtAssignSkill, baFloating, 2);
    add(246, rtAssignSkill, baFloating, 3);
    add(251, rtAssignSkill, baFloating, 4);
    add(253, rtAssignSkill, baFloating, 5);
    add(259, rtAssignSkill, baFloating, 6);
    add(261, rtAssignSkill, baFloating, 7);
    add(267, rtAssignSkill, baFloating, 8);
    add(416, rtAssignSkill, baBuilding, 41);
    add(514, rtAssignSkill, baDigging, 8);
    add(549, rtAssignSkill, baBashing, 8);

    r.savetofile('from the boundary line.bin');
    r.savetotxt('from the boundary line.txt');

  finally
  r.free;
  end;

  R:=TReplayItems.create;
  try

    rr(19, rtStartIncreaseRR, 1);
    rr(19, rtStopChangingRR, 85);
    rr(19, rtStartIncreaseRR, 85);
    rr(19, rtStopChangingRR, 86);
    add(55, rtAssignSkill, baClimbing, 0);
    rr(58, rtStartDecreaseRR, 86);
    rr(58, rtStopChangingRR, 1);
    add(66, rtAssignSkill, baClimbing, 1);
    add(77, rtAssignSkill, baExplosion, 0);
    add(85, rtAssignSkill, baFloating, 0);
    add(96, rtAssignSkill, baFloating, 1);
    add(118, rtAssignSkill, baFloating, 2);
    add(184, rtAssignSkill, baBuilding, 1);
    add(202, rtAssignSkill, baBuilding, 1);
    add(219, rtAssignSkill, baBashing, 1);
    add(226, rtAssignSkill, baBuilding, 1);
    add(243, rtAssignSkill, baBashing, 1);
    add(249, rtAssignSkill, baBuilding, 1);
    rr(412, rtStartIncreaseRR, 1);
    rr(412, rtStopChangingRR, 99);
    add(518, rtAssignSkill, baBlocking, 1);

    r.savetofile('pillars.bin');
    r.savetotxt('pillars.txt');

  finally
  r.free;
  end;



end;

  (*

//  if actplay.enabled then exit;
  if not (key in ['c'..'x']) then exit;
  fcc := fcc + key;
  if length(fcc) > 10 then
    fcc := '';
  if length(fcc) = 9 then
  begin
    if decryptstr(fcc) = '[דn' then  // ccexplore encrypted
    begin
      actplay.Enabled := true;
      keypreview := false;
      onkeypress:=nil;
    end;
    fcc := '';
  end;

//  umisc
  //'ccexplore'
  *)

procedure TMapEditorForm.SaveDlg_TypeChange(Sender: TObject);
{-------------------------------------------------------------------------------
  Temporary eventhandler when saving a file.
-------------------------------------------------------------------------------}
var
  Dlg: TSaveDialog absolute Sender;
  S: string;
begin
  if not (Sender is TSaveDialog) then
    Exit;
  S := Dlg.FileName;
//  Deb([s]);
  if S <> '' then
  begin
    case Dlg.FilterIndex of
      2:
        begin

//          S := ChangeFileExt(S, '.ini');
  //        Deb([s]);
    //      Dlg.FileName := s;//:= ChangeFileExt(Dlg.FileName, '.ini');
        end;
      3:
        begin
          Dlg.FileName := ChangeFileExt(Dlg.FileName, '.lvl');
        end;
    end;
//    S := ReplaceFileExt()
//    S := ChangeFileExt(S, )
//    Deb([dlg.filterindex]);

//  SaveAsFilters : array[1..3] of string = (
//    'All files (*.*)|*.*|',
//    'Lemmini file (*.ini)|*.ini|',
//    'WinLemmings file (*.lvl)|*.lvl'
//  );

  end;
end;



procedure TMapEditorForm.ActSaveInDosPack_Execute(Sender: TObject);
begin
  if Editor = nil then
    Exit;
  if Assigned(Editor.Style) and Assigned(Editor.Graph) then
  begin
    if Editor.Loader is TFormEditDosLevelPack then
    begin
      FormEditDosLevelPack.SaveInPack;

    end
    else
      showmessage('boooh');

  end;
end;

(*
procedure TMapEditorForm.Test;
var
  iCount:       Integer;
  iSize:        Integer;
  iProp:        Integer;
  pProps:       PPropList;
//  iValue:       Integer;
//  fValue:       Extended;
//  sValue:       String; umisc
aClass: TClass;
begin
   aClass := TSteel;
    // Vraag het aantal properties op en reken het benodigde geheugen uit...
    iCount  := GetPropList(aClass.ClassInfo, tkProperties, nil);
    iSize   := iCount * SizeOf(TPropInfo);

    // Reserveer het geheugen...
    GetMem(pProps, iSize);

    try
      // Vraag de properties lijst op...
      GetPropList(aClass.ClassInfo, tkProperties, pProps);
      for iProp := 0 to iCount - 1 do
      begin
        //Deb([pProps^[iProp]^.Name]);
        //aMethod(aClass, pProps^[iProp]^.Name, pProps^[iProp]^.PropType^.Kind);
      end;
    finally
      FreeMem(pProps, iSize);
    end;
end;

