{$include lem_directives.inc}
unit FNewLevel;

{-------------------------------------------------------------------------------
  This unit contains a dialog for selecting a style/graphicset for a new level.
-------------------------------------------------------------------------------}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls,
  LemTypes,
  LemCore,
  LemStrings,
  Kernel_Resource,
  UMasterControls, Buttons, ExtCtrls, ComCtrls, ActnList, USplit,
  VirtualTrees;

type
  TMyData = class
  private
    NodeType    : (ntNone, ntStyle, ntGraph);
    fGraph      : TBaseGraphicSet;
    fStyle      : TBaseLemmingStyle;
  public
    property Style: TBaseLemmingStyle read fStyle;
    property Graph: TBaseGraphicSet read fGraph;
  end;

  TNewLevelForm = class(TForm)
    ActionList: TActionList;
    ActOK: TAction;
    MainPanel: TSplitPanel;
    BitBtn2: TBitBtn;
    BitBtn1: TBitBtn;
    VTree: TVirtualStringTree;
    procedure Form_Create(Sender: TObject);
    procedure ActOK_Execute(Sender: TObject);
    procedure ActOKUpdate(Sender: TObject);
    procedure Form_CloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure VTree_GetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType;
      var CellText: WideString);
    procedure VTree_FreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure VTree_FocusChanged(Sender: TBaseVirtualTree;
      Node: PVirtualNode; Column: TColumnIndex);
  private
    SelNode      : PVirtualNode;
    SelectedStyle: TBaseLemmingStyle;
    SelectedGraph: TBaseGraphicSet;
    { Private declarations }
    function GetTreeViewResult(var aStyle: TBaseLemmingStyle; var aGraph: TBaseGraphicSet): Boolean;
    function GetMyNodeData(aNode: PVirtualNode): TMyData;
  public
    { Public declarations }
  end;

function ExecuteNewLevelForm(var aStyle: TBaseLemmingStyle; var aGraph: TBaseGraphicSet): Boolean;

implementation

{$R *.dfm}

type
  PDataRec = ^TDataRec;
  TDataRec = record
    Obj: TObject;
  end;


function ExecuteNewLevelForm(var aStyle: TBaseLemmingStyle; var aGraph: TBaseGraphicSet): Boolean;
{-------------------------------------------------------------------------------
  Execute function for the form.
-------------------------------------------------------------------------------}
var
  F: TNewLevelForm;
begin
  aStyle := nil;
  aGraph := nil;

  F := TNewLevelForm.Create(nil);
  try
    F.Position := poDesktopCenter;
    Result := F.ShowModal = mrOK;
    if Result then
      F.GetTreeViewResult(aStyle, aGraph)
  finally
    F.Free;
  end;
end;

procedure TNewLevelForm.Form_Create(Sender: TObject);
(*
{-------------------------------------------------------------------------------
  Fill the Treeview with all the descriptions of styles and their graphicsets.
  The objects are added to the nodes.
-------------------------------------------------------------------------------}
var
  i, j: Integer;
  S: TBaseLemmingStyle;
  G: TBaseGraphicSet;
  It: TTreeNode;
begin
  Tree.Items.BeginUpdate;
  try
    with StyleMgr do
      for i := 0 to StyleList.Count - 1 do
      begin
        S := StyleList[i];
        It := Tree.Items.AddChildObject(nil, S.StyleDescription, S);
        for j := 0 to S.GraphicSetList.Count - 1 do
        begin
          G := S.GraphicSetList[j];
          Tree.Items.AddChildObject(It, G.GraphicSetName, G);
        end;
      end;
    //Tree.FullExpand;
    if Tree.Items.Count > 0 then
      Tree.Selected := Tree.Items[0];
  finally
    Tree.Items.EndUpdate;
  end;
  *)

{-------------------------------------------------------------------------------
  Fill the Treeview with all the descriptions of styles and their graphicsets.
  The objects are added to the nodes.
-------------------------------------------------------------------------------}
var
  i, j: Integer;
  S: TBaseLemmingStyle;
  G: TBaseGraphicSet;
//  It: TTreeNode;
  //CompileCount: Integer;

  SData, GData: TMyData;
  SNode, GNode: PVirtualNode;
begin

  VTree.Images := MxImages10;
  VTree.NodeDataSize := SizeOf(TDataRec);

//  VTree.TreeOptions.MiscOptions := VTree.TreeOptions.MiscOptions + [toCheckSupport];

  with StyleMgr do
    for i := 0 to StyleList.Count - 1 do
    begin
      S := StyleList[i];
      SData := TMyData.Create;
      SData.fStyle := S;
      SData.NodeType := ntStyle;

      //CompileCount := 0;

      SNode := VTree.AddChild(nil, SData);
      for j := 0 to S.GraphicSetList.Count - 1 do
      begin
        G := S.GraphicSetList[j];
        GData := TMyData.Create;
        GData.fGraph := G;
        GData.NodeType := ntGraph;

        GNode := VTree.AddChild(SNode, GData);
        VTree.CheckType[GNode] := ctCheckBox;

        if G.GraphicSetArchive <> '' then
          if FileExists(G.FullArchiveName) then
          begin
//            GData.fIsCompiled := 1;
  //          Inc(CompileCount);
          end;
//        VTree.CheckState[GNode] := csCheckedNormal;


        //Tree.Items.AddChildObject(It, G.GraphicSetName, G);
      end;

//      if CompileCount = S.GraphicSetList.Count then
  //      SData.fIsCompiled := 1;

    end;






end;

function TNewLevelForm.GetTreeViewResult(var aStyle: TBaseLemmingStyle; var aGraph: TBaseGraphicSet): Boolean;
{-------------------------------------------------------------------------------
  Get the style and graphicset from the selected treenode
-------------------------------------------------------------------------------}
//var
//  O: TObject;
//  S: TBaseLemmingStyle;
//  G: TBaseGraphicSet;
//  It: TTreeNode;
begin
  aStyle := SelectedStyle;
  aGraph := SelectedGraph;
  Result := True;
  Exit;
  (*


  Result := False;
  It := Tree.Selected;
  if It = nil then
    Exit;
  O := TObject(It.Data);
  if O is TBaseGraphicSet then
    G := TBaseGraphicSet(O)
  else
    Exit;

  It := It.Parent;
  if It = nil then
    Exit;

  O := TObject(It.Data);
  if O is TBaseLemmingStyle then
    S := TBaseLemmingStyle(O)
  else
    Exit;

  aStyle := S;
  aGraph := G;
  Result := True;
    //  aGraph := Tree.S
    *)
end;

procedure TNewLevelForm.ActOK_Execute(Sender: TObject);
begin
   //
end;

procedure TNewLevelForm.ActOKUpdate(Sender: TObject);
begin
//  ActOK.Enabled := (VTree.Selected <> nil) and (TObject(Tree.Selected.Data) is TBaseGraphicSet);
end;

procedure TNewLevelForm.Form_CloseQuery(Sender: TObject; var CanClose: Boolean);
begin
//  if ModalResult = mrOK then
  //  CanClose := (Tree.Selected <> nil) and (TObject(Tree.Selected.Data) is TBaseGraphicSet);
end;

function TNewLevelForm.GetMyNodeData(aNode: PVirtualNode): TMyData;
var
  D: PDataRec;
begin
  Result := nil;
  D := VTree.GetNodeData(aNode);
  if Assigned(D) and (D^.Obj is TMyData) then
    Result := TMyData(D^.Obj);
end;

procedure TNewLevelForm.VTree_FocusChanged(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex);
var
  D: TMyData;
begin
    SelectedStyle:=nil;
    SelectedGraph:=nil;
  D := GetMyNodeData(Node); //virtualtrees
  if D = nil then
    Exit;
{  if D.Graph <> nil then
    CellText := D.Graph.GraphicSetName
  else if D.Style <> nil then
    CellText := D.Style.StyleName
  else
    CellText := ''; }
  selnode:=node;

  if node = nil then
  begin
//    button1.caption := 'compile';
  //  button1.tag := 0;
    ActOK.Enabled := False;////button1.enabled := false;
    SelectedStyle:=nil;
    SelectedGraph:=nil;
  end
  else if D.Graph <> nil then
  begin
    //button1.Caption := 'compile graphic set';
  //  button1.tag := 2;
//    button1.enabled := true;
    ActOK.Enabled := True;
    SelectedGraph:=D.Graph;//TBaseGraphicSet(TObject(Tree.Selected.Data));
    SelectedStyle := SelectedGraph.OwnerStyle;
  end
  else if D.Style <> nil then
  begin
    //button1.Caption := 'compile all graphicsets of this style';
  //  button1.tag := 1;
//    button1.enabled := true;
    ActOK.Enabled := False;////button1.enabled := false;
    SelectedStyle := D.Style;
  end
end;

procedure TNewLevelForm.VTree_FreeNode(Sender: TBaseVirtualTree;
  Node: PVirtualNode);
begin
   GetMyNodeData(Node).Free;
end;

procedure TNewLevelForm.VTree_GetText(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
  var CellText: WideString);
var
  D: TMyData;
begin

  D := GetMyNodeData(Node);
  if D = nil then
    Exit;
  case D.NodeType of
    ntStyle: CellText := D.Style.StyleDescription;
    ntGraph: CellText := D.Graph.GraphicSetName;
  else
    CellText := '';
  end;


end;

end.


