{$include lem_directives.inc}

{-------------------------------------------------------------------------------
  This is the main form of Player.exe
  It handles the logical order/interaction of the existing dos-screens:
  o DosScreenMenu
  o DosScreenPreview
  o FGame
  o DosScreenPostview
  o DosScreenLevelCode
-------------------------------------------------------------------------------}

unit FPlayer;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  lemmisc,umisc,  Dialogs,
  UBasePlayingForm,           fgame,
lemdosstyles,  lemgame, LemCore,

  lemdospreview,
  lemdospostview,

  GR32, GR32_Image, GR32_Layers, StdCtrls;

const
  LM_START = WM_USER + 1;
type
  TLMStart = packed record
    Msg: Cardinal;

  end;

type
  TProgramState = (
    psNone,
    psMenu,
    psPreview,
    psGame,
    psPostView
  );

  TFormPlayer = class(TBasePlayingForm)
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
//    fLevelLoader
    fGameForm       : TFormGame;
    fGame           : TLemmingGame; // reference (for now)
    fLevel          : TLevel;
    fMainDatFile    : string;
    fGameResults    : TGameResultsRec;
    fPreviewScreen  : TDosGamePreview;
    fPostviewScreen : TDosGamePostview;

    fPlayRec: TDosGamePlayInfoRec;
    fState: TProgramState;

    procedure LMStart(var Msg: TMessage); message LM_START;
    procedure LoadNextLevel;
    procedure UnloadLevel;
  { eventhandlers preview }
    procedure PreviewScreen_MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);
    procedure PreviewScreen_KeyDown(Sender: TObject; var Key: Word; SHift: TShiftState);

    procedure PostViewScreen_MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; Layer: TCustomLayer);


    procedure ShowPreview;
    procedure RemovePreview;
    procedure ShowPostview;
    procedure RemovePostView;
//    procedure InternalSetState(Value: TProgramState);
    procedure SetState(Value: TProgramState);
  public
    property State: TProgramState read fState write SetState;
  end;

var
  FormPlayer: TFormPlayer;


implementation

{$R *.dfm}

uses
  LemControls, LemTypes, LemStyles, lemdosfiles;




procedure TFormPlayer.FormCreate(Sender: TObject);
begin
  InitializeScaleAndRect(640, 350);
  color:=clblack;

  fMainDatFile := DosOrigStyle.IncludeCommonPath(DosOrigStyle.MainDataFile);
  fLevel := TLevel.Create(Self);

 // here we should show the menu


   (*

  // initial game for debugging
  with fPlayRec do
  begin
    dValid         := False;
    dPack          := 0;
    dSection       := 1;
    dLevel         := 0;
    dSectionName   := '';
  end; *)

  LoadNextLevel; // put this elsewhere
  ShowPreview; // put this elsewhere
end;


procedure TFormPlayer.LMStart(var Msg: TMessage);
begin
  RemovePostView;
  RemovePreview;
  fLevel.LockLevel;
  //LoadNextLevel;
//  fGameForm := TFormGame.Create(nil);
//  fGameForm.Game.Level := fLevel; // whoops bad programming
//  fGame := fGameForm.Game;
//  fMainDatFile := DosOrigStyle.IncludeCommonPath(DosOrigStyle.MainDataFile);
  fGameForm.LoadAndPlay(fLevel);
  fLevel.UnLockLevel;
  fGameResults := fGame.GameResultRec;
//  FreeAndNil(fGameForm);
  ShowPostView;
end;

procedure TFormPlayer.LoadNextLevel;
var
  Loader: TBaseDosLevelLoader;

begin
  UnloadLevel;

  fGameForm := TFormGame.Create(nil);
  fGame := fGameForm.Game;
  fGame.Level := fLevel;

  Loader := TBaseDosLevelLoader(DosOrigStyle.LevelLoader);

  if fPlayRec.dValid then
  begin
    if fGameResults.gSuccess then
    begin
      if Loader.FindNextLevel(fPlayRec) then
        with fPlayRec do
          Loader.LoadSingleLevel(dPack, dSection, dLevel, flevel)
      else
        raise exception.create('level not found');
    end;    
  end
  else begin
    if Loader.FindLevel(fPlayRec) then
      with fPlayRec do
        Loader.LoadSingleLevel(dPack, dSection, dLevel, flevel);

    (*
    if Loader.FindFirstLevel(fPlayRec) then
      with fPlayRec do
        Loader.LoadSingleLevel(dPack, dSection, dLevel, flevel)
    else
      raise exception.create('level not found');
    *)  
  end;

end;

procedure TFormPlayer.PreviewScreen_MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer;
  Layer: TCustomLayer);
{-------------------------------------------------------------------------------
  We cannot free the preview screen here, in the middle of its eventhandler.
-------------------------------------------------------------------------------}
begin
  PostMessage(Handle, LM_START, 0, 0);
end;

procedure TFormPlayer.PostViewScreen_MouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer;
  Layer: TCustomLayer);
begin
  LoadNextLevel;
  ShowPreview;
//  PostMessage(Handle, LM_START, 0, 0);
end;


procedure TFormPlayer.RemovePreview;
begin
  FreeAndNil(fPreviewScreen);
end;

procedure TFormPlayer.ShowPostview;
//var
  //GameResult: TGameResultsRec;
begin
  fPostviewScreen := TDosGamePostview.Create(nil);
    {
  with GameResult do
  begin
    gCount    := 80;
    gRescue   := 40;
    gDone     := 40;
    gTimeIsUp := False;
  end;
  }

  with fPostviewScreen do
  begin
    MainDatFile := Self.fMainDatFile;
    BuildBitmap(fGameResults, fPlayRec.dSectionName, fPlayRec.dLevel);
    Parent := Self;
    Left := MainRect.Left;
    Top := MainRect.Top;
    Left := 10;
    Top := 30;
    Width := 1000;//960;
    Height := 768; //700;
    ScaleMode := smOptimalScaled;
    Scale := 1.8;
    Show;
    OnMouseDown := PostViewScreen_MouseDown;
  end;

end;

procedure TFormPlayer.ShowPreview;
(*
// allright
  there are a few solutions:
  -integer scale 1:1 or 2:1 for big screens (1280 * 800);
  -non-integer scale calc max

*)
begin
  RemovePostView;

  fPreviewScreen := TDosGamePreview.Create(nil);

  with fPreviewScreen do
  begin
    MainDatFile := Self.fMainDatFile;
    BuildBitmap(Self.fGame, fPlayRec.dSectionName, fPlayRec.dLevel);
    Parent := Self;
    Left := MainRect.Left;
    Top := MainRect.Top;
    Left := 0;
    Top := 30;
    Width := 1024;//960;
    Height := 700;
    ScaleMode := smOptimalScaled;
    Scale := 1.8;
    Show;
//    Self.ActiveControl := fPreviewScreen;
    OnMouseDown := PreviewScreen_MouseDown;
    OnKeyDown := PreviewScreen_KeyDown;
    //Free;
  end;

end;

procedure TFormPlayer.FormDestroy(Sender: TObject);
begin
  // make sure of disposing
  fPreviewScreen.Free;
  fPostviewScreen.Free;
  UnloadLevel;
end;

procedure TFormPlayer.RemovePostView;
begin
   FreeAndNil(fPostViewScreen);
end;

procedure TFormPlayer.UnloadLevel;
begin
  FreeAndNil(fGameForm);
  fGame := nil;
end;

procedure TFormPlayer.SetState(Value: TProgramState);
var
  OldState: TProgramState;
begin
  OldState := fState;
  if OldState = Value then
    Exit;

  // leave old state
  case OldState of
//    psNone,
//    psMenu,
    psPreview: RemovePreview;
//    psGame
    psPostView: RemovePostView;
  end;

  fState := Value;

  // enter new state
  case fState of
    //psNone,
    //psMenu,
    psPreview: ShowPreview;
//    psGame
    psPostView: ShowPostView;
  end;
  
end;

procedure TFormPlayer.PreviewScreen_KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_CANCEL: State := psMenu;
  end;
end;

end.


(*

    MENUSCREEN
      F1            = PREVIEWSCREEN
      F2            = LEVELCODESCREEN
      F3            = sound
      F4            = INPUTSCREEN
      ESCAPE        = quit

    PREVIEWSCREEN
      escape        = MENUSCREEN
      rightclick    = MENUSCREEN
      Leftclick     = GAMESCREEN

    LEVELCODESCREEN
      return        = check code
      escape        = MAINMENUSCREEN

    GAMESCREEN
      escape        = POSTVIEWSCREEN

    POSTVIEWSCREEN
      leftclick     = GAMESCREEN
      rightclick    = MENUSCREEN




*)
(*
procedure TFormPlayer.Button1Click(Sender: TObject);
var
//  aLevel: TLevel;
  Result: Boolean;
  Needed, Done: Integer;
  res: TGameResultsRec;
  datfile: string;
//  GF: TFormGame;
begin
  (*
  RemovePostView;


//  fLevel := TLevel.Create(Self);
  GF := TFormGame.Create(nil);

  begin

//    fLevel.LoadFromFile('isteve.lvl', DosOrigStyle, lffLVL);
//    fLevel.LoadFromFile('builders will help you here.lvl', DosOrigStyle, lffLVL);
//    fLevel.LoadFromFile('justdig.lvl', DosOrigStyle, lffLVL);
    with TBaseDosLevelLoader(DosOrigStyle.LevelLoader) do
    begin
      if fPlayRec.dValid then
      begin
        if FindNextLevel(fPlayRec) then
          with fPlayRec do
            DosOrigStyle.LevelLoader.LoadSingleLevel(dPack, dSection, dLevel, flevel)
        else
          raise exception.create('level not found');
      end
      else begin
        if FindFirstLevel(fPlayRec) then
          with fPlayRec do
            DosOrigStyle.LevelLoader.LoadSingleLevel(dPack, dSection, dLevel, flevel)
        else
          raise exception.create('level not found');
      end;
    end;

    fGameForm.Game.Level := fLevel; // whoops bad programming
    fGame := fGameForm.Game;
    fMainDatFile := DosOrigStyle.IncludeCommonPath(DosOrigStyle.MainDataFile);

    ShowPreview;
  end;

//  fLevel.Free;
            *)
end;

  with fPreviewScreen, Bitmap do
  begin
    // this is fading
//    DrawMode := dmBlend;
  //  MasterAlpha := 140;
  end;

