unit UDataSetExport;

interface

uses
  Classes, Db,
  UMisc, ULog;

procedure DataSetToCSV(Dataset: TDataset; FileName: string);
//function CSVToCursor(FileName: string): TMasterCursor;
//procedure DataSetToXML(Dataset: TMasterDataset; FileName: string);

implementation

uses
  SysUtils;


procedure DataSetToCSV(Dataset: TDataset; FileName: string);
var
  i: integer;
  F: TextFile;
  S: string;
  BkMark: String;

    function HeaderStr: string;
    var
      i: integer;
    begin
      Result := '';
      with DataSet do
        for i := 0 to FieldCount - 1 do
          with Fields[i] {.Def} do
            Result := Result + '"' + FieldName + '"' + ';';
    end;

    function RecordStr: string;
    var
      i: integer;
    begin
      Result := '';
      with DataSet do
        for i := 0 to FieldCount - 1 do
          with Fields[i] do
            if IsNull then Result := Result + '""' + ';' else
              Result := Result + '"' + AsString + '"' + ';';
    end;
begin
  with DataSet do
  begin
    AssignFile(F, FileName);
    Rewrite(F);
    BkMark := {Get}Bookmark;
    DisableControls;
    try
      First;
      WriteLn(F, HeaderStr);
      while not DataSet.Eof do
      begin
        S := RecordStr;
        WriteLn(F, S);
        Next;
      end;
    finally
      Bookmark := BkMark;
      EnableControls;
      CloseFile(F);
    end;
  end;
end;

(*
function CSVToCursor(FileName: string): TMasterCursor;
var
  F: TextFile;
  S: string;
  FldNames: TOpenStringArray;
  FldValues: TOpenVariantArray;
  HeaderFldCount, DataFldCount: integer;
  Lines: integer;
  Cursor: TMasterCursor;

    { lees header }
    function GetFldNames(const S: string): boolean;
    var
      i: integer;
    begin
      Result := False;
      FldNames := SplitString_To_StringArray(S, ';', True);
      HeaderFldCount := Length(FldNames);
      Result := HeaderFldCount > 0;
      for i := 0 to HeaderFldCount - 1 do
        FldNames[i] := UnQuotedStr(FldNames[i]);
      //Logarray(FldNames);
    end;

    { lees data }
    function GetFldValues(const S: string): boolean;
    var
      i: integer;
      Temp: TOpenStringArray;
    begin
      Result := False;
      Temp := SplitString_To_StringArray(S, ';', True);
      DataFldCount := Length(Temp);
      Result := (DataFldCount > 0) and (DataFldCount = HeaderFldCount);
      if Result then
      begin
        //FldValues := StringArray_To_VariantArray(Temp);
        SetLength(FldValues, DataFldCount);
        for i := 0 to DataFldCount - 1 do
        begin
          Temp[i] := UnQuotedStr(Temp[i]);
          FldValues[i] := Temp[i];
        end;
      end;
    end;

    { maak cursor }
    procedure CreateCursor;
    var
      i: integer;
      FldTypes: TDataTypeArray;
    begin
      SetLength(FldTypes, HeaderFldCount);
      for i := 0 to HeaderFldCount - 1 do
        FldTypes[i] := dtString;
      Cursor := CreateSimpleCursor('Cursor', FldNames, FldTypes);
    end;

    { append aan dataset }
    procedure AppendData;
    var
      i: integer;
    begin
      if Cursor = nil then
        CreateCursor;
      Cursor.AppendRecordByNames(FldNames, FldValues);
    end;

begin
  Cursor := nil;
  Result := nil;
  AssignFile(F, FileName);
  Reset(F);
  Lines := 0;
  try
    while not Eof(F) do
    begin
      ReadLn(F, S);
      Inc(Lines);
      { header }
      if Lines = 1 then
      begin
        if not GetFldNames(S) then Break;
      end
      { rest }
      else begin
        if GetFldValues(S) then AppendData;
      end;
    end;
    Cursor.Open;
    Cursor.First;
    Result := Cursor;
  finally
    CloseFile(F);
  end;
end;


{ *** DataSetToXML *** }

var
  SourceBuffer: PChar;

procedure WriteString(Stream: TFileStream; s: string);
begin
  StrPCopy(SourceBuffer, s);
  Stream.Write(SourceBuffer[0], StrLen(SourceBuffer));
end;

procedure WriteFileBegin(Stream: TFileStream; Dataset: TMasterDataset);

  function XMLFieldType(fld: TField): string;
  begin
    case fld.def.DataType of
      dtString: Result := '"string" WIDTH="' + IntToStr(Length(fld.AsString)) + '"';
//      ftSmallint: Result := '"i4"'; //??
      dtInteger: Result := '"i4"';
//      ftWord: Result := '"i4"'; //??
//      ftBoolean: Result := '"boolean"';
//      ftAutoInc: Result := '"i4" SUBTYPE="Autoinc"';
//      ftFloat: Result := '"r8"';
//      ftCurrency: Result := '"r8" SUBTYPE="Money"';
//      ftBCD: Result := '"r8"'; //??
      dtDate: Result := '"date"';
//      ftTime: Result := '"time"'; //??
//      ftDateTime: Result := '"datetime"';
    else
    end;
    if fld.Def.Required then
      Result := Result + ' required="true"';
//    if fld.Def.Readonly then
  //    Result := Result + ' readonly="true"';
  end;

var
  i: Integer;
begin
  WriteString(Stream, '<?xml version="1.0" standalone="yes"?><!-- Generated by SMExport -->  ' +
                      '<DATAPACKET Version="2.0">');
  WriteString(Stream, '<METADATA><FIELDS>');

  {write th metadata}
  with Dataset do
    for i := 0 to FieldCount-1 do
    begin
      WriteString(Stream, '<FIELD attrname="' +
                          Fields[i].FieldName +
                          '" fieldtype=' +
                          XMLFieldType(Fields[i]) +
                          '/>');
    end;
  WriteString(Stream, '</FIELDS>');
  WriteString(Stream, '<PARAMS DEFAULT_ORDER="1" PRIMARY_KEY="1" LCID="1033"/>');
  WriteString(Stream, '</METADATA><ROWDATA>');
end;

procedure WriteFileEnd(Stream: TFileStream);
begin
  WriteString(Stream, '</ROWDATA></DATAPACKET>');
end;

procedure WriteRowStart(Stream: TFileStream; IsAddedTitle: Boolean);
begin
  if not IsAddedTitle then
    WriteString(Stream, '<ROW');
end;

procedure WriteRowEnd(Stream: TFileStream; IsAddedTitle: Boolean);
begin
  if not IsAddedTitle then
    WriteString(Stream, '/>');
end;

procedure WriteData(Stream: TFileStream; fld: TField; AString: ShortString);
begin
  if Assigned(fld) and (AString <> '') then
    WriteString(Stream, ' ' + fld.FieldName + '="' + AString + '"');
end;

function GetFieldStr(Field: TField): string;

  function GetDig(i, j: Word): string;
  begin
    Result := IntToStr(i);
    while (Length(Result) < j) do
      Result := '0' + Result;
  end;

var Hour, Min, Sec, MSec: Word;
begin
  case Field.Def.DataType of
//    ftBoolean: Result := UpperCase(Field.AsString);
    dtDate: Result := Field.AsString;//FormatDateTime('yyyymmdd', Field.AsDateTime);
//    ftTime: Result := FormatDateTime('hhnnss', Field.AsDateTime);
{    ftDateTime: begin
                  Result := FormatDateTime('yyyymmdd', Field.AsDateTime);
                  DecodeTime(Field.AsDateTime, Hour, Min, Sec, MSec);
                  if (Hour <> 0) or (Min <> 0) or (Sec <> 0) or (MSec <> 0) then
                    Result := Result + 'T' + GetDig(Hour, 2) + ':' + GetDig(Min, 2) + ':' + GetDig(Sec, 2) + GetDig(MSec, 3);
                end; }
  else
    Result := Field.AsString;
  end;
end;

procedure DatasetToXML(Dataset: TMasterDataset; FileName: string);
var
  Stream: TFileStream;
  bkmark: integer;//TBookmark;
  i: Integer;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  SourceBuffer := StrAlloc(1024);
  WriteFileBegin(Stream, Dataset);

  with DataSet do
  begin
    DisableControls;
    bkmark := GetBookmark;
    First;

    {write a title row}
    WriteRowStart(Stream, True);
    for i := 0 to FieldCount-1 do
      WriteData(Stream, nil, Fields[i].Def.Prompt{Fields[i].DisplayLabel});
    {write the end of row}
    WriteRowEnd(Stream, True);

    while (not EOF) do
    begin
      WriteRowStart(Stream, False);
      for i := 0 to FieldCount-1 do
        WriteData(Stream, Fields[i], GetFieldStr(Fields[i]));
      {write the end of row}
      WriteRowEnd(Stream, False);

      Next;
    end;

    GotoBookmark(bkmark);
    EnableControls;
  end;

  WriteFileEnd(Stream);
  Stream.Free;
  StrDispose(SourceBuffer);
end;
*)

end.

