{$include lem_directives.inc}

{-------------------------------------------------------------------------------
   The dos postview screen, which shows you how you've done it.
-------------------------------------------------------------------------------}

unit LemDosPostview;

interface

uses
  Classes, SysUtils,
  Gr32, Gr32_Image,
  UMisc,
  LemCore, LemGame, LemDosFiles, LemDosStyles, LemControls,
  LemDosScreen;

{-------------------------------------------------------------------------------
   The dos postview screen, which shows you how you've done it.
-------------------------------------------------------------------------------}
type
  TDosGamePostview = class(TBaseDosScreenImage)
  private
    fSection     : string;
    fLevelNumber : Integer;
    fGameResult  : TGameResultsRec;
    function GetScreenText(aLevelNumber: Integer): string;
  protected
    procedure LoadImages; override;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure BuildBitmap(const aResult: TGameResultsRec;
       const aSection: string; aLevelNumber: Integer);
  published
  end;

implementation

const
  // results from worst to best
  OrigResultStrings: array[0..8] of string = (
    'ROCK BOTTOM! I hope for your sake'      + #13 + 'that you nuked that level.',
    'Better rethink your strategy before'    + #13 + 'you try this level again!',
    'A little more practice on this level'   + #13 + 'is definitely recommended.',
    'You got pretty close that time.'        + #13 + 'Now try again for that few % extra.',
    'OH NO, So near and yet so far (teehee)' + #13 + 'Maybe this time.....',
    'RIGHT ON. You can''t get much closer'   + #13 + 'than that. Let''s try the next...',
    'That level seemed no problem to you on' + #13 + 'that attempt. Onto the next....',
    'You totally stormed that level!'        + #13 + 'Let''s see if you can storm the next...',
    'Superb! You rescued every lemmings on'  + #13 +  'that level. Can you do it again....'
  );

{ TDosGamePreview }

procedure TDosGamePostview.BuildBitmap(const aResult: TGameResultsRec;
  const aSection: string; aLevelNumber: Integer);
var
  Temp: TBitmap32;
  DstRect: TRect;
begin
  fSection := aSection;
  fLevelNumber := aLevelNumber;
  fGameResult := aResult;

  Temp := TBitmap32.Create;
  try
    Temp.SetSize(640, 350);
    Temp.Clear(0);
    // draw background
    TileBackgroundBitmap(Temp, 0, 0);
    // draw text
    DrawPurpleText(Temp, GetScreenText(aLevelNumber), 0, 0);

//    DrawPurpleText(Temp, GetScreenText, 0, 80);
//    Temp.SaveToFile(apppath + 'testscreen.bmp');
    Bitmap.Assign(Temp);
  finally
    Temp.Free;
  end;
end;

constructor TDosGamePostview.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  Visible := False;
end;

destructor TDosGamePostview.Destroy;
begin
  inherited Destroy;
end;

function TDosGamePostview.GetScreenText(aLevelNumber: Integer): string;
var
  N, i: Integer;
  SToRescuePerc: string;
  SRescuedPerc: string;

    procedure Add(const S: string; LineFeeds: Integer);
    begin
      if S <> '' then
        Result := Result + PadC(S, 40) + StringOfChar(#13, LineFeeds)
      else
        Result := Result + #13 + StringOfChar(#13, LineFeeds)
    end;

    procedure AddResultText(const S: string);
    // adds result, splitting and adding 6 linefeeds
    begin
      Result := Result + PadC(SplitString(S, 0, #13), 40) + #13;
      Result := Result + PadC(SplitString(S, 1, #13), 40);
      Result := Result + StringOfChar(#13, 6);
    end;


begin

  Result := '';
  with fGameResult do
  begin
    // init some local strings
    SToRescuePerc := i2s(gTarget) + '%';
    SRescuedPerc  := i2s(gDone) + '%';

    // top text
    Add('', 0);
    if gTimeIsUp then
      Add('Your time is up!', 2)
    else
      Add('All lemmings accounted for.', 2);

    Add('You rescued ' + SRescuedPerc, 1);
    Add('You needed  ' + SToRescuePerc, 2);


    // result text
    if gDone = 100 then
      i := 8
    else if gDone = 0 then
      i := 0
    else if gDone < gTarget div 2 then
      i := 1
    else if gDone < gTarget - 5 then
      i := 2
    else if gDone < gTarget - 1 then
      i := 3
    else if gDone = gTarget - 1 then
      i := 4
    else if gDone = gTarget then
      i := 5
    else if gDone < gTarget + 20 then
      i := 6
    else if gDone >= gTarget + 20 then
      i := 7
    else raise exception.Create('leveltext error');

    AddResultText(OrigResultStrings[i]);

    if gSuccess then
    begin
      // new levelcode text
      Add('Your Access Code for Level ' + i2s(aLevelNumber + 2), 1);
      Add('is AAAAABBBBB', 3);
    end
    else begin
      // else fake space
      Add('', 1);
      Add('', 3);
    end;

    // bottom text
    if gSuccess then
      Add('Press left mouse button for next level', 1)
    else
      Add('Press left mouse button to retry level', 1);
    Add('Press right mouse button for menu', 0);

  end;

end;

procedure TDosGamePostview.LoadImages;
begin
  inherited LoadImages;

end;

end.


