{$include lem_directives.inc}

unit LemDosPreview;

interface

uses
  Classes, SysUtils,
  Gr32, Gr32_Image,
  UMisc,
  LemCore, LemGame, LemDosFiles, LemDosStyles, LemControls,
  LemDosScreen;

const
  SPreviewString =
     'Level %d ' + '%s'                     + #13#13#13 +
     '          Number of Lemmings %d'      + #13#13 +
     '          %s To Be Saved'             + #13#13 +
     '          Release Rate %d'            + #13#13 +
     '          Time %d Minutes'            + #13#13 +
     '          Rating  %s'                 + #13#13#13 +
     '     Press mouse button to continue';

  {-------------------------------------------------------------------------------
     The dos preview screen
  -------------------------------------------------------------------------------}
type
  TDosGamePreview = class(TBaseDosScreenImage)
  private
    fGame        : TLemmingGame; // reference
    fSection     : string;
    fLevelNumber : Integer;
    function GetScreenText: string;
  protected
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure BuildBitmap(aGame: TLemmingGame; const aSection: string; aLevelNumber: Integer);
  published
  end;

implementation

{ TDosGamePreview }

procedure TDosGamePreview.BuildBitmap(aGame: TLemmingGame; const aSection: string; aLevelNumber: Integer);
// levelnumber = zerobased
var
  Temp, W: TBitmap32;
  DstRect: TRect;
begin
  fGame := aGame;
  fSection := aSection;
  fLevelNumber := aLevelNumber;

  Temp := TBitmap32.Create;
  W := TBitmap32.Create;
  try
    Temp.SetSize(640, 350);
    Temp.Clear(0);
    // draw level preview
    fGame.Renderer.LevelBitmap := W;
    fGame.Renderer.RenderOptions := [];
    fGame.Renderer.RenderTypes := [litTerrain, litObject];
    fGame.Renderer.Preview := True;
    fGame.Renderer.RenderLevel(fGame.Level);
    fGame.Renderer.Preview := False;
//    DstRect := Rect(120, 20, 120 + W.Width div 4, 20 + W.Height div 4);
    DstRect := Rect(0, 0, 400, 40);
    RectMove(DstRect, 120, 20);
    W.DrawTo(Temp, DstRect, W.BoundsRect);
    // draw background
    TileBackgroundBitmap(Temp, 0, 78);
    // draw text
    DrawPurpleText(Temp, GetScreenText, 0, 80);
//    Temp.SaveToFile(apppath + 'testscreen.bmp');
    Bitmap.Assign(Temp);
  finally
    W.Free;
    Temp.Free;
  end;
end;

constructor TDosGamePreview.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  Visible := False;
  Width := 640;
  Height := 350;
end;

destructor TDosGamePreview.Destroy;
begin
  inherited Destroy;
end;

function TDosGamePreview.GetScreenText: string;
{-------------------------------------------------------------------------------
  Internal function to retrieve level preview text
-------------------------------------------------------------------------------}
var
  Perc: string;
begin
  with fGame.Level, Statics do
  begin
    Perc := PadR(i2s(Percentage(LemmingsCount, RescueCount)) + '%', 4);
    Result := Format(SPreviewString, [fLevelNumber + 1,
                                LevelName,
                                LemmingsCount,
                                Perc,
                                ReleaseRate,
                                TimeLimit,
                                fSection
                                ]);
  end;

end;

end.
