unit LemDosScreen;

interface


uses
  Classes, SysUtils,
  Gr32, Gr32_Image,
  UMisc,
  LemCore, LemGame, LemDosFiles, LemDosStyles, LemControls;

type
  {-------------------------------------------------------------------------------
    abstract ancestor for dos-screens
  -------------------------------------------------------------------------------}
  TBaseDosScreenImage = class(TImage32)
  private
    fPurpleFont   : TPurpleFont;
    fBackground   : TBitmap32;
    fMainDatFile  : string;
    fDataExtractor: TMainDatExtractor;
    procedure SetMainDatFile(const Value: string);
  protected
    procedure LoadImages; virtual;
    procedure TileBackgroundBitmap(Dst: TBitmap32; X, Y: Integer);
    property DataExtractor: TMainDatExtractor read fDataExtractor;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure DrawPurpleText(Dst: TBitmap32; const S: string; X, Y: Integer);
    property MainDatFile: string read fMainDatFile write SetMainDatFile;
  published
  end;

implementation

{ TBaseDosScreenImage }

constructor TBaseDosScreenImage.Create(aOwner: TComponent);
begin
  inherited;
  fDataExtractor := TMainDatExtractor.Create('');
  fPurpleFont := TPurpleFont.Create(Self);
  fBackGround := TBitmap32.Create;

end;

destructor TBaseDosScreenImage.Destroy;
begin
  fBackground.Free;
  fDataExtractor.Free;
  inherited;
end;

procedure TBaseDosScreenImage.TileBackgroundBitmap(Dst: TBitmap32; X, Y: Integer);
var
  aX, aY: Integer;
  S, D: TBitmap32;
begin
  Assert(fBackground.Width > 0);
  Assert(fBackground.Height > 0);

  aY := Y;
  aX := X;
  while aY <= Dst.Height do
  begin
    while aX <= Dst.Width do
    begin
      fBackground.DrawTo(Dst, aX, aY);
      Inc(aX, fBackground.Width);
    end;
    Inc(aY, fBackground.Height);
    aX := X;
  end;
end;


procedure TBaseDosScreenImage.DrawPurpleText(Dst: TBitmap32; const S: string; X, Y: Integer);
{-------------------------------------------------------------------------------
  Linefeeds increment 16 pixels
  Spaces increment 16 pixels
-------------------------------------------------------------------------------}
var
  C: Char;
  CX, CY, i: Integer;
begin
  CX := X;
  CY := Y;
  for i := 1 to Length(S) do
  begin
    C := S[i];
    case C of
      #13:
        begin
          Inc(CY, 16);
          CX := X;
        end;
      ' ':
        begin
          Inc(CX, 16);
        end;
      '!'..'~':
        begin
          fPurpleFont.BitmapOfChar[C].DrawTo(Dst, CX, CY);
          Inc(CX, 16);
        end;
    end;
  end;
end;

procedure TBaseDosScreenImage.LoadImages;
var
  i: Integer;
begin
  with DataExtractor do
  begin
    // background
    ExtractBrownBackground(fBackground);

    // purple font
    DoExtract(4);
    fMems[4].Seek($69B0, soFromBeginning);
    Planar.DirectRGB := True;
    for i := 0 to PURPLEFONTCOUNT - 1 do
      Planar.LoadFromStream(fMems[4], fPurpleFont.fBitmaps[i], -1, 16,16, 3, DosMainMenuPalette);
  end;
end;

procedure TBaseDosScreenImage.SetMainDatFile(const Value: string);
begin
  Assert(Value <> '');
  if CompareText(fMainDatFile, Value) = 0 then
    Exit;
  fMainDatFile := Value;
  DataExtractor.MainFile := fMainDatFile;
  LoadImages;
end;


end.

