{$include lem_directives.inc}

unit LemFiles;

interface

uses
  Windows, Classes, SysUtils, Graphics, GR32,
  UMisc;

type
  TColorArray = array of DWORD;

const
  LVL_MAXOBJECTCOUNT  = 32;
  LVL_MAXTERRAINCOUNT = 400;
  LVL_MAXSTEELCOUNT   = 32;

  GRAPHICSET_DIRT      = 0;
  GRAPHICSET_FIRE      = 1;
  GRAPHICSET_MARBLE    = 2;
  GRAPHICSET_PILLAR    = 3;
  GRAPHICSET_CRYSTAL  = 4;
  GRAPHICSET_BRICK     = 5;
  GRAPHICSET_ROCK      = 6;
  GRAPHICSET_SNOW      = 7;
  GRAPHICSET_BUBBLE    = 8;
  GRAPHICSET_XMAS      = 9;

  EXTGRAPHICSET_01     = 1;
  EXTGRAPHICSET_02     = 2;
  EXTGRAPHICSET_03     = 3;
  EXTGRAPHICSET_04     = 4;

//  OBJECT_EXIT

  // game object terrain modifier consts
  OBJECTMODIFIER_NOOVERWRITE         = $80; // do not overwrite existing terrain
  OBJECTMODIFIER_NEEDSTERRAIN        = $40; // must have terrain underneath to be visible
  OBJECTMODIFIER_ALWAYSDRAW          = $00; // always draw full graphic

  OBJECTDISPLAY_NORMAL         = $0F;
  OBJECTDISPLAY_UPSIDEDOWN     = $8F;

  //graphics set 0
  OBJECT_G00_EXIT                          = 0;
  OBJECT_G00_START           	             = 1;
  OBJECT_G00_WAVING_GREEN_FLAG	           = 2;
  OBJECT_G00_ONE_WAY_BLOCK_LEFT            = 3;
  OBJECT_G00_ONE_WAY_BLOCK_RIGHT	         = 4;
  OBJECT_G00_WATER               	         = 5;
  OBJECT_G00_BEAR_TRAP          	         = 6;
  OBJECT_G00_EXIT_DECORATION_FLAMES        = 7;
  OBJECT_G00_ROCK_SQUISHING_TRAP           = 8;
  OBJECT_G00_WAVING_BLUE_FLAG           	 = 9;
  OBJECT_G00_10_TON_SQUISHING_TRAP         = 10;

  //graphics set 1
  OBJECT_G01_EXIT                          = 0;
  OBJECT_G01_START                         = 1;
  OBJECT_G01_WAVING_GREEN_FLAG             = 2;
  OBJECT_G01_ONE_WAY_BLOCK_LEFT            = 3;
  OBJECT_G01_ONE_WAY_BLOCK_RIGHT           = 4;
  OBJECT_G01_RED_LAVA                      = 5;
  OBJECT_G01_EXIT_FLAMES                   = 6; //
  OBJECT_G01_FIRE_PIT_TRAP                 = 7;
  OBJECT_G01_FIRE_SHOOTER_TRAP_FROM_LEFT   = 8;
  OBJECT_G01_WAVING_BLUE_FLAG              = 9;
  OBJECT_G01_FIRE_SHOOTER_TRAP_FROM_RIGHT  = 10;

  //graphics set 2
	OBJECT_G02_EXIT                          = 0;
	OBJECT_G02_START                         = 1;
	OBJECT_G02_WAVING_GREEN_FLAG             = 2;
	OBJECT_G02_ONE_WAY_BLOCK_LEFT            = 3;
	OBJECT_G02_ONE_WAY_BLOCK_RIGHT           = 4;
	OBJECT_G02_GREEN_LIQUID                  = 5;
	OBJECT_G02_EXIT_FLAMES                   = 6; //
	OBJECT_G02_WAVING_BLUE_FLAG              = 7;
	OBJECT_G02_PILLAR_SQUISHING_TRAP         = 8;
	OBJECT_G02_SPINNING_DEATH_TRAP           = 9;

  OBJECT_G03_EXIT                         = 0;
  OBJECT_G03_START                        = 1;
  OBJECT_G03_WAVING_GREENFLAG             = 2;
  OBJECT_G03_ONEWAY_BLOCK_LEFT            = 3;
  OBJECT_G03_ONEWAY_BLOCK_RIGHT           = 4;
  OBJECT_G03_WATER                        = 5;
  OBJECT_G03_EXIT_FLAMES                  = 6; //
  OBJECT_G03_WAVING_BLUE_FLAG             = 7;
  OBJECT_G03_SPINNY_ROPE_TRAP             = 8;
  OBJECT_G03_SPIKES_FROM_LEFT_TRAP        = 9;
  OBJECT_G03_SPIKES_FROM_RIGHT_TRAP       = 10;


type
  TGraphicSetRange = GRAPHICSET_DIRT..GRAPHICSET_BUBBLE;
//  TExtendedGraphicSetRange
const
  TheGraphicSets = [GRAPHICSET_DIRT..GRAPHICSET_BUBBLE];

type
  TLVLFile = class;
  TSPRFile = class;
  TPALFile = class;

  TGenericFileFormat = class(TPersistent)
  private
  protected
  public
    procedure LoadFromFile(const aFileName: string); virtual;
    procedure LoadFromStream(S: TStream); virtual;
  published
  end;

  

  {-------------------------------------------------------------------------------
    .LVL files (WinLemmings)
  -------------------------------------------------------------------------------}

  // sometimes swapping needed of Lo Hi!!!!!!!!!!!!!!

  // sub record interactive objects
  TLVLObject = packed record
    case Byte of
      0: (
            XPos              : Word; // swap lo and hi
            YPos              : Word; // swap lo and hi
            ObjectID          : Word;
            Modifier          : Byte;  // objectmodifier_xxx
            DisplayMode       : Byte;
//            DisplayMode       : Byte;  // objectdisplay_xxx
  //          _NotUsed          : array[0..2] of Byte
          );
      1: (
           AsInt64: Int64;
         );
      2: (
            B0, B1, B2, B3, B4, B5, B6, B7: Byte;
         );
  end;

  // sub record terrain
  {
    bits 0..3    = modifier
    bits 4..7    shl 8 for xpos
    bits 8..15   add to xpos
    bits 16..24  9 bits number YPos


  }
  TLVLTerrain = packed record
    case Byte of
      0: (
            XPos              : Word;
            YPos              : Byte; // 9 bits
            TerrainID         : Byte;
         );
      1: (
            AsDWORD: DWORD;
         );
      2: (
            W0: Word;//Lo: Word;
            W1: Word;//Hi: Word;
         );
      3: (
            B0, B1, B2, B3: Byte;
         );
  end;


  TLVLSteel = packed record
    case Byte of
      0: (
           XPos              : Byte; // 9 bits
           YPos              : Byte; // 7 bits bits 1..7
           Area              : Byte; // bits 0..3 is height in 4 pixel units (then add 4)
                                     // bit 4..7 is width in 4 pixel units (then add 4)
           b                 : Byte; // always zero
         );
      1: (
            AsDWORD: DWORD;
         );
      2: (
            W0: Word;
            W1: Word;
         );
      3: (
            B0, B1, B2, B3: Byte;
         );
  end;

  // the main record 2048 bytes
  PLVLFileRec = ^TLVLFileRec;
  TLVLFileRec = packed record
    {0000}  _ReleaseRateNotUsed        : Byte; // little endian's
    {    }  ReleaseRate                : Byte;
    {    }  _LemmingsCountNotUsed      : Byte;
    {    }  LemmingsCount              : Byte;
    {    }  _RescueCountNotUsed        : Byte;
    {    }  RescueCount                : Byte;
    {    }  _TimeLimitNotUsed          : Byte;
    {    }  TimeLimit                  : Byte;
    {0008}  _ClimberNotUsed            : Byte;
    {    }  ClimberCount               : Byte;
    {    }  _FloaterNotUsed            : Byte;
    {    }  FloaterCount               : Byte;
    {    }  _BomberNotUsed             : Byte;
    {    }  BomberCount                : Byte;
    {    }  _BlockerNotUsed            : Byte;
    {    }  BlockerCount               : Byte;
    {    }  _BuilderNotUsed            : Byte;
    {    }  BuilderCount               : Byte;
    {    }  _BasherNotUsed             : Byte;
    {    }  BasherCount                : Byte;
    {    }  _MinerNotUsed              : Byte;
    {    }  MinerCount                 : Byte;
    {    }  _DiggerNotUsed             : Byte;
    {    }  DiggerCount                : Byte;
    {0018}  ScreenPosition             : Word;
    {001A}  _GraphicSetNoUsed          : Byte;
    {    }  GraphicSet                 : Byte; // graphicset_xxx
    {    }  _GraphicSetExNotUsed       : Byte;
    {    }  GraphicSetEx               : Byte;
    {001E}  Reserved1                  : Byte; // I think it is a filler until the 32 bytes
    {    }  Reserved2                  : Byte; // I think it is a filler until the 32 bytes
                                               // And it was, but for the superlemminglevel
                                               // in "ohno more lemmings" both bytes are $FF
    {0020}  Objects                    : array[0..LVL_MAXOBJECTCOUNT - 1] of TLVLObject;
    {0120}  Terrain                    : array[0..LVL_MAXTERRAINCOUNT - 1] of TLVLTerrain;
    {0760}  Steel                      : array[0..LVL_MAXSTEELCOUNT - 1] of TLVLSteel;
    {07E0}  LevelName                  : array[0..31] of Char;
  end;

  TLVLFile = class(TGenericFileFormat)
  private
    function GetLevelName: string;
  protected
  public
    fRec: TLVLFileRec;
    procedure LoadFromStream(S: TStream); override;
    //function
  published
    property ReleaseRate: Byte read fRec.ReleaseRate;
    property LemmingsCount: Byte read fRec.LemmingsCount;
    property RescueCount: Byte read fRec.RescueCount;
    property TimeLimit: Byte read fRec.TimeLimit;
    property ClimberCount: Byte read fRec.ClimberCount;
    property FloaterCount: Byte read fRec.FloaterCount;
    property BomberCount: Byte read fRec.BomberCount;
    property BlockerCount: Byte read fRec.BlockerCount;
    property BuilderCount: Byte read fRec.BuilderCount;
    property BasherCount: Byte read fRec.BasherCount;
    property MinerCount: Byte read fRec.MinerCount;
    property DiggerCount: Byte read fRec.DiggerCount;
    property GraphicSet: Byte read fRec.GraphicSet;
    property LevelName: string read GetLevelName;
  end;


  {-------------------------------------------------------------------------------
    .SPR files (WinLemmings)
  -------------------------------------------------------------------------------}

  // sub record: there are several of this in SPRFileRec
  TSPRAnimationInfoRec = packed record
    XOffset      : Word;
    YOffset      : Word;
    MaxDataWidth : Word;
    DataHeight   : Word;  // Lines of data
    ImageWidth   : Word;
    ImageHeight  : Word;
    // imagedata follows here
  end;

  TSPRFrameRec = record
    Data: array of string;
  end;

  TSPRFileHeaderRec = packed record
    Signature        : array[0..3] of Char; //'SRLE'
    FrameCount       : Word;
    FirstFrameOffset : Word;
    // AnimationInfoRecs follow here
  end;

  TSPRAnimationInfo = class
  private
    fFilePosition: Integer;
    fRec: TSPRAnimationInfoRec;
  protected
  public
  published
  end;
         (*
  TSPRAnimationInfoList = class(TObjectList)
  private
    function GetItem(Index: Integer): TSPRAnimationInfo;
  protected
  public
    function Add(Item: TSPRAnimationInfo): Integer;
    procedure Insert(Index: Integer; Item: TSPRAnimationInfo);
    property Items[Index: Integer]: TSPRAnimationInfo read GetItem; default;
  published
  end; *)

  TSPRFile = class(TGenericFileFormat)
  private
    procedure Clear;
  protected
  public
    fHeader        : TSPRFileHeaderRec;
    fAnimationRecs : array of TSPRAnimationInfoRec;
    fFrames        : array of TSPRFrameRec; // the encoded bitmap lines in string format
    destructor Destroy; override;
    procedure PrepareFromStream(S: TStream);
    procedure LoadFromStream(S: TStream); override;
    procedure ConstructBitmap(Index: Integer; BMP: TBitmap32; const aColors: TColorArray = nil; DebugIt: Boolean = False);
  published
  end;

  (*
  TSPRFileEx = class(TGenericFileFormat)
  private
  protected
  public
    fRec           : TSPRFileRec;
    fAnimationRecs : array of TSPRAnimationInfoRec;
    destructor Destroy; override;
    procedure PrepareFromStream(S: TStream);
    procedure LoadFromStream(S: TStream); override;
//    procedure ConstructBitmap(Index: Integer; BMP: TBitmap32; const aColors: TColorArray = nil; DebugIt: Boolean = False);
  published
  end; *)

  {-------------------------------------------------------------------------------
    .PAL files (WinLemmings)
  -------------------------------------------------------------------------------}
  TPALHeaderRec = packed record
    ID: Byte; // $20
    Header: array[0..2] of Char; //'LAP'
    EntryCount: Word;
  end;

  TPALFile = class(TGenericFileFormat)
  private
    procedure Clear;
  protected
  public
    fRec: TPALHeaderRec;
    fColors: TColorArray;//array of DWORD;    //types
    destructor Destroy; override;
    procedure LoadFromStream(S: TStream); override;
  published
  end;

  {-------------------------------------------------------------------------------
    Win lemmings dbfile
  -------------------------------------------------------------------------------}
  TWinLemDBHeader = packed record
    wUnknown: array[0..17] of Byte; // 18 bytes;
  end;

  TWinLemDBLevelItem = packed record
    wPseudoTitle: array[0..39] of Char;          // 40 bytes 40
    wLevelFileName: array[0..33] of Char;        // 34 bytes 74
    wLevelBitmapFileName: array[0..43] of Char;  // 44 bytes 118
    wUnknown: array[0..18] of Byte;              // 19 bytes 137
  end;


  (*
0000: 2 (=?)
0001
0002
0003
0004: 220 (=number of levels?)

18: Original Levels....
58: orig
155 Fun....

292 just dig (40 bytes)
332 lvl0091.lvl (34 bytes)
366 ls0091.bmp (44 bytes)
410 unknown (19 bytes)

429 Only floaters

566 title

703 title


137 bytes per level
  *)


  //the file is 30148 bytes
  // this record is 30158 bytes. that's probably because some pointer misses at the end.
  PWinLemDB = ^TWinLemDB;
  TWinLemDB = packed record
    Header: TWinLemDBHeader;
    Recs: array[0..219] of TWinLemDBLEvelItem;
  end;


implementation

uses
  LemMisc;

{ TGenericFileFormat }

procedure TGenericFileFormat.LoadFromFile(const aFileName: string);
var
  F: TFileStream;
begin
  F := TFileStream.Create(aFileName, fmOpenRead);
  try
    LoadFromStream(F);
  finally
    F.Free;
  end;
end;

procedure TGenericFileFormat.LoadFromStream(S: TStream);
begin
  //raise Exception.Create('TGenericFileFormatL');
end;


{ TLVLFile }

function TLVLFile.GetLevelName: string;
begin
  Result := Trim(fRec.LevelName);
end;

procedure TLVLFile.LoadFromStream(S: TStream);
var
  R: Integer;
  Buf: TLVLFileRec;
begin
  R := S.Read(Buf, 2048);
  if R <> 2048 then
    raise Exception.Create('LVL read error');
  fRec := Buf;
end;

{ TSPRFile }

procedure TSPRFile.Clear;
begin
  FillChar(fHeader, SizeOf(fHeader), 0);
  fAnimationRecs := nil;
  fFrames := nil;
end;


procedure TSPRFile.ConstructBitmap(Index: Integer; BMP: TBitmap32; const aColors: TColorArray = nil; DebugIt: Boolean = False);

type
  TLineState = (
    lsBegin,
    lsReady,
    lsSpace
  );

var
  CurrentParsingLine: Integer;
  LineBuffer: array of Byte;

    procedure DoPixel(X, Y: Integer; Entry: Byte);
    begin
      if aColors = nil then
        BMP.Canvas.Pixels[x, y] := clyellow//RGB(Entry, Entry, Entry)
      else begin

        if Entry >= Length(aColors) then
        begin
          if Entry > 128 then
            Dec(Entry, 128)
          {else if Entry > 64 then
            Dec(Entry, 64); }

//          if Entry >= Length(Pal.fColors) then
  //          raise Exception.CreateFmt('TSPRFile.ContructBitmap color index error %d %d', [Entry, Length(Pal.fColors)]);
        end;

        // #EL changed mei 2006
        if Entry < Length(aColors) then
          if Entry > 0 then
            BMP.Pixel[x,y] := Color32(aColors[Entry]);


            (*
        if Entry < Length(aColors) then
          BMP.Canvas.Pixels[x, y] := aColors[Entry]
        else
          BMP.Canvas.Pixels[x, y] := clYellow;
          *)


      end;
    end;

    procedure ParseLine(const S: string; aOffset: Integer);

    // parseline is nu goed maar dient netjes gemaakt!!!!!
    var
      i, j: Integer;
      B, CurrentByte: Byte;
      XOffset, PixelCount: Byte;
//      State: TLineState;
      P: Integer;
      BufferPos: Integer;

        function Valid(aBufferPos: Integer): Boolean;
        begin
          Result := aBufferPos < Length(LineBuffer);
        end;

        procedure Check(aBufferPos: Integer);
        begin
          if aBufferPos >= Length(LineBuffer) then
            raise Exception.Create('bufferpos error ' + i2s(abufferpos) + ', ' + i2s(length(linebuffer)));
        end;

    begin

      FillChar(LineBuffer[0], Length(LineBuffer), 0);

      i := 1;
      BufferPos := aOffset;
      XOffset := 0;
//      if CurrentParsingLine = 48 then
  //      Log(['starting 48']);

      repeat
        if i > Length(S) then
          Break;
        CurrentByte := Byte(S[i]);
        if CurrentByte = $80 then // end of line
        begin
          Break;
        end
        else if CurrentByte = 255 then // 127 bytes to follow
        begin
          PixelCount := 127;
          Inc(i);
          for P := i to i + PixelCount - 1 do
          begin
            B := Byte(S[P]);
            //Log([bufferpos]);
            Check(BufferPos);
            LineBuffer[BufferPos] := B;
            Inc(BufferPos);
          end;
          //i := P + 1;
          Inc(i, PixelCount);
        end
        else if CurrentByte < $7F then // x offset
        begin
          XOffset := CurrentByte;
          Inc(BufferPos, XOffset);
//          Log([bufferpos]);
          Inc(i);
        end
        else if CurrentByte > $7F then // omitted x offset
        begin
    //      if CurrentParsingLine = 48 then if DebugIt then
      //      Log([i, 'OMITTED OFFSET']);
          XOffset := 0;
          PixelCount := CurrentByte - $80;//7F;
          Inc(i);
          for P := i to i + PixelCount - 1 do
          begin
            B := Byte(S[P]);
            (*if not Valid(BufferPos) then
            begin
              Log(['validation error']);
              Log(['Position', p]);
              Log(['Pixelcount', PixelCount]);
              for j := 1 to Length(S) do
                Log(['v', LeadZeroStr(j, 3), Byte(S[j])]);
            end; *)
            Check(BufferPos);
            LineBuffer[BufferPos] := B;
            Inc(BufferPos);
          end;
          Inc(i, PixelCount);
        end
        else if CurrentByte = $7F then // extended offset, there is always an extra byte
        begin
          //if CurrentParsingLine = 48 then if DebugIt then
            //Log([i, 'EXTENDED OFFSET']);
          Inc(BufferPos, CurrentByte);
          (*
          repeat
            Inc(i);
            B := Byte(S[i]);
            Inc(BufferPos, B);
            // the byte AFTER the row of 7F-bytes is 1) count byte 2) offset byte
            if B <> $7F then
              Break;
          until False;

          *)
{          Log(['CurrentParsingLine', CurrentParsingLine]);
          for j := 1 to Length(S) do
            Log(['v', LeadZeroStr(j, 3), Byte(S[j])]);
          raise exception.create('$7F'); }
          Inc(i); // go to the data-count-byte
        end

      until False;

    end;

    procedure TransferLine(Y: Integer);
    var
      i: Integer;
    begin
      //Log([length(linebuffer)]);
      for i := 0 to Length(LineBuffer) - 1 do
        DoPixel(i, Y, LineBuffer[i]);
    end;



var
  A: ^TSPRAnimationInfoRec;
  F: ^TSPRFrameRec;
  i, Lines, Q: Integer;
  PalIndex, XPos, YPos: Byte;
  x, y: integer;
  B: Byte;
  S: string;
  IncrementX: Byte;
  PixelCountSpecifier: Byte;
  CurrentByte: Byte;
  StringPos, CountedPixels: Integer;

begin
  if Index >= Length(fAnimationRecs) then
    raise Exception.Create('TSPRFile.ContructBitmap error 1');
  if Index >= Length(fFrames) then
    raise Exception.Create('TSPRFile.ContructBitmap error 2');

  A := @fAnimationRecs[Index];
  F := @fFrames[Index];

  BMP.Width := A^.ImageWidth;
  BMP.Height := A.ImageHeight;
  BMP.Clear(0); // clear with #transparent black
//  BMP.Canvas.Brush.Color := clBlack;
//  BMP.Canvas.FillRect(Rect(0, 0, A^.ImageWidth, A^.ImageHeight));

  SetLength(LineBuffer, A^.ImageWidth);

  for Lines := 0 to Length(F^.Data) - 1 do
  begin
    S := F^.Data[Lines];
    if Length(S) <= 2 then
      Continue;
//      Break;
    CurrentParsingLine := Lines;


(*    if DebugIt then
      if Lines = 48 then
      begin
        Log(['line 48']);
        for Q := 1 to Length(S) do
          Log([leadzerostr(q,3), Byte(S[Q])]);
      end; *)

    ParseLine(S, A^.XOffset);
    TransferLine(A^.YOffset + Lines);

  end;

end;


destructor TSPRFile.Destroy;
begin
  Clear;
  inherited;
end;

procedure TSPRFile.LoadFromStream(S: TStream);
var
  Ani: ^TSPRAnimationInfoRec;
  Fra: ^TSPRFrameRec;
  Rd, StreamPos: Integer;
  LineStr: string;
  HeaderStr: string;
  B: Byte;
  OK: Boolean;
  i: Integer;
  AniCounter, LineCounter: Integer;


    function ReadSpriteLine(var aLine: string): Boolean;
    var
      XOffset, LineSize, Eol, B: Byte;
      i, SavePos, Current, Count: Integer;
      EmptyLine: Boolean;
//      Buf: PBytes;
    begin

      aLine := '';
      Result := False;
      Count := 0;
      //SavePos := S.Position;

      while S.Read(B, 1) = 1 do
      begin
        Inc(Count);
        {$ifdef debug_winspr_loading}
        Debug(dmWinSprLoading, ['streampos, byte', LeadZeroStr(S.Position - 1, 6), B], 2);
        {$endif}
        aLine := aLine + Chr(B);
        if B = 128 then
          Break;
      end;

      if Count = 0 then
        Exit;

      //SetLength(aLine, Count);
      Result := True;
    end;


begin
  Clear;

  try

  {$ifdef debug_winspr_loading}
  Debug(dmWinSprLoading, ['start reading spr file']);
  Debug(dmWinSprLoading, ['streamsize', S.Size]);
  {$endif}

  // read header
  Rd := S.Read(fHeader, SizeOf(fHeader));

  if Rd <> SizeOf(fHeader) then
    raise Exception.Create('SPRFile error 1');
  if fHeader.Signature <> 'SRLE' then
    raise Exception.Create('SPRFile error 2');


  HeaderStr := fHeader.Signature;
  {$ifdef debug_winspr_loading}
  Debug(dmWinSprLoading, ['header', HeaderStr]);
  Debug(dmWinSprLoading, ['framecount', fRec.FrameCount]);
  Debug(dmWinSprLoading, ['frameoffset', frec.FirstFrameOffset]);
  {$endif}

  {.$define debug_winspr_loading}

  StreamPos := S.Seek(fHeader.FirstFrameOffset, soFromBeginning);

  if StreamPos <> fHeader.FirstFrameOffset then
    raise Exception.Create('SPRFile error 3');

  // init animation records
  SetLength(fAnimationRecs, fHeader.FrameCount);
  for i := 0 to fHeader.FrameCount - 1 do
    FillChar(fAnimationRecs[i], SizeOf(TSPRAnimationInfoRec), 0);

  // init sprite records
  SetLength(fFrames, fHeader.FrameCount);
  for i := 0 to fHeader.FrameCount - 1 do
    fFrames[i].Data := nil;

  // Walk through animations
  for AniCounter := 0 to fHeader.FrameCount - 1 do
  begin


    {$ifdef debug_winspr_loading}
    Debug(dmWinSprLoading, ['animationcounter', AniCounter]);
    Debug(dmWinSprLoading, ['streampos, hex streampos', S.Position, HexStr(S.Position)]);
    {$endif}

    Ani := @fAnimationRecs[AniCounter];


//    if S.Position = $A841 then
  //    S.Position := $A8AE;

    // read animation header
    Rd := S.Read(Ani^, SizeOf(TSPRAnimationInfoRec));

    if Rd <> SizeOf(TSPRAnimationInfoRec) then
      raise Exception.Create('SPRFile error 4 at frame ' + i2s(AniCounter) +
        ' at position ' + i2s(S.Position) + ' streamsize=' + i2s(s.size));

    {$ifdef debug_winspr_loading}
      Debug(dmWinSprLoading, ['ani.xoffset', Ani^.XOffset]);
      Debug(dmWinSprLoading, ['ani.yoffset', Ani^.YOffset]);
      Debug(dmWinSprLoading, ['ani.maxdatawidth', Ani^.MaxDataWidth]);
      Debug(dmWinSprLoading, ['ani.dataheight', Ani^.DataHeight]);
      Debug(dmWinSprLoading, ['ani.imagewidth', Ani^.ImageWidth]);
      Debug(dmWinSprLoading, ['ani.imageheight', Ani^.ImageHeight]);
    {$endif}

    // Make enough frames in the array
    Fra := @fFrames[AniCounter];
    SetLength(Fra^.Data, Ani^.DataHeight);

    LineCounter := 0;
    for i := 0 to Ani^.DataHeight - 1 do
    begin
      {$ifdef debug_winspr_loading}
      Debug(dmWinSprLoading, ['start reading line', i], 1);
      {$endif}

      if ReadSpriteLine(LineStr) then
      begin
        Inc(LineCounter);
        Fra^.Data[i] := LineStr;
      end
      else
        raise Exception.Create('readspriteline error');

//        Break;
      // else error
    end;

  end;


  except
    Clear;
    raise;
  end;

end;

procedure TSPRFile.PrepareFromStream(S: TStream);
begin
//
end;

{ TPALFile }

procedure TPALFile.Clear;
begin
  FillChar(fRec, SizeOf(fRec), 0);
  fColors := nil;
end;

destructor TPALFile.Destroy;
begin
  inherited;
end;

procedure TPALFile.LoadFromStream(S: TStream);
var
  Rd, i: Integer;
begin
  Clear;
  try
    if S.Read(fRec, SizeOf(fRec)) <> SizeOf(fRec) then
      raise Exception.Create('TPALFile error 1');
    if fRec.ID <> $20 then
      raise Exception.Create('TPALFile error 2');
    if fRec.Header <> 'LAP' then
      raise Exception.Create('TPALFile error 3');
    SetLength(fColors, fRec.EntryCount);
    for i := 0 to Length(fColors) - 1 do
      fColors[i] := 0;
    Rd := S.Read(fColors[0], fRec.EntryCount * 4);
    //Log(['pal entries', fRec.entrycount]);
  except
    Clear;
    raise;
  end;

end;

end.










    (*
    for i := 3 to Length(S) - 1 do // eind string = $80
    begin
      CurrentByte := Byte(S[i]);

      if CountedPixels = PixelCount then
      begin
        //windlg(['oe', currentbyte])
        Inc(IncrementX, CurrentByte);
      end;

      x := A^.XOffset + XPos + i - 2 + IncrementX;
      y := A^.YOffset + Lines;
      PalIndex := CurrentByte;
//      Log([x,y,palindex]);

      if Pal = nil then
        BMP.Canvas.Pixels[x, y] := RGB(PalIndex, PalIndex, PalIndex)// TColor(PalIndex)
      else begin
        if PalIndex >= Length(Pal.fColors) then
          raise Exception.CreateFmt('TSPRFile.ContructBitmap color index error %d %d', [PalIndex, Length(Pal.fColors)]);
        BMP.Canvas.Pixels[x, y] := Pal.fColors[PalIndex];
      end;

      Inc(CountedPixels);

    end; *)




    (*

    IncrementX := 0;
    CountedPixels := 0;
    XPos := Byte(S[1]);
    if XPos <= $7F then // common situation: x offset
    begin
      //if XPos = $7f then Log(['7F']);
      PixelCountSpecifier := Byte(S[2]);
      StringPos := 3;
    end
    else begin // check omitted xoffset (then its zero), then this byte is the pixelcountspecifier
      XPos := 0;
      PixelCountSpecifier := Byte(S[1]);
      StringPos := 2;
    end;

    repeat
      CurrentByte := Byte(S[StringPos]);
      //if lines = 0 then Log([stringpos, currentbyte]);
      if CountedPixels = PixelCountSpecifier - 128 then // check sublines
      begin
        Inc(IncrementX, CurrentByte);
        Inc(StringPos);
        CountedPixels := 0;
        PixelCountSpecifier := Byte(S[StringPos]);
        Inc(StringPos); // now go to the next pixel
        //log(['trans', lines, currentbyte, pixelcountspecifier, incrementx]);
      end
      else begin
        x := A^.XOffset + XPos + {StringPos - 3} + IncrementX;
        y := A^.YOffset + Lines;
        DoPixel(x, y, CurrentByte);
        Inc(StringPos);
        Inc(CountedPixels);
        Inc(IncrementX);
      end;
      if StringPos >= Length(S) then
        Break;
      if CurrentByte = 128 then // end of line marker
        Break;
    until False;
    *)

  TLVLTerrain = packed record
     B0, B1, B2, B3: Byte;
  end;
(*

brick objects in brickos.spr file

0 = 1        0/0

1 = 10       1/10

2 = 14       11/24

3 = 7        25/31

4 = 7        32/38

5 = 9        39/47

6 = 18       48/65

7 = 20       66/85

8 = 14       86/99

9 = 4        100/103

*)

(*

  LVL restrictions:

  width = 1584

  objects:
  --------
  x position must be dividable by 8 (OUCH)


  terrain:
  --------
  x limits: -16 and 4079
  y limits: -256 and 255

  steel:
  ------
  x limits: -16 and 2028 dividable by 4
  y limits: 0 to 508 dividable by 4




*)

