{$include lem_directives.inc}

unit FPickTerrain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, UMx, GR32, Math, umisc,
  LemStrings,
  LemCore, StdCtrls, Buttons;

const
  MAX_PICTURE_WIDTH = 512;
  MAX_PICTURE_HEIGHT = 256;

type
  TTerrainMatrixData = class(TCellData)
  private
    Index: Integer;
    Bmp: TBitmap32; // ref
    Frame: TBitmap32; // instance
  protected
    procedure Clear; override;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TTOMode = (
    tNone,
    tTerrains,
    tObjects
  );

  TPickTerrainForm = class(TForm)
    TerrainMatrix: TSuperMatrix;
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure TerrainMatrix_GetCellData(Sender: TObject; aCol, aRow: Integer; out Data: TCellData);
    procedure TerrainMatrix_PaintCell(Sender: TObject; Params: TCellParams; Data: TCellData);
    procedure TerrainMatrixCellDblClick(Sender: TObject; aCol,
      aRow: Integer);
  private
    fGraphicSet: TBaseGraphicSet;//Integer;
    fCellData: TTerrainMatrixData;
    fScale: Integer;
    fMode: TTOMode;
    fOK: Boolean;
//    fObjBmp: TBitmap32;
    procedure AdjustMatrix;
    procedure SetGraphicSet(const Value: TBaseGraphicSet);
    procedure SetMode(const Value: TTOMode);
    function GetCurrentID: Integer;
    procedure SetCurrentID(const Value: Integer);
  public
    property GraphicSet: TBaseGraphicSet read fGraphicSet write SetGraphicSet;
//    property Graph: TBaseGraphicSet read fGraph write SetGraph;
    property Mode: TTOMode read fMode write SetMode;
    property CurrentID: Integer read GetCurrentID write SetCurrentID;
  end;

function PickListTerrain(aGraphicSet: TBaseGraphicSet; var aTerrainID: Integer): Boolean;
function PicklistObject(aGraphicSet: TBaseGraphicSet; var aObjectID: Integer): Boolean;

implementation

{$R *.dfm}

const
  picture_column = 1;

function DoPicklist(aGraphicSet: TBaseGraphicSet; aMode: TTOMode; var aID: Integer): Boolean;
var
  F: TPickTerrainForm;
begin
  Result := False;
//  aTerrainID := -1;
  F := TPickTerrainForm.Create(nil);
  with F do
  begin
    Position := poDesktopCenter;
    GraphicSet := aGraphicSet;
    Mode := aMode;
    CurrentID := aID;
    if ShowModal = mrOK then
    begin
      Result := True;
      aID := CurrentID;
    end;
    Free;
  end;
end;

function PicklistTerrain(aGraphicSet: TBaseGraphicSet; var aTerrainID: Integer): Boolean;
begin
  Result := DoPicklist(aGraphicSet, tTerrains, aTerrainID);
end;

function PicklistObject(aGraphicSet: TBaseGraphicSet; var aObjectID: Integer): Boolean;
begin
  Result := DoPicklist(aGraphicSet, tObjects, aObjectID);
end;

{ TTerrainMatrixData }

procedure TTerrainMatrixData.Clear;
begin
  inherited Clear;
  Index := 0;
  Bmp := nil;
//  Frame
end;


constructor TTerrainMatrixData.Create;
begin
  inherited;
  Frame := TBitmap32.Create;
end;

destructor TTerrainMatrixData.Destroy;
begin
  Frame.Free;
  inherited;
end;

{ TPickTerrainForm }

procedure TPickTerrainForm.AdjustMatrix;
var
  W, H, i: Integer;
  Bmp: TBitmap32;
begin
  if fGraphicSet = nil then
    Exit;
  if fMode = tNone then
    Exit;

  case fMode of
    tTerrains : Caption := 'Pick Terrain';
    tObjects  : Caption := 'Pick Object';
  end;


  with TerrainMatrix do
  begin
    Bmp := nil;
    PaintLock;
    try
      case fMode of
        tNone: RowCount := 1;
        tTerrains: RowCount := GraphicSet.MetaTerrainCollection.Count;// PictureMgr.TerrainBitmapCount[fGraphicSet];
        tObjects: RowCount := GraphicSet.MetaObjectCollection.Count;// PictureMgr.ObjectBitmapCount[fGraphicSet];
      end;


      for i := 0 to RowCount - 1 do
      begin
        //Bmp := PictureMgr.TerrainBitmaps[fGraphicSet, i];
        case fMode of
          tTerrains: Bmp := GraphicSet.TerrainBitmaps[i];
          tObjects: Bmp := GraphicSet.ObjectBitmaps[i];
        end;
        if Bmp <> nil then
        begin
//          if Bmp.Width * fScale > ColWidths[2]  * fScale then ColWidths[2] := Bmp.Width  * fScale;
          W := Bmp.Width * fScale - fScale + 4 {extra};
          if W > MAX_PICTURE_WIDTH then W := MAX_PICTURE_WIDTH;
          //ColWidths[picture_column] := Max(ColWidths[picture_column], Bmp.Width * fScale - fScale + 4 {extra});
          ColWidths[picture_column] := Max(ColWidths[picture_column], W);

          if fMode = tObjects then
            H := GraphicSet.MetaObjectCollection[i].Height
          else
            H := Bmp.Height;
          H := H * fScale - fScale + 4;
          if H > MAX_PICTURE_HEIGHT then H := MAX_PICTURE_HEIGHT;
          RowHeights[i] := Max(H(*H * fScale - fScale + 4{extra}*), 19);
          
        end
      end;
    finally
      PaintUnlock;
    end;
  end;
end;

procedure TPickTerrainForm.SetGraphicSet(const Value: TBaseGraphicSet);
begin
  fGraphicSet := Value;
  AdjustMatrix;
end;

procedure TPickTerrainForm.FormCreate(Sender: TObject);
begin
  fGraphicSet := nil;//-1;
  fScale := 1;
  fCellData := TTerrainMatrixData.Create;
//  fObjBmp := TBitmap32.Create;
end;

procedure TPickTerrainForm.FormDestroy(Sender: TObject);
begin
  fCellData.Free;
//  fObjBitmap.Free;
end;

procedure TPickTerrainForm.TerrainMatrix_GetCellData(Sender: TObject; aCol, aRow: Integer; out Data: TCellData);
begin
  Data := fCellData;
  //fCellData.Index := aRow;
//  fCellData.Bmp := nil;
  fCellData.Clear;

  if aCol = 0 then
    fCellData.Text := i2s(aRow);

  if aCol = picture_column then
    if fGraphicSet <> nil then //>= 0 then
    begin
      case fMode of
        tTerrains: fCellData.Bmp := GraphicSet.TerrainBitmaps[aRow];// PictureMgr.TerrainBitmaps[fGraphicSet, aRow];
        tObjects:
          begin
//            fCellData.Bmp := GraphicSet.ObjectBitmaps[aRow];// PictureMgr.ObjectBitmaps[fGraphicSet, aRow];
            GraphicSet.GetObjectFrame(aRow, 0, fCellData.Frame);
            fCellData.Bmp := fCellData.Frame;
          end;
      end;
    end;
end;

procedure TPickTerrainForm.TerrainMatrix_PaintCell(Sender: TObject; Params: TCellParams; Data: TCellData);
{-------------------------------------------------------------------------------
  Paint terrain bitmap
-------------------------------------------------------------------------------}
var
  Bmp: TBitmap32;
  R, D: TRect;
  OldDrawMode: TDrawMode;
begin
  if fGraphicSet = nil then //< 0 then
    Exit;
  with Params do
  begin
    if (pCol = picture_column) and (pStage = pcsContents) then
    begin
      Bmp := fCellData.Bmp;
      if Bmp = nil then
        Exit;
      with Bmp do
        R := Rect(0, 0, Width - 1, Height - 1);
      D := R;
      D.Right := D.Right * fScale;
      D.Bottom := D.Bottom * fScale;

      OldDrawMode := Bmp.DrawMode;
      Bmp.Drawmode := dmTransparent;
      try
        Bmp.DrawTo(pCellCanvas.Handle, D, R);
      finally
        Bmp.DrawMode := OldDrawMode;
      end;
    end;
  end;
end;


procedure TPickTerrainForm.SetMode(const Value: TTOMode);
begin
  if fMode = Value then Exit;
  fMode := Value;

  AdjustMatrix;
end;

function TPickTerrainForm.GetCurrentID: Integer;
begin
  Result := TerrainMatrix.Row;
end;

procedure TPickTerrainForm.TerrainMatrixCellDblClick(Sender: TObject; aCol, aRow: Integer);
begin
  ModalResult := mrOK;
end;

procedure TPickTerrainForm.SetCurrentID(const Value: Integer);
begin
  if Between(Value, 0, TerrainMatrix.RowCount - 1) then
    TerrainMatrix.Row := Value;
end;

end.
//
